@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2015-2017 Canonical Ltd
+ * Copyright (C) 2015-2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -29,7 +29,6 @@ import (
 
 	"github.com/snapcore/snapd/asserts"
 	"github.com/snapcore/snapd/snap"
-
 	"github.com/snapcore/snapd/testutil"
 )
 
@@ -119,43 +118,14 @@ func (s *attrConstraintsSuite) TestSimple(c *C) {
 	c.Check(err, ErrorMatches, `attribute "bar" has constraints but is unset`)
 }
 
-func (s *attrConstraintsSuite) TestSimpleAnchorsVsRegexpAlt(c *C) {
+func (s *attrConstraintsSuite) TestMissingCheck(c *C) {
 	m, err := asserts.ParseHeaders([]byte(`attrs:
-  bar: BAR|BAZ`))
+  foo: $MISSING`))
 	c.Assert(err, IsNil)
 
 	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
 	c.Assert(err, IsNil)
-
-	plug := attrerObject(map[string]interface{}{
-		"bar": "BAR",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, IsNil)
-
-	plug = attrerObject(map[string]interface{}{
-		"bar": "BARR",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, ErrorMatches, `attribute "bar" value "BARR" does not match \^\(BAR|BAZ\)\$`)
-
-	plug = attrerObject(map[string]interface{}{
-		"bar": "BBAZ",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, ErrorMatches, `attribute "bar" value "BAZZ" does not match \^\(BAR|BAZ\)\$`)
-
-	plug = attrerObject(map[string]interface{}{
-		"bar": "BABAZ",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, ErrorMatches, `attribute "bar" value "BABAZ" does not match \^\(BAR|BAZ\)\$`)
-
-	plug = attrerObject(map[string]interface{}{
-		"bar": "BARAZ",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, ErrorMatches, `attribute "bar" value "BARAZ" does not match \^\(BAR|BAZ\)\$`)
+	c.Check(asserts.RuleFeature(cstrs, "dollar-attr-constraints"), Equals, true)
 }
 
 func (s *attrConstraintsSuite) TestNested(c *C) {
@@ -208,80 +178,68 @@ baz: BAZ
 	c.Check(err, ErrorMatches, `attribute "bar\.bar2" must be a scalar or list`)
 }
 
-func (s *attrConstraintsSuite) TestAlternative(c *C) {
+func (s *attrConstraintsSuite) TestAlternativeMatchingComplex(c *C) {
+	toMatch := attrs(`
+mnt: [{what: "/dev/x*", where: "/foo/*", options: ["rw", "nodev"]}, {what: "/bar/*", where: "/baz/*", options: ["rw", "bind"]}]
+`)
+
 	m, err := asserts.ParseHeaders([]byte(`attrs:
-  -
-    foo: FOO
-    bar: BAR
-  -
-    foo: FOO
-    bar: BAZ`))
-	c.Assert(err, IsNil)
-
-	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].([]interface{}))
-	c.Assert(err, IsNil)
-
-	plug := attrerObject(map[string]interface{}{
-		"foo": "FOO",
-		"bar": "BAR",
-		"baz": "BAZ",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, IsNil)
-
-	plug = attrerObject(map[string]interface{}{
-		"foo": "FOO",
-		"bar": "BAZ",
-		"baz": "BAZ",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, IsNil)
-
-	plug = attrerObject(map[string]interface{}{
-		"foo": "FOO",
-		"bar": "BARR",
-		"baz": "BAR",
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, ErrorMatches, `no alternative matches: attribute "bar" value "BARR" does not match \^\(BAR\)\$`)
-}
-
-func (s *attrConstraintsSuite) TestNestedAlternative(c *C) {
-	m, err := asserts.ParseHeaders([]byte(`attrs:
-  foo: FOO
-  bar:
-    bar1: BAR1
-    bar2:
-      - BAR2
-      - BAR22`))
+  mnt:
+    -
+      what: /(bar/|dev/x)\*
+      where: /(foo|baz)/\*
+      options: rw|bind|nodev`))
 	c.Assert(err, IsNil)
 
 	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
 	c.Assert(err, IsNil)
 
-	err = cstrs.Check(attrs(`
-foo: FOO
-bar:
-  bar1: BAR1
-  bar2: BAR2
-`), nil)
+	err = cstrs.Check(toMatch, nil)
 	c.Check(err, IsNil)
 
-	err = cstrs.Check(attrs(`
-foo: FOO
-bar:
-  bar1: BAR1
-  bar2: BAR22
-`), nil)
+	m, err = asserts.ParseHeaders([]byte(`attrs:
+  mnt:
+    -
+      what: /dev/x\*
+      where: /foo/\*
+      options:
+        - nodev
+        - rw
+    -
+      what: /bar/\*
+      where: /baz/\*
+      options:
+        - rw
+        - bind`))
+	c.Assert(err, IsNil)
+
+	cstrsExtensive, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
+	c.Assert(err, IsNil)
+
+	err = cstrsExtensive.Check(toMatch, nil)
 	c.Check(err, IsNil)
 
-	err = cstrs.Check(attrs(`
-foo: FOO
-bar:
-  bar1: BAR1
-  bar2: BAR3
-`), nil)
-	c.Check(err, ErrorMatches, `no alternative for attribute "bar\.bar2" matches: attribute "bar\.bar2" value "BAR3" does not match \^\(BAR2\)\$`)
+	// not matching case
+	m, err = asserts.ParseHeaders([]byte(`attrs:
+  mnt:
+    -
+      what: /dev/x\*
+      where: /foo/\*
+      options:
+        - rw
+    -
+      what: /bar/\*
+      where: /baz/\*
+      options:
+        - rw
+        - bind`))
+	c.Assert(err, IsNil)
+
+	cstrsExtensiveNoMatch, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
+	c.Assert(err, IsNil)
+
+	err = cstrsExtensiveNoMatch.Check(toMatch, nil)
+	c.Check(err, ErrorMatches, `no alternative for attribute "mnt\.0" matches: no alternative for attribute "mnt\.0.options\.1" matches:.*`)
 }
 
 func (s *attrConstraintsSuite) TestOtherScalars(c *C) {
@@ -298,13 +256,6 @@ foo: 1
 bar: true
 `), nil)
 	c.Check(err, IsNil)
-
-	plug := attrerObject(map[string]interface{}{
-		"foo": int64(1),
-		"bar": true,
-	})
-	err = cstrs.Check(plug, nil)
-	c.Check(err, IsNil)
 }
 
 func (s *attrConstraintsSuite) TestCompileErrors(c *C) {
@@ -313,16 +264,6 @@ func (s *attrConstraintsSuite) TestCompileErrors(c *C) {
 	})
 	c.Check(err, ErrorMatches, `cannot compile "foo" constraint "\[": error parsing regexp:.*`)
 
-	_, err = asserts.CompileAttributeConstraints(map[string]interface{}{
-		"foo": []interface{}{"foo", "["},
-	})
-	c.Check(err, ErrorMatches, `cannot compile "foo/alt#2/" constraint "\[": error parsing regexp:.*`)
-
-	_, err = asserts.CompileAttributeConstraints(map[string]interface{}{
-		"foo": []interface{}{"foo", []interface{}{"bar", "baz"}},
-	})
-	c.Check(err, ErrorMatches, `cannot nest alternative constraints directly at "foo/alt#2/"`)
-
 	_, err = asserts.CompileAttributeConstraints("FOO")
 	c.Check(err, ErrorMatches, `first level of non alternative constraints must be a set of key-value contraints`)
 
@@ -340,52 +281,15 @@ func (s *attrConstraintsSuite) TestCompileErrors(c *C) {
 		_, err := asserts.CompileAttributeConstraints(map[string]interface{}{
 			"foo": wrong,
 		})
-		c.Check(err, ErrorMatches, fmt.Sprintf(`cannot compile "foo" constraint "%s": not a valid \$SLOT\(\)/\$PLUG\(\) constraint`, regexp.QuoteMeta(wrong)))
+		c.Check(err, ErrorMatches, fmt.Sprintf(`cannot compile "foo" constraint "%s": not a valid \$SLOT\(\)/\$PLUG\(\)/\$PLUG_PUBLISHER_ID/\$SLOT_PUBLISHER_ID constraint`, regexp.QuoteMeta(wrong)))
 
 	}
 }
 
-func (s *attrConstraintsSuite) TestMatchingListsSimple(c *C) {
-	m, err := asserts.ParseHeaders([]byte(`attrs:
-  foo: /foo/.*`))
-	c.Assert(err, IsNil)
-
-	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
-	c.Assert(err, IsNil)
-
-	err = cstrs.Check(attrs(`
-foo: ["/foo/x", "/foo/y"]
-`), nil)
-	c.Check(err, IsNil)
-
-	err = cstrs.Check(attrs(`
-foo: ["/foo/x", "/foo"]
-`), nil)
-	c.Check(err, ErrorMatches, `attribute "foo\.1" value "/foo" does not match \^\(/foo/\.\*\)\$`)
-}
-
-func (s *attrConstraintsSuite) TestMissingCheck(c *C) {
-	m, err := asserts.ParseHeaders([]byte(`attrs:
-  foo: $MISSING`))
-	c.Assert(err, IsNil)
-
-	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
-	c.Assert(err, IsNil)
-	c.Check(asserts.RuleFeature(cstrs, "dollar-attr-constraints"), Equals, true)
-
-	err = cstrs.Check(attrs(`
-bar: baz
-`), nil)
-	c.Check(err, IsNil)
-
-	err = cstrs.Check(attrs(`
-foo: ["x"]
-`), nil)
-	c.Check(err, ErrorMatches, `attribute "foo" is constrained to be missing but is set`)
-}
-
 type testEvalAttr struct {
-	comp func(side string, arg string) (interface{}, error)
+	comp            func(side string, arg string) (interface{}, error)
+	plugPublisherID string
+	slotPublisherID string
 }
 
 func (ca testEvalAttr) SlotAttr(arg string) (interface{}, error) {
@@ -396,6 +300,14 @@ func (ca testEvalAttr) PlugAttr(arg string) (interface{}, error) {
 	return ca.comp("plug", arg)
 }
 
+func (ca testEvalAttr) PlugPublisherID() string {
+	return ca.plugPublisherID
+}
+
+func (ca testEvalAttr) SlotPublisherID() string {
+	return ca.slotPublisherID
+}
+
 func (s *attrConstraintsSuite) TestEvalCheck(c *C) {
 	m, err := asserts.ParseHeaders([]byte(`attrs:
   foo: $SLOT(foo)
@@ -421,7 +333,7 @@ bar: bar
 	err = cstrs.Check(attrs(`
 foo: foo
 bar: bar.baz
-`), testEvalAttr{comp1})
+`), testEvalAttr{comp: comp1})
 	c.Check(err, IsNil)
 
 	c.Check(calls, DeepEquals, map[[2]string]bool{
@@ -439,7 +351,7 @@ bar: bar.baz
 	err = cstrs.Check(attrs(`
 foo: foo
 bar: bar.baz
-`), testEvalAttr{comp2})
+`), testEvalAttr{comp: comp2})
 	c.Check(err, ErrorMatches, `attribute "bar" constraint \$PLUG\(bar\.baz\) cannot be evaluated: boom`)
 
 	comp3 := func(op string, arg string) (interface{}, error) {
@@ -452,32 +364,72 @@ bar: bar.baz
 	err = cstrs.Check(attrs(`
 foo: foo
 bar: bar.baz
-`), testEvalAttr{comp3})
+`), testEvalAttr{comp: comp3})
 	c.Check(err, ErrorMatches, `attribute "foo" does not match \$SLOT\(foo\): foo != other-value`)
 }
 
-func (s *attrConstraintsSuite) TestMatchingListsMap(c *C) {
+func (s *attrConstraintsSuite) TestCheckWithAttrPlugPublisherID(c *C) {
 	m, err := asserts.ParseHeaders([]byte(`attrs:
-  foo:
-    p: /foo/.*`))
+  my-attr: $PLUG_PUBLISHER_ID`))
 	c.Assert(err, IsNil)
 
 	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
 	c.Assert(err, IsNil)
+	c.Check(asserts.RuleFeature(cstrs, "publisher-id-constraints"), Equals, true)
+
+	helper := testEvalAttr{plugPublisherID: "my-account"}
 
 	err = cstrs.Check(attrs(`
-foo: [{p: "/foo/x"}, {p: "/foo/y"}]
+my-attr: my-account
 `), nil)
+	c.Check(err, ErrorMatches, `attribute "my-attr" cannot be matched without context`)
+
+	err = cstrs.Check(attrs(`
+my-attr: my-account
+`), helper)
 	c.Check(err, IsNil)
 
 	err = cstrs.Check(attrs(`
-foo: [{p: "zzz"}, {p: "/foo/y"}]
-`), nil)
-	c.Check(err, ErrorMatches, `attribute "foo\.0\.p" value "zzz" does not match \^\(/foo/\.\*\)\$`)
+my-attr: other-account
+`), helper)
+	c.Check(err, ErrorMatches, `.*attribute "my-attr" does not match \$PLUG_PUBLISHER_ID\: other-account != my-account`)
+
+	err = cstrs.Check(attrs(`
+my-attr: 1
+`), helper)
+	c.Check(err, ErrorMatches, `.*attribute "my-attr" is not expected string type: int64`)
 }
 
-func (s *attrConstraintsSuite) TestAlwaysMatchAttributeConstraints(c *C) {
-	c.Check(asserts.AlwaysMatchAttributes.Check(nil, nil), IsNil)
+func (s *attrConstraintsSuite) TestCheckWithAttrSlotPublisherID(c *C) {
+	m, err := asserts.ParseHeaders([]byte(`attrs:
+  my-attr: $SLOT_PUBLISHER_ID`))
+	c.Assert(err, IsNil)
+
+	cstrs, err := asserts.CompileAttributeConstraints(m["attrs"].(map[string]interface{}))
+	c.Assert(err, IsNil)
+	c.Check(asserts.RuleFeature(cstrs, "publisher-id-constraints"), Equals, true)
+
+	helper := testEvalAttr{slotPublisherID: "my-account"}
+
+	err = cstrs.Check(attrs(`
+my-attr: my-account
+`), nil)
+	c.Check(err, ErrorMatches, `attribute "my-attr" cannot be matched without context`)
+
+	err = cstrs.Check(attrs(`
+my-attr: my-account
+`), helper)
+	c.Check(err, IsNil)
+
+	err = cstrs.Check(attrs(`
+my-attr: other-account
+`), helper)
+	c.Check(err, ErrorMatches, `.*attribute "my-attr" does not match \$SLOT_PUBLISHER_ID\: other-account != my-account`)
+
+	err = cstrs.Check(attrs(`
+my-attr: 1
+`), helper)
+	c.Check(err, ErrorMatches, `.*attribute "my-attr" is not expected string type: int64`)
 }
 
 func (s *attrConstraintsSuite) TestNeverMatchAttributeConstraints(c *C) {
@@ -800,6 +752,7 @@ func (s *plugSlotRulesSuite) TestCompilePlugRuleInstalationConstraintsIDConstrai
 	rule, err := asserts.CompilePlugRule("iface", map[string]interface{}{
 		"allow-installation": map[string]interface{}{
 			"plug-snap-type": []interface{}{"core", "kernel", "gadget", "app"},
+			"plug-snap-id":   []interface{}{"snapidsnapidsnapidsnapidsnapid01", "snapidsnapidsnapidsnapidsnapid02"},
 		},
 	})
 	c.Assert(err, IsNil)
@@ -807,6 +760,7 @@ func (s *plugSlotRulesSuite) TestCompilePlugRuleInstalationConstraintsIDConstrai
 	c.Assert(rule.AllowInstallation, HasLen, 1)
 	cstrs := rule.AllowInstallation[0]
 	c.Check(cstrs.PlugSnapTypes, DeepEquals, []string{"core", "kernel", "gadget", "app"})
+	c.Check(cstrs.PlugSnapIDs, DeepEquals, []string{"snapidsnapidsnapidsnapidsnapid01", "snapidsnapidsnapidsnapidsnapid02"})
 }
 
 func (s *plugSlotRulesSuite) TestCompilePlugRuleInstallationConstraintsOnClassic(c *C) {
@@ -1277,19 +1231,19 @@ func (s *plugSlotRulesSuite) TestCompilePlugRuleErrors(c *C) {
 		{`iface:
   allow-connection:
     slot-snap-ids:
-      - foo`, `allow-connection in plug rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, slot-publisher-id, slot-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
+      - foo`, `allow-connection in plug rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, slot-publisher-id, slot-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-core-desktop, on-store, on-brand, on-model`},
 		{`iface:
   deny-connection:
     slot-snap-ids:
-      - foo`, `deny-connection in plug rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, slot-publisher-id, slot-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
+      - foo`, `deny-connection in plug rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, slot-publisher-id, slot-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-core-desktop, on-store, on-brand, on-model`},
 		{`iface:
   allow-auto-connection:
     slot-snap-ids:
-      - foo`, `allow-auto-connection in plug rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, slot-publisher-id, slot-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
+      - foo`, `allow-auto-connection in plug rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, slot-publisher-id, slot-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-core-desktop, on-store, on-brand, on-model`},
 		{`iface:
   deny-auto-connection:
     slot-snap-ids:
-      - foo`, `deny-auto-connection in plug rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, slot-publisher-id, slot-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
+      - foo`, `deny-auto-connection in plug rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, slot-publisher-id, slot-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-core-desktop, on-store, on-brand, on-model`},
 		{`iface:
   allow-connect: true`, `plug rule for interface "iface" must specify at least one of allow-installation, deny-installation, allow-connection, deny-connection, allow-auto-connection, deny-auto-connection`},
 		{`iface:
@@ -1644,6 +1598,7 @@ func (s *plugSlotRulesSuite) TestCompileSlotRuleInstallationConstraintsIDConstra
 	rule, err := asserts.CompileSlotRule("iface", map[string]interface{}{
 		"allow-installation": map[string]interface{}{
 			"slot-snap-type": []interface{}{"core", "kernel", "gadget", "app"},
+			"slot-snap-id":   []interface{}{"snapidsnapidsnapidsnapidsnapid01", "snapidsnapidsnapidsnapidsnapid02"},
 		},
 	})
 	c.Assert(err, IsNil)
@@ -1651,6 +1606,7 @@ func (s *plugSlotRulesSuite) TestCompileSlotRuleInstallationConstraintsIDConstra
 	c.Assert(rule.AllowInstallation, HasLen, 1)
 	cstrs := rule.AllowInstallation[0]
 	c.Check(cstrs.SlotSnapTypes, DeepEquals, []string{"core", "kernel", "gadget", "app"})
+	c.Check(cstrs.SlotSnapIDs, DeepEquals, []string{"snapidsnapidsnapidsnapidsnapid01", "snapidsnapidsnapidsnapidsnapid02"})
 }
 
 func (s *plugSlotRulesSuite) TestCompileSlotRuleInstallationConstraintsOnClassic(c *C) {
@@ -1804,6 +1760,7 @@ func (s *plugSlotRulesSuite) TestCompileSlotRuleInstallationConstraintsSlotNames
 func (s *plugSlotRulesSuite) TestCompileSlotRuleConnectionConstraintsIDConstraints(c *C) {
 	rule, err := asserts.CompileSlotRule("iface", map[string]interface{}{
 		"allow-connection": map[string]interface{}{
+			"slot-snap-type":    []interface{}{"core"},
 			"plug-snap-type":    []interface{}{"core", "kernel", "gadget", "app"},
 			"plug-snap-id":      []interface{}{"snapidsnapidsnapidsnapidsnapid01", "snapidsnapidsnapidsnapidsnapid02"},
 			"plug-publisher-id": []interface{}{"pubidpubidpubidpubidpubidpubid09", "canonical", "$SAME"},
@@ -1813,6 +1770,7 @@ func (s *plugSlotRulesSuite) TestCompileSlotRuleConnectionConstraintsIDConstrain
 
 	c.Assert(rule.AllowConnection, HasLen, 1)
 	cstrs := rule.AllowConnection[0]
+	c.Check(cstrs.SlotSnapTypes, DeepEquals, []string{"core"})
 	c.Check(cstrs.PlugSnapTypes, DeepEquals, []string{"core", "kernel", "gadget", "app"})
 	c.Check(cstrs.PlugSnapIDs, DeepEquals, []string{"snapidsnapidsnapidsnapidsnapid01", "snapidsnapidsnapidsnapidsnapid02"})
 	c.Check(cstrs.PlugPublisherIDs, DeepEquals, []string{"pubidpubidpubidpubidpubidpubid09", "canonical", "$SAME"})
@@ -2096,6 +2054,10 @@ func (s *plugSlotRulesSuite) TestCompileSlotRuleErrors(c *C) {
     plug-snap-id:
       - foo`, `plug-snap-id in allow-connection in slot rule for interface "iface" contains an invalid element: "foo"`},
 		{`iface:
+  allow-connection:
+    slot-snap-type:
+      - foo`, `slot-snap-type in allow-connection in slot rule for interface "iface" contains an invalid element: "foo"`},
+		{`iface:
   allow-connection:
     plug-snap-type:
       - foo`, `plug-snap-type in allow-connection in slot rule for interface "iface" contains an invalid element: "foo"`},
@@ -2114,19 +2076,19 @@ func (s *plugSlotRulesSuite) TestCompileSlotRuleErrors(c *C) {
 		{`iface:
   allow-connection:
     plug-snap-ids:
-      - foo`, `allow-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
+      - foo`, `allow-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-core-desktop, on-store, on-brand, on-model`},
 		{`iface:
   deny-connection:
     plug-snap-ids:
-      - foo`, `deny-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
+      - foo`, `deny-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-core-desktop, on-store, on-brand, on-model`},
 		{`iface:
   allow-auto-connection:
     plug-snap-ids:
-      - foo`, `allow-auto-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
+      - foo`, `allow-auto-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-core-desktop, on-store, on-brand, on-model`},
 		{`iface:
   deny-auto-connection:
     plug-snap-ids:
-      - foo`, `deny-auto-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-store, on-brand, on-model`},
+      - foo`, `deny-auto-connection in slot rule for interface "iface" must specify at least one of plug-names, slot-names, plug-attributes, slot-attributes, slot-snap-type, plug-snap-type, plug-publisher-id, plug-snap-id, slots-per-plug, plugs-per-slot, on-classic, on-core-desktop, on-store, on-brand, on-model`},
 		{`iface:
   allow-connect: true`, `slot rule for interface "iface" must specify at least one of allow-installation, deny-installation, allow-connection, deny-connection, allow-auto-connection, deny-auto-connection`},
 		{`iface:
@@ -2263,6 +2225,7 @@ func (s *plugSlotRulesSuite) TestSlotRuleFeatures(c *C) {
 }
 
 func (s *plugSlotRulesSuite) TestValidOnStoreBrandModel(c *C) {
+	// more extensive testing is now done in deviceScopeConstraintSuite
 	tests := []struct {
 		constr string
 		value  string
@@ -2272,43 +2235,20 @@ func (s *plugSlotRulesSuite) TestValidOnStoreBrandModel(c *C) {
 		{"on-store", "foo", true},
 		{"on-store", "F_o-O88", true},
 		{"on-store", "foo!", false},
-		{"on-store", "foo.", false},
-		{"on-store", "foo/", false},
 		{"on-brand", "", false},
 		// custom set brands (length 2-28)
 		{"on-brand", "dwell", true},
 		{"on-brand", "Dwell", false},
 		{"on-brand", "dwell-88", true},
 		{"on-brand", "dwell_88", false},
-		{"on-brand", "dwell.88", false},
-		{"on-brand", "dwell:88", false},
-		{"on-brand", "dwell!88", false},
-		{"on-brand", "a", false},
-		{"on-brand", "ab", true},
 		{"on-brand", "0123456789012345678901234567", true},
 		// snappy id brands (fixed length 32)
 		{"on-brand", "01234567890123456789012345678", false},
-		{"on-brand", "012345678901234567890123456789", false},
-		{"on-brand", "0123456789012345678901234567890", false},
 		{"on-brand", "01234567890123456789012345678901", true},
-		{"on-brand", "abcdefghijklmnopqrstuvwxyz678901", true},
-		{"on-brand", "ABCDEFGHIJKLMNOPQRSTUVWCYZ678901", true},
-		{"on-brand", "ABCDEFGHIJKLMNOPQRSTUVWCYZ678901X", false},
-		{"on-brand", "ABCDEFGHIJKLMNOPQ!STUVWCYZ678901", false},
-		{"on-brand", "ABCDEFGHIJKLMNOPQ_STUVWCYZ678901", false},
-		{"on-brand", "ABCDEFGHIJKLMNOPQ-STUVWCYZ678901", false},
 		{"on-model", "", false},
-		{"on-model", "/", false},
 		{"on-model", "dwell/dwell1", true},
 		{"on-model", "dwell", false},
 		{"on-model", "dwell/", false},
-		{"on-model", "dwell//dwell1", false},
-		{"on-model", "dwell/-dwell1", false},
-		{"on-model", "dwell/dwell1-", false},
-		{"on-model", "dwell/dwell1-23", true},
-		{"on-model", "dwell/dwell1!", false},
-		{"on-model", "dwell/dwe_ll1", false},
-		{"on-model", "dwell/dwe.ll1", false},
 	}
 
 	check := func(constr, value string, valid bool) {
