@@ -20,6 +20,7 @@
 package main
 
 import (
+	"context"
 	"os"
 	"os/user"
 	"time"
@@ -29,12 +30,14 @@ import (
 	"github.com/snapcore/snapd/client"
 	"github.com/snapcore/snapd/cmd/snaplock/runinhibit"
 	"github.com/snapcore/snapd/image"
+	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/sandbox/cgroup"
 	"github.com/snapcore/snapd/sandbox/selinux"
+	"github.com/snapcore/snapd/seed/seedwriter"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/store"
 	"github.com/snapcore/snapd/store/tooling"
-	usersessionclient "github.com/snapcore/snapd/usersession/client"
+	"github.com/snapcore/snapd/testutil"
 )
 
 var RunMain = run
@@ -94,7 +97,8 @@ var (
 
 	IsStopping = isStopping
 
-	GetSnapDirOptions = getSnapDirOptions
+	GetSnapDirOptions                   = getSnapDirOptions
+	SnapInstancesAndComponentsFromNames = snapInstancesAndComponentsFromNames
 )
 
 func HiddenCmd(descr string, completeHidden bool) *cmdInfo {
@@ -108,11 +112,19 @@ func HiddenCmd(descr string, completeHidden bool) *cmdInfo {
 type ChangeTimings = changeTimings
 
 func NewInfoWriter(w writeflusher) *infoWriter {
+	return NewInfoWriterWithFmtTime(w, nil)
+}
+
+func NewInfoWriterWithFmtTime(w writeflusher, fmtTime func(time.Time) string) *infoWriter {
+	if fmtTime == nil {
+		fmtTime = func(t time.Time) string { return t.Format(time.Kitchen) }
+	}
+
 	return &infoWriter{
 		writeflusher: w,
 		termWidth:    20,
 		esc:          &escapes{dash: "--", tick: "*"},
-		fmtTime:      func(t time.Time) string { return t.Format(time.Kitchen) },
+		fmtTime:      fmtTime,
 	}
 }
 
@@ -138,9 +150,10 @@ var (
 	MaybePrintSum               = (*infoWriter).maybePrintSum
 	MaybePrintCohortKey         = (*infoWriter).maybePrintCohortKey
 	MaybePrintHealth            = (*infoWriter).maybePrintHealth
-	WaitInhibitUnlock           = waitInhibitUnlock
+	MaybePrintRefreshInfo       = (*infoWriter).maybePrintRefreshInfo
 	WaitWhileInhibited          = waitWhileInhibited
-	IsLocked                    = isLocked
+	NewInhibitionFlow           = newInhibitionFlow
+	ErrSnapRefreshConflict      = errSnapRefreshConflict
 )
 
 func MockPollTime(d time.Duration) (restore func()) {
@@ -344,6 +357,14 @@ func MockConfirmSystemdServiceTracking(fn func(securityTag string) error) (resto
 	}
 }
 
+func MockConfirmSystemdAppTracking(fn func(securityTag string) error) (restore func()) {
+	old := cgroupConfirmSystemdAppTracking
+	cgroupConfirmSystemdAppTracking = fn
+	return func() {
+		cgroupConfirmSystemdAppTracking = old
+	}
+}
+
 func MockApparmorSnapAppFromPid(f func(pid int) (string, string, string, error)) (restore func()) {
 	old := apparmorSnapAppFromPid
 	apparmorSnapAppFromPid = f
@@ -369,10 +390,10 @@ func MockSyscallUmount(f func(string, int) error) (restore func()) {
 }
 
 func MockIoutilTempDir(f func(string, string) (string, error)) (restore func()) {
-	old := ioutilTempDir
-	ioutilTempDir = f
+	old := osMkdirTemp
+	osMkdirTemp = f
 	return func() {
-		ioutilTempDir = old
+		osMkdirTemp = old
 	}
 }
 
@@ -408,45 +429,19 @@ func MockOsChmod(f func(string, os.FileMode) error) (restore func()) {
 	}
 }
 
-func MockWaitInhibitUnlock(f func(snapName string, waitFor runinhibit.Hint) (bool, error)) (restore func()) {
-	old := waitInhibitUnlock
-	waitInhibitUnlock = f
-	return func() {
-		waitInhibitUnlock = old
-	}
+func MockWaitWhileInhibited(f func(ctx context.Context, snapName string, notInhibited func(ctx context.Context) error, inhibited func(ctx context.Context, hint runinhibit.Hint, inhibitInfo *runinhibit.InhibitInfo) (cont bool, err error), interval time.Duration) (flock *osutil.FileLock, retErr error)) (restore func()) {
+	restore = testutil.Backup(&runinhibitWaitWhileInhibited)
+	runinhibitWaitWhileInhibited = f
+	return restore
 }
 
-func MockIsLocked(f func(snapName string) (runinhibit.Hint, error)) (restore func()) {
-	old := isLocked
-	isLocked = f
-	return func() {
-		isLocked = old
-	}
-}
-
-func MockIsGraphicalSession(graphical bool) (restore func()) {
-	old := isGraphicalSession
-	isGraphicalSession = func() bool {
-		return graphical
+func MockInhibitionFlow(flow inhibitionFlow) (restore func()) {
+	old := newInhibitionFlow
+	newInhibitionFlow = func(cli *client.Client, name string) inhibitionFlow {
+		return flow
 	}
 	return func() {
-		isGraphicalSession = old
-	}
-}
-
-func MockPendingRefreshNotification(f func(refreshInfo *usersessionclient.PendingSnapRefreshInfo) error) (restore func()) {
-	old := pendingRefreshNotification
-	pendingRefreshNotification = f
-	return func() {
-		pendingRefreshNotification = old
-	}
-}
-
-func MockFinishRefreshNotification(f func(refreshInfo *usersessionclient.FinishedSnapRefreshInfo) error) (restore func()) {
-	old := finishRefreshNotification
-	finishRefreshNotification = f
-	return func() {
-		finishRefreshNotification = old
+		newInhibitionFlow = old
 	}
 }
 
@@ -470,3 +465,9 @@ func ParseQuotaValues(maxMemory, cpuMax, cpuSet, threadsMax, journalSizeMax, jou
 
 	return quotas.parseQuotas()
 }
+
+func MockSeedWriterReadManifest(f func(manifestFile string) (*seedwriter.Manifest, error)) (restore func()) {
+	restore = testutil.Backup(&seedwriterReadManifest)
+	seedwriterReadManifest = f
+	return restore
+}
