@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2015-2020 Canonical Ltd
+ * Copyright (C) 2015-2023 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -25,7 +25,6 @@ import (
 	"errors"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"mime/multipart"
 	"os"
 	"path/filepath"
@@ -63,6 +62,9 @@ type FileReference struct {
 	TmpPath  string
 }
 
+// RemoveAllExcept removes all temporary files uploaded with form, except for
+// the given paths. Should be called once the files uploaded with the form are
+// no longer needed.
 func (f *Form) RemoveAllExcept(paths []string) {
 	for _, refs := range f.FileRefs {
 		for _, ref := range refs {
@@ -123,7 +125,7 @@ type sideloadFlags struct {
 	dangerousOK bool
 }
 
-func sideloadOrTrySnap(c *Command, body io.ReadCloser, boundary string, user *auth.UserState) Response {
+func sideloadOrTrySnap(ctx context.Context, c *Command, body io.ReadCloser, boundary string, user *auth.UserState) Response {
 	route := c.d.router.Get(stateChangeCmd.Path)
 	if route == nil {
 		return InternalError("cannot find route for change")
@@ -192,9 +194,9 @@ func sideloadOrTrySnap(c *Command, body io.ReadCloser, boundary string, user *au
 
 	var chg *state.Change
 	if len(snapFiles) > 1 {
-		chg, errRsp = sideloadManySnaps(st, snapFiles, sideloadFlags, user)
+		chg, errRsp = sideloadManySnaps(ctx, st, snapFiles, sideloadFlags, user)
 	} else {
-		chg, errRsp = sideloadSnap(st, snapFiles[0], sideloadFlags)
+		chg, errRsp = sideloadSnap(ctx, st, snapFiles[0], sideloadFlags)
 	}
 	if errRsp != nil {
 		return errRsp
@@ -214,16 +216,22 @@ func sideloadOrTrySnap(c *Command, body io.ReadCloser, boundary string, user *au
 	return AsyncResponse(nil, chg.ID())
 }
 
-func sideloadManySnaps(st *state.State, snapFiles []*uploadedSnap, flags sideloadFlags, user *auth.UserState) (*state.Change, *apiError) {
+// sideloadedInfo contains information from a bunch of sideloaded snaps
+type sideloadedInfo struct {
+	sideInfos                  []*snap.SideInfo
+	names, origPaths, tmpPaths []string
+}
+
+func sideloadInfo(st *state.State, snapFiles []*uploadedSnap, flags sideloadFlags) (*sideloadedInfo, *apiError) {
 	deviceCtx, err := snapstate.DevicePastSeeding(st, nil)
 	if err != nil {
 		return nil, InternalError(err.Error())
 	}
 
-	sideInfos := make([]*snap.SideInfo, len(snapFiles))
 	names := make([]string, len(snapFiles))
-	tempPaths := make([]string, len(snapFiles))
 	origPaths := make([]string, len(snapFiles))
+	tmpPaths := make([]string, len(snapFiles))
+	sideInfos := make([]*snap.SideInfo, len(snapFiles))
 
 	for i, snapFile := range snapFiles {
 		si, apiError := readSideInfo(st, snapFile.tmpPath, snapFile.filename, flags, deviceCtx.Model())
@@ -233,8 +241,18 @@ func sideloadManySnaps(st *state.State, snapFiles []*uploadedSnap, flags sideloa
 
 		sideInfos[i] = si
 		names[i] = si.RealName
-		tempPaths[i] = snapFile.tmpPath
 		origPaths[i] = snapFile.filename
+		tmpPaths[i] = snapFile.tmpPath
+	}
+
+	return &sideloadedInfo{sideInfos: sideInfos, names: names,
+		origPaths: origPaths, tmpPaths: tmpPaths}, nil
+}
+
+func sideloadManySnaps(ctx context.Context, st *state.State, snapFiles []*uploadedSnap, flags sideloadFlags, user *auth.UserState) (*state.Change, *apiError) {
+	slInfo, apiErr := sideloadInfo(st, snapFiles, flags)
+	if apiErr != nil {
+		return nil, apiErr
 	}
 
 	var userID int
@@ -242,19 +260,19 @@ func sideloadManySnaps(st *state.State, snapFiles []*uploadedSnap, flags sideloa
 		userID = user.ID
 	}
 
-	tss, err := snapstateInstallPathMany(context.TODO(), st, sideInfos, tempPaths, userID, &flags.Flags)
+	tss, err := snapstateInstallPathMany(ctx, st, slInfo.sideInfos, slInfo.tmpPaths, userID, &flags.Flags)
 	if err != nil {
-		return nil, errToResponse(err, tempPaths, InternalError, "cannot install snap files: %v")
+		return nil, errToResponse(err, slInfo.names, InternalError, "cannot install snap files: %v")
 	}
 
-	msg := fmt.Sprintf(i18n.G("Install snaps %s from files %s"), strutil.Quoted(names), strutil.Quoted(origPaths))
-	chg := newChange(st, "install-snap", msg, tss, names)
-	chg.Set("api-data", map[string][]string{"snap-names": names})
+	msg := fmt.Sprintf(i18n.G("Install snaps %s from files %s"), strutil.Quoted(slInfo.names), strutil.Quoted(slInfo.origPaths))
+	chg := newChange(st, "install-snap", msg, tss, slInfo.names)
+	chg.Set("api-data", map[string][]string{"snap-names": slInfo.names})
 
 	return chg, nil
 }
 
-func sideloadSnap(st *state.State, snapFile *uploadedSnap, flags sideloadFlags) (*state.Change, *apiError) {
+func sideloadSnap(_ context.Context, st *state.State, snapFile *uploadedSnap, flags sideloadFlags) (*state.Change, *apiError) {
 	var instanceName string
 	if snapFile.instanceName != "" {
 		// caller has specified desired instance name
@@ -269,9 +287,31 @@ func sideloadSnap(st *state.State, snapFile *uploadedSnap, flags sideloadFlags)
 		return nil, InternalError(err.Error())
 	}
 
+	// These two are filled only for components
+	var compInfo *snap.ComponentInfo
+	var snapInfo *snap.Info
+
 	sideInfo, apiErr := readSideInfo(st, snapFile.tmpPath, snapFile.filename, flags, deviceCtx.Model())
 	if apiErr != nil {
-		return nil, apiErr
+		// TODO:COMPS: installation of local but asserted components
+		// needs to addressed yet. This will also help with deciding
+		// whether we are dealing with a snap or a component.
+		// Try to load as a component
+		var compErr *apiError
+		compInfo, snapInfo, compErr = readComponentInfo(st, snapFile.tmpPath, instanceName, flags)
+		if compErr != nil {
+			logger.Noticef("cannot sideload as a snap: %v", apiErr)
+			logger.Noticef("cannot sideload as a component: %v", compErr)
+			// If the snap owning the component was not found, we already read
+			// the component information, so this is a valid component and we
+			// report the snap not found error. Otherwise, we don't know and
+			// we report the error while trying to read the file as a snap.
+			if compErr.Kind == client.ErrorKindSnapNotInstalled {
+				return nil, compErr
+			}
+			return nil, apiErr
+		}
+		sideInfo = &snapInfo.SideInfo
 	}
 
 	if instanceName != "" {
@@ -283,14 +323,39 @@ func sideloadSnap(st *state.State, snapFile *uploadedSnap, flags sideloadFlags)
 		instanceName = sideInfo.RealName
 	}
 
-	tset, _, err := snapstateInstallPath(st, sideInfo, snapFile.tmpPath, instanceName, "", flags.Flags)
+	var tset *state.TaskSet
+	contType := "snap"
+	message := fmt.Sprintf("%q snap", instanceName)
+	if compInfo == nil {
+		// TODO pass per request context
+		tset, _, err = snapstateInstallPath(st, sideInfo, snapFile.tmpPath, instanceName, "", flags.Flags, nil)
+	} else {
+		// It is a component
+		contType = "component"
+		message = fmt.Sprintf("%q component for %q snap",
+			compInfo.Component.ComponentName, instanceName)
+		tset, err = snapstateInstallComponentPath(st, snap.NewComponentSideInfo(compInfo.Component, snap.Revision{}), snapInfo, snapFile.tmpPath, flags.Flags)
+	}
 	if err != nil {
-		return nil, errToResponse(err, []string{sideInfo.RealName}, InternalError, "cannot install snap file: %v")
+		return nil, errToResponse(err, []string{sideInfo.RealName}, InternalError, "cannot install %s file: %v", contType)
 	}
 
-	msg := fmt.Sprintf(i18n.G("Install %q snap from file %q"), instanceName, snapFile.filename)
-	chg := newChange(st, "install-snap", msg, []*state.TaskSet{tset}, []string{instanceName})
-	chg.Set("api-data", map[string]string{"snap-name": instanceName})
+	msg := fmt.Sprintf(i18n.G("Install %s from file %q"), message, snapFile.filename)
+	chg := newChange(st, "install-"+contType, msg, []*state.TaskSet{tset}, []string{instanceName})
+	apiData := map[string]interface{}{}
+	if compInfo == nil {
+		apiData = map[string]interface{}{
+			"snap-name":  instanceName,
+			"snap-names": []string{instanceName},
+		}
+	} else {
+		// Installing only a component, so snap name is inside components entry
+		// (snap-name would be included if installing snap+components)
+		apiData["components"] = map[string][]string{
+			instanceName: {compInfo.Component.ComponentName},
+		}
+	}
+	chg.Set("api-data", apiData)
 
 	return chg, nil
 }
@@ -303,7 +368,7 @@ func readSideInfo(st *state.State, tempPath string, origPath string, flags sidel
 		switch {
 		case err == nil:
 			sideInfo = si
-		case asserts.IsNotFound(err):
+		case errors.Is(err, &asserts.NotFoundError{}):
 			// with devmode we try to find assertions but it's ok
 			// if they are not there (implies --dangerous)
 			if !flags.DevMode {
@@ -330,6 +395,66 @@ func readSideInfo(st *state.State, tempPath string, origPath string, flags sidel
 	return sideInfo, nil
 }
 
+var readComponentInfoFromCont = readComponentInfoFromContImpl
+
+func readComponentInfoFromContImpl(tempPath string, csi *snap.ComponentSideInfo) (*snap.ComponentInfo, error) {
+	compf, err := snapfile.Open(tempPath)
+	if err != nil {
+		return nil, fmt.Errorf("cannot open container: %w", err)
+	}
+
+	// hook information isn't loaded here, but it shouldn't be needed in this
+	// context
+	return snap.ReadComponentInfoFromContainer(compf, nil, csi)
+}
+
+// readComponentInfo reads ComponentInfo from a snap component file and the
+// snap.Info of the matching installed snap. If instanceName is not empty, it
+// is used to find the right instance, otherwise the SnapName from the
+// component is used.
+func readComponentInfo(st *state.State, tempPath, instanceName string, flags sideloadFlags) (*snap.ComponentInfo, *snap.Info, *apiError) {
+	if !flags.dangerousOK {
+		// TODO:COMPS: read assertions for components
+		return nil, nil, BadRequest("only unasserted installation of local component with --dangerous is supported at the moment")
+	}
+
+	// TODO:COMPS: will this need to take a non-nil snap.ComponentSideInfo?
+	// not sure where it would get it from, i guess whatever assertion we
+	// end up receiving
+	ci, err := readComponentInfoFromCont(tempPath, nil)
+	if err != nil {
+		return nil, nil, BadRequest("cannot read component metadata: %v", err)
+	}
+
+	// If no instance was provided in the request we use the snap name from the component
+	if instanceName == "" {
+		instanceName = ci.Component.SnapName
+	}
+	si, err := installedSnapInfo(st, instanceName)
+	if err != nil {
+		if errors.Is(err, state.ErrNoState) {
+			return nil, nil, SnapNotInstalled(instanceName, fmt.Errorf("snap owning %q not installed", ci.Component))
+		}
+		return nil, nil, BadRequest("cannot retrieve information for %q: %v", instanceName, err)
+	}
+
+	return ci, si, nil
+}
+
+func installedSnapInfo(st *state.State, instanceName string) (*snap.Info, error) {
+	var snapst snapstate.SnapState
+	if err := snapstate.Get(st, instanceName, &snapst); err != nil {
+		return nil, err
+	}
+
+	snapInfo, err := snapst.CurrentInfo()
+	if err != nil {
+		return nil, err
+	}
+
+	return snapInfo, nil
+}
+
 // maxReadBuflen is the maximum buffer size for reading the non-file parts in the snap upload form
 const maxReadBuflen = 1024 * 1024
 
@@ -420,7 +545,7 @@ func readForm(reader *multipart.Reader) (_ *Form, apiErr *apiError) {
 // its path. If the path is not empty then a file was written and it's the
 // caller's responsibility to clean it up (even if the error is non-nil).
 func writeToTempFile(reader io.Reader) (path string, err error) {
-	tmpf, err := ioutil.TempFile(dirs.SnapBlobDir, dirs.LocalInstallBlobTempPrefix)
+	tmpf, err := os.CreateTemp(dirs.SnapBlobDir, dirs.LocalInstallBlobTempPrefix+"*.snap")
 	if err != nil {
 		return "", fmt.Errorf("cannot create temp file for form data file part: %v", err)
 	}
@@ -466,6 +591,7 @@ func trySnap(st *state.State, trydir string, flags snapstate.Flags) Response {
 		return BadRequest("cannot read snap info for %s: %s", trydir, err)
 	}
 
+	// TODO consider support for trying snaps plus components
 	tset, err := snapstateTryPath(st, info.InstanceName(), trydir, flags)
 	if err != nil {
 		return errToResponse(err, []string{info.InstanceName()}, BadRequest, "cannot try %s: %s", trydir)
@@ -473,7 +599,10 @@ func trySnap(st *state.State, trydir string, flags snapstate.Flags) Response {
 
 	msg := fmt.Sprintf(i18n.G("Try %q snap from %s"), info.InstanceName(), trydir)
 	chg := newChange(st, "try-snap", msg, []*state.TaskSet{tset}, []string{info.InstanceName()})
-	chg.Set("api-data", map[string]string{"snap-name": info.InstanceName()})
+	chg.Set("api-data", map[string]interface{}{
+		"snap-name":  info.InstanceName(),
+		"snap-names": []string{info.InstanceName()},
+	})
 
 	ensureStateSoon(st)
 
