@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2018 Canonical Ltd
+ * Copyright (C) 2018-2024 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -16,9 +16,22 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
  */
+
 package features
 
+import (
+	"github.com/snapcore/snapd/testutil"
+)
+
 // NumberOfFeature returns the number of known features.
 func NumberOfFeatures() int {
 	return int(lastFeature)
 }
+
+var FeaturesSupportedCallbacks = featuresSupportedCallbacks
+
+func MockReleaseSystemctlSupportsUserUnits(f func() bool) (restore func()) {
+	r := testutil.Backup(&releaseSystemctlSupportsUserUnits)
+	releaseSystemctlSupportsUserUnits = f
+	return r
+}
