@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2018 Canonical Ltd
+ * Copyright (C) 2018-2024 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -25,6 +25,9 @@ import (
 
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/osutil"
+	"github.com/snapcore/snapd/release"
+	"github.com/snapcore/snapd/sandbox/apparmor"
+	"github.com/snapcore/snapd/systemd"
 )
 
 // SnapdFeature is a named feature that may be on or off.
@@ -53,6 +56,8 @@ const (
 	DbusActivation
 	// HiddenSnapDataHomeDir controls if the snaps' data dir is ~/.snap/data instead of ~/snap
 	HiddenSnapDataHomeDir
+	// MoveSnapHomeDir controls whether snap user data under ~/snap (or ~/.snap/data) can be moved to ~/Snap.
+	MoveSnapHomeDir
 	// CheckDiskSpaceRemove controls free disk space check on remove whenever automatic snapshot needs to be created.
 	CheckDiskSpaceRemove
 	// CheckDiskSpaceInstall controls free disk space check on snap install.
@@ -61,9 +66,17 @@ const (
 	CheckDiskSpaceRefresh
 	// GateAutoRefreshHook enables refresh control from snaps via gate-auto-refresh hook.
 	GateAutoRefreshHook
-
-	// QuotaGroups enable creating resource quota groups for snaps via the rest API and cli.
+	// QuotaGroups enables any current experimental features related to the Quota Groups API, on top of the features
+	// already graduated past experimental:
+	//  * journal quotas are still experimental
+	// while guota groups creation and management and memory, cpu, quotas are no longer experimental.
 	QuotaGroups
+	// RefreshAppAwarenessUX enables experimental UX improvements for refresh-app-awareness.
+	RefreshAppAwarenessUX
+	// Registries enables experimental configuration based on registries and views.
+	Registries
+	// AppArmorPrompting enables AppArmor to prompt the user for permission when apps perform certain operations.
+	AppArmorPrompting
 
 	// lastFeature is the final known feature, it is only used for testing.
 	lastFeature
@@ -95,6 +108,7 @@ var featureNames = map[SnapdFeature]string{
 	DbusActivation: "dbus-activation",
 
 	HiddenSnapDataHomeDir: "hidden-snap-folder",
+	MoveSnapHomeDir:       "move-snap-home-dir",
 
 	CheckDiskSpaceInstall: "check-disk-space-install",
 	CheckDiskSpaceRefresh: "check-disk-space-refresh",
@@ -103,6 +117,11 @@ var featureNames = map[SnapdFeature]string{
 	GateAutoRefreshHook: "gate-auto-refresh-hook",
 
 	QuotaGroups: "quota-groups",
+
+	RefreshAppAwarenessUX: "refresh-app-awareness-ux",
+	Registries:            "registries",
+
+	AppArmorPrompting: "apparmor-prompting",
 }
 
 // featuresEnabledWhenUnset contains a set of features that are enabled when not explicitly configured.
@@ -123,6 +142,38 @@ var featuresExported = map[SnapdFeature]bool{
 	ClassicPreservesXdgRuntimeDir: true,
 	RobustMountNamespaceUpdates:   true,
 	HiddenSnapDataHomeDir:         true,
+	MoveSnapHomeDir:               true,
+
+	RefreshAppAwarenessUX: true,
+	Registries:            true,
+	AppArmorPrompting:     true,
+}
+
+var (
+	releaseSystemctlSupportsUserUnits = release.SystemctlSupportsUserUnits
+)
+
+// featuresSupportedCallbacks maps features to a callback function which may be
+// run to determine if the feature is supported and, if not, return false along
+// with a reason why the feature is unsupported. If a function has no callback
+// defined, it should be assumed to be supported.
+var featuresSupportedCallbacks = map[SnapdFeature]func() (bool, string){
+	// QuotaGroups requires systemd version 230 or higher
+	QuotaGroups: func() (bool, string) {
+		if err := systemd.EnsureAtLeast(230); err != nil {
+			return false, err.Error()
+		}
+		return true, ""
+	},
+	// UserDaemons requires user units
+	UserDaemons: func() (bool, string) {
+		if !releaseSystemctlSupportsUserUnits() {
+			return false, "user session daemons are not supported on this system's distribution version"
+		}
+		return true, ""
+	},
+	// AppArmorPrompting requires that AppArmor supports prompting.
+	AppArmorPrompting: apparmor.PromptingSupported,
 }
 
 // String returns the name of a snapd feature.
@@ -170,6 +221,19 @@ func (f SnapdFeature) ConfigOption() (snapName, confName string) {
 	return "core", "experimental." + f.String()
 }
 
+// IsSupported returns true if the feature's supported callback returns true, or
+// if it has no supportedCallback. If the feature is unsupported, the returned
+// string details as to why.
+func (f SnapdFeature) IsSupported() (supported bool, whyNot string) {
+	if callback, exists := featuresSupportedCallbacks[f]; exists {
+		supported, whyNot = callback()
+		if !supported {
+			return false, whyNot
+		}
+	}
+	return true, ""
+}
+
 // IsEnabled checks if a given exported snapd feature is enabled.
 //
 // The function panics for features that are not exported.
@@ -204,3 +268,49 @@ func Flag(tr confGetter, feature SnapdFeature) (bool, error) {
 	}
 	return false, fmt.Errorf("%s can only be set to 'true' or 'false', got %q", feature, isEnabled)
 }
+
+// FeatureInfo records whether a particular feature is supported and/or enabled.
+//
+// If the feature is not supported, it should also contain a reason describing
+// why the feature is not supported. A feature is enabled if its feature flag is
+// set to true, regardless of whether or not it is supported.
+type FeatureInfo struct {
+	Supported         bool   `json:"supported"`
+	UnsupportedReason string `json:"unsupported-reason,omitempty"`
+	Enabled           bool   `json:"enabled"`
+}
+
+// All returns a map from feature flags to information about that feature.
+//
+// In particular, the information contains whether the feature is supported
+// and/or enabled. If the feature is not supported, it should also contain a
+// reason describing why the feature is not supported. If a feature's value is
+// not set to true or false, it is excluded from the list, since it is not in
+// this case considered to be a feature flag.
+func All(tr confGetter) map[string]FeatureInfo {
+	knownFeatures := KnownFeatures()
+	allFeaturesInfo := make(map[string]FeatureInfo, len(knownFeatures))
+	for _, feature := range knownFeatures {
+		enabled, err := Flag(tr, feature)
+		if err != nil {
+			// Skip features with values other than true or false
+			continue
+		}
+		// Features implicitly supported if no callback exists
+		supported := true
+		var unsupportedReason string
+		if callback, exists := featuresSupportedCallbacks[feature]; exists {
+			supported, unsupportedReason = callback()
+		}
+		name := feature.String()
+		info := FeatureInfo{
+			Supported: supported,
+			Enabled:   enabled,
+		}
+		if !supported {
+			info.UnsupportedReason = unsupportedReason
+		}
+		allFeaturesInfo[name] = info
+	}
+	return allFeaturesInfo
+}
