@@ -1,6 +1,5 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 //go:build !nosecboot
-// +build !nosecboot
 
 /*
  * Copyright (C) 2019-2020 Canonical Ltd
@@ -25,17 +24,19 @@ import (
 	"fmt"
 	"os"
 	"path/filepath"
-	"syscall"
+	"sort"
 
 	"github.com/snapcore/snapd/asserts"
 	"github.com/snapcore/snapd/boot"
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/gadget"
 	"github.com/snapcore/snapd/gadget/quantity"
+	"github.com/snapcore/snapd/kernel"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/osutil/disks"
 	"github.com/snapcore/snapd/secboot"
 	"github.com/snapcore/snapd/secboot/keys"
+	"github.com/snapcore/snapd/strutil"
 	"github.com/snapcore/snapd/timings"
 )
 
@@ -79,8 +80,9 @@ func roleNeedsEncryption(role string) bool {
 	return role == gadget.SystemData || role == gadget.SystemSave
 }
 
-func saveStorageTraits(mod gadget.Model, allLaidOutVols map[string]*gadget.LaidOutVolume, optsPerVol map[string]*gadget.DiskVolumeValidationOptions, hasSavePartition bool) error {
-	allVolTraits, err := gadget.AllDiskVolumeDeviceTraits(allLaidOutVols, optsPerVol)
+func saveStorageTraits(mod gadget.Model, allVols map[string]*gadget.Volume, optsPerVol map[string]*gadget.DiskVolumeValidationOptions, hasSavePartition bool) error {
+
+	allVolTraits, err := gadget.AllDiskVolumeDeviceTraits(allVols, optsPerVol)
 	if err != nil {
 		return err
 	}
@@ -97,7 +99,9 @@ func saveStorageTraits(mod gadget.Model, allLaidOutVols map[string]*gadget.LaidO
 	return nil
 }
 
-func maybeEncryptPartition(laidOut *gadget.LaidOutStructure, encryptionType secboot.EncryptionType, sectorSize quantity.Size, perfTimings timings.Measurer) (fsParams *mkfsParams, encryptionKey keys.EncryptionKey, err error) {
+func maybeEncryptPartition(dgpair *gadget.OnDiskAndGadgetStructurePair, encryptionType secboot.EncryptionType, sectorSize quantity.Size, perfTimings timings.Measurer) (fsParams *mkfsParams, encryptionKey keys.EncryptionKey, err error) {
+	diskPart := dgpair.DiskStructure
+	volStruct := dgpair.GadgetStructure
 	mustEncrypt := (encryptionType != secboot.EncryptionTypeNone)
 	// fsParams.Device is the kernel device that carries the
 	// filesystem, which is either the raw /dev/<partition>, or
@@ -105,17 +109,17 @@ func maybeEncryptPartition(laidOut *gadget.LaidOutStructure, encryptionType secb
 	// the latter, it will be filled below in this function).
 	fsParams = &mkfsParams{
 		// Filesystem and label are as specified in the gadget
-		Type:  laidOut.Filesystem(),
-		Label: laidOut.Label(),
+		Type:  volStruct.Filesystem,
+		Label: volStruct.Label,
 		// Rest come from disk data
-		Device:     laidOut.Node,
-		Size:       laidOut.Size,
+		Device:     diskPart.Node,
+		Size:       diskPart.Size,
 		SectorSize: sectorSize,
 	}
 
-	if mustEncrypt && roleNeedsEncryption(laidOut.Role()) {
-		timings.Run(perfTimings, fmt.Sprintf("make-key-set[%s]", laidOut.Role()),
-			fmt.Sprintf("Create encryption key set for %s", laidOut.Role()),
+	if mustEncrypt && roleNeedsEncryption(volStruct.Role) {
+		timings.Run(perfTimings, fmt.Sprintf("make-key-set[%s]", volStruct.Role),
+			fmt.Sprintf("Create encryption key set for %s", volStruct.Role),
 			func(timings.Measurer) {
 				encryptionKey, err = keys.NewEncryptionKey()
 				if err != nil {
@@ -125,14 +129,15 @@ func maybeEncryptPartition(laidOut *gadget.LaidOutStructure, encryptionType secb
 		if err != nil {
 			return nil, nil, err
 		}
-		logger.Noticef("encrypting partition device %v", laidOut.Node)
+		logger.Noticef("encrypting partition device %v", diskPart.Node)
 		var dataPart encryptedDevice
 		switch encryptionType {
 		case secboot.EncryptionTypeLUKS, secboot.EncryptionTypeLUKSWithICE:
-			timings.Run(perfTimings, fmt.Sprintf("new-encrypted-device[%s] (%v)", laidOut.Role(), encryptionType),
-				fmt.Sprintf("Create encryption device for %s (%s)", laidOut.Role(), encryptionType),
+			timings.Run(perfTimings, fmt.Sprintf("new-encrypted-device[%s] (%v)", volStruct.Role, encryptionType),
+				fmt.Sprintf("Create encryption device for %s (%s)", volStruct.Role, encryptionType),
 				func(timings.Measurer) {
-					dataPart, err = newEncryptedDeviceLUKS(&laidOut.OnDiskStructure, encryptionType, encryptionKey, laidOut.PartitionFSLabel, laidOut.Name())
+					dataPart, err = newEncryptedDeviceLUKS(diskPart.Node, encryptionType, encryptionKey, volStruct.Label, volStruct.Name)
+					// TODO close device???
 				})
 			if err != nil {
 				return nil, nil, err
@@ -170,12 +175,12 @@ func createFilesystem(part *gadget.OnDiskStructure, fsParams *mkfsParams, partDi
 }
 
 // TODO probably we won't need to pass partDisp when we include storage in laidOut
-func writePartitionContent(laidOut *gadget.LaidOutStructure, fsDevice string, observer gadget.ContentObserver, partDisp string, perfTimings timings.Measurer) error {
+func writePartitionContent(laidOut *gadget.LaidOutStructure, kSnapInfo *KernelSnapInfo, fsDevice string, observer gadget.ContentObserver, partDisp string, perfTimings timings.Measurer) error {
 	var err error
 	timings.Run(perfTimings, fmt.Sprintf("write-content[%s]", partDisp),
 		fmt.Sprintf("Write content for %s", partDisp),
 		func(timings.Measurer) {
-			err = writeFilesystemContent(laidOut, fsDevice, observer)
+			err = writeFilesystemContent(laidOut, kSnapInfo, fsDevice, observer)
 		})
 	if err != nil {
 		return err
@@ -183,22 +188,37 @@ func writePartitionContent(laidOut *gadget.LaidOutStructure, fsDevice string, ob
 	return nil
 }
 
-func installOnePartition(laidOut *gadget.LaidOutStructure, encryptionType secboot.EncryptionType, sectorSize quantity.Size, observer gadget.ContentObserver, perfTimings timings.Measurer) (fsDevice string, encryptionKey keys.EncryptionKey, err error) {
+func installOnePartition(dgpair *gadget.OnDiskAndGadgetStructurePair,
+	kernelInfo *kernel.Info, kernelSnapInfo *KernelSnapInfo, gadgetRoot string,
+	encryptionType secboot.EncryptionType, sectorSize quantity.Size,
+	observer gadget.ContentObserver, perfTimings timings.Measurer,
+) (fsDevice string, encryptionKey keys.EncryptionKey, err error) {
 	// 1. Encrypt
-	role := laidOut.Role()
-	fsParams, encryptionKey, err := maybeEncryptPartition(laidOut, encryptionType, sectorSize, perfTimings)
+	diskPart := dgpair.DiskStructure
+	vs := dgpair.GadgetStructure
+	role := vs.Role
+	fsParams, encryptionKey, err := maybeEncryptPartition(dgpair, encryptionType, sectorSize, perfTimings)
 	if err != nil {
 		return "", nil, fmt.Errorf("cannot encrypt partition %s: %v", role, err)
 	}
 	fsDevice = fsParams.Device
 
 	// 2. Create filesystem
-	if err := createFilesystem(&laidOut.OnDiskStructure, fsParams, role, perfTimings); err != nil {
+	if err := createFilesystem(diskPart, fsParams, role, perfTimings); err != nil {
 		return "", nil, err
 	}
 
 	// 3. Write content
-	if err := writePartitionContent(laidOut, fsDevice, observer, role, perfTimings); err != nil {
+	opts := &gadget.LayoutOptions{
+		GadgetRootDir: gadgetRoot,
+		KernelRootDir: kernelSnapInfo.MountPoint,
+		EncType:       encryptionType,
+	}
+	los, err := gadget.LayoutVolumeStructure(dgpair, kernelInfo, opts)
+	if err != nil {
+		return "", nil, err
+	}
+	if err := writePartitionContent(los, kernelSnapInfo, fsDevice, observer, role, perfTimings); err != nil {
 		return "", nil, err
 	}
 
@@ -229,47 +249,34 @@ func resolveBootDevice(bootDevice string, bootVol *gadget.Volume) (string, error
 // volume name where partitions have been created, the on disk
 // structures after that, the laidout volumes, and the disk sector
 // size.
-func createPartitions(model gadget.Model, gadgetRoot, kernelRoot, bootDevice string, options Options,
-	perfTimings timings.Measurer) (bootVolGadgetName string, created []gadget.LaidOutStructure, allLaidOutVols map[string]*gadget.LaidOutVolume, bootVolSectorSize quantity.Size, err error) {
-	logger.Noticef("installing a new system")
-	logger.Noticef("        gadget data from: %v", gadgetRoot)
-	logger.Noticef("        encryption: %v", options.EncryptionType)
-
-	if gadgetRoot == "" {
-		return "", nil, nil, 0, fmt.Errorf("cannot use empty gadget root directory")
-	}
-
-	if model.Grade() == asserts.ModelGradeUnset {
-		return "", nil, nil, 0,
-			fmt.Errorf("cannot run install mode on pre-UC20 system")
-	}
-
-	// Find boot laid out data
-	laidOutBootVol, allLaidOutVols, err := gadget.LaidOutVolumesFromGadget(gadgetRoot, kernelRoot, model, options.EncryptionType)
+func createPartitions(model gadget.Model, info *gadget.Info, gadgetRoot, kernelRoot, bootDevice string, options Options,
+	perfTimings timings.Measurer) (
+	bootVolGadgetName string, created []*gadget.OnDiskAndGadgetStructurePair, bootVolSectorSize quantity.Size, err error) {
+	// Find boot volume
+	bootVol, err := gadget.FindBootVolume(info.Volumes)
 	if err != nil {
-		return "", nil, nil, 0, fmt.Errorf("cannot layout volumes: %v", err)
+		return "", nil, 0, err
 	}
-	bootVol := laidOutBootVol.Volume
 
 	bootDevice, err = resolveBootDevice(bootDevice, bootVol)
 	if err != nil {
-		return "", nil, nil, 0, err
+		return "", nil, 0, err
 	}
 
 	diskVolume, err := gadget.OnDiskVolumeFromDevice(bootDevice)
 	if err != nil {
-		return "", nil, nil, 0, fmt.Errorf("cannot read %v partitions: %v", bootDevice, err)
+		return "", nil, 0, fmt.Errorf("cannot read %v partitions: %v", bootDevice, err)
 	}
 
 	// check if the current partition table is compatible with the gadget,
 	// ignoring partitions added by the installer (will be removed later)
 	if _, err := gadget.EnsureVolumeCompatibility(bootVol, diskVolume, nil); err != nil {
-		return "", nil, nil, 0, fmt.Errorf("gadget and system-boot device %v partition table not compatible: %v", bootDevice, err)
+		return "", nil, 0, fmt.Errorf("gadget and system-boot device %v partition table not compatible: %v", bootDevice, err)
 	}
 
 	// remove partitions added during a previous install attempt
 	if err := removeCreatedPartitions(gadgetRoot, bootVol, diskVolume); err != nil {
-		return "", nil, nil, 0, fmt.Errorf("cannot remove partitions from previous install: %v", err)
+		return "", nil, 0, fmt.Errorf("cannot remove partitions from previous install: %v", err)
 	}
 	// at this point we removed any existing partition, nuke any
 	// of the existing sealed key files placed outside of the
@@ -277,7 +284,7 @@ func createPartitions(model gadget.Model, gadgetRoot, kernelRoot, bootDevice str
 	sealedKeyFiles, _ := filepath.Glob(filepath.Join(boot.InitramfsSeedEncryptionKeyDir, "*.sealed-key"))
 	for _, keyFile := range sealedKeyFiles {
 		if err := os.Remove(keyFile); err != nil && !os.IsNotExist(err) {
-			return "", nil, nil, 0, fmt.Errorf("cannot cleanup obsolete key file: %v", keyFile)
+			return "", nil, 0, fmt.Errorf("cannot cleanup obsolete key file: %v", keyFile)
 		}
 	}
 
@@ -285,15 +292,15 @@ func createPartitions(model gadget.Model, gadgetRoot, kernelRoot, bootDevice str
 		opts := &CreateOptions{
 			GadgetRootDir: gadgetRoot,
 		}
-		created, err = createMissingPartitions(diskVolume, laidOutBootVol, opts)
+		created, err = createMissingPartitions(diskVolume, bootVol, opts)
 	})
 	if err != nil {
-		return "", nil, nil, 0, fmt.Errorf("cannot create the partitions: %v", err)
+		return "", nil, 0, fmt.Errorf("cannot create the partitions: %v", err)
 	}
 
 	bootVolGadgetName = bootVol.Name
 	bootVolSectorSize = diskVolume.SectorSize
-	return bootVolGadgetName, created, allLaidOutVols, bootVolSectorSize, nil
+	return bootVolGadgetName, created, bootVolSectorSize, nil
 }
 
 func createEncryptionParams(encTyp secboot.EncryptionType) gadget.StructureEncryptionParameters {
@@ -308,12 +315,38 @@ func createEncryptionParams(encTyp secboot.EncryptionType) gadget.StructureEncry
 	return gadget.StructureEncryptionParameters{}
 }
 
+func onDiskStructsSortedIdx(vss map[int]*gadget.OnDiskStructure) []int {
+	yamlIdxSl := []int{}
+	for idx := range vss {
+		yamlIdxSl = append(yamlIdxSl, idx)
+	}
+	sort.Ints(yamlIdxSl)
+	return yamlIdxSl
+}
+
 // Run creates partitions, encrypts them when expected, creates
 // filesystems, and finally writes content on them.
-func Run(model gadget.Model, gadgetRoot, kernelRoot, bootDevice string, options Options, observer gadget.ContentObserver, perfTimings timings.Measurer) (*InstalledSystemSideData, error) {
+func Run(model gadget.Model, gadgetRoot string, kernelSnapInfo *KernelSnapInfo, bootDevice string, options Options, observer gadget.ContentObserver, perfTimings timings.Measurer) (*InstalledSystemSideData, error) {
+	logger.Noticef("installing a new system")
+	logger.Noticef("        gadget data from: %v", gadgetRoot)
+	logger.Noticef("        encryption: %v", options.EncryptionType)
+
+	if gadgetRoot == "" {
+		return nil, fmt.Errorf("cannot use empty gadget root directory")
+	}
+	if model.Grade() == asserts.ModelGradeUnset {
+		return nil, fmt.Errorf("cannot run install mode on pre-UC20 system")
+	}
+
+	info, err := gadget.ReadInfoAndValidate(gadgetRoot, model, nil)
+	if err != nil {
+		return nil, err
+	}
+
 	// Step 1: create partitions
-	bootVolGadgetName, created, allLaidOutVols, bootVolSectorSize, err :=
-		createPartitions(model, gadgetRoot, kernelRoot, bootDevice, options, perfTimings)
+	kernelRoot := kernelSnapInfo.MountPoint
+	bootVolGadgetName, created, bootVolSectorSize, err :=
+		createPartitions(model, info, gadgetRoot, kernelRoot, bootDevice, options, perfTimings)
 	if err != nil {
 		return nil, err
 	}
@@ -321,22 +354,26 @@ func Run(model gadget.Model, gadgetRoot, kernelRoot, bootDevice string, options
 	// Step 2: layout content in the created partitions
 	var keyForRole map[string]keys.EncryptionKey
 	devicesForRoles := map[string]string{}
-
 	partsEncrypted := map[string]gadget.StructureEncryptionParameters{}
-
+	kernelInfo, err := kernel.ReadInfo(kernelRoot)
+	if err != nil {
+		return nil, err
+	}
 	hasSavePartition := false
-
 	// Note that all partitions here will have a role (see
-	// gadget.IsCreatableAtInstall() which defines the list)
-	for _, laidOut := range created {
+	// gadget.IsCreatableAtInstall() which defines the list). We do it in
+	// the order in which partitions were specified in the gadget.
+	for _, dgpair := range created {
+		diskPart := dgpair.DiskStructure
+		vs := dgpair.GadgetStructure
 		logger.Noticef("created new partition %v for structure %v (size %v) with role %s",
-			laidOut.Node, laidOut, laidOut.Size.IECString(), laidOut.Role())
-		if laidOut.Role() == gadget.SystemSave {
+			diskPart.Node, vs, diskPart.Size.IECString(), vs.Role)
+		if vs.Role == gadget.SystemSave {
 			hasSavePartition = true
 		}
 		// keep track of the /dev/<partition> (actual raw
 		// device) for each role
-		devicesForRoles[laidOut.Role()] = laidOut.Node
+		devicesForRoles[vs.Role] = diskPart.Node
 
 		// use the diskLayout.SectorSize here instead of lv.SectorSize, we check
 		// that if there is a sector-size specified in the gadget that it
@@ -346,7 +383,8 @@ func Run(model gadget.Model, gadgetRoot, kernelRoot, bootDevice string, options
 
 		// for encrypted device the filesystem device it will point to
 		// the mapper device otherwise it's the raw device node
-		fsDevice, encryptionKey, err := installOnePartition(&laidOut, options.EncryptionType,
+		fsDevice, encryptionKey, err := installOnePartition(dgpair,
+			kernelInfo, kernelSnapInfo, gadgetRoot, options.EncryptionType,
 			bootVolSectorSize, observer, perfTimings)
 		if err != nil {
 			return nil, err
@@ -356,13 +394,13 @@ func Run(model gadget.Model, gadgetRoot, kernelRoot, bootDevice string, options
 			if keyForRole == nil {
 				keyForRole = map[string]keys.EncryptionKey{}
 			}
-			keyForRole[laidOut.Role()] = encryptionKey
-			partsEncrypted[laidOut.Name()] = createEncryptionParams(options.EncryptionType)
+			keyForRole[vs.Role] = encryptionKey
+			partsEncrypted[vs.Name] = createEncryptionParams(options.EncryptionType)
 		}
-		if options.Mount && laidOut.Label() != "" && laidOut.HasFilesystem() {
+		if options.Mount && vs.Label != "" && vs.HasFilesystem() {
 			// fs is taken from gadget, as on disk one might be displayed as
 			// crypto_LUKS, which is not useful for formatting.
-			if err := mountFilesystem(fsDevice, laidOut.VolumeStructure.Filesystem, getMntPointForPart(laidOut.VolumeStructure)); err != nil {
+			if err := mountFilesystem(fsDevice, vs.LinuxFilesystem(), getMntPointForPart(vs)); err != nil {
 				return nil, err
 			}
 		}
@@ -377,8 +415,9 @@ func Run(model gadget.Model, gadgetRoot, kernelRoot, bootDevice string, options
 			ExpectedStructureEncryption: partsEncrypted,
 		},
 	}
+
 	// save the traits to ubuntu-data host and optionally to ubuntu-save if it exists
-	if err := saveStorageTraits(model, allLaidOutVols, optsPerVol, hasSavePartition); err != nil {
+	if err := saveStorageTraits(model, info.Volumes, optsPerVol, hasSavePartition); err != nil {
 		return nil, err
 	}
 
@@ -418,34 +457,6 @@ func laidOutStructureForDiskStructure(laidVols map[string]*gadget.LaidOutVolume,
 	return nil, fmt.Errorf("cannot find laid out structure for %q", onDiskStruct.Name)
 }
 
-// sysfsPathForBlockDevice returns the sysfs path for a block device.
-var sysfsPathForBlockDevice = func(device string) (string, error) {
-	syfsLink := filepath.Join("/sys/class/block", filepath.Base(device))
-	partPath, err := os.Readlink(syfsLink)
-	if err != nil {
-		return "", fmt.Errorf("cannot read link %q: %v", syfsLink, err)
-	}
-	// Remove initial ../../ from partPath, and make path absolute
-	return filepath.Join("/sys/class/block", partPath), nil
-}
-
-// onDiskVolumeFromPartitionSysfsPath creates an OnDiskVolume that
-// matches the disk that contains the given partition sysfs path
-func onDiskVolumeFromPartitionSysfsPath(partPath string) (*gadget.OnDiskVolume, error) {
-	// Removing the last component will give us the disk path
-	diskPath := filepath.Dir(partPath)
-	disk, err := disks.DiskFromDevicePath(diskPath)
-	if err != nil {
-		return nil, fmt.Errorf("cannot retrieve disk information for %q: %v", partPath, err)
-	}
-	onDiskVol, err := gadget.OnDiskVolumeFromDisk(disk)
-	if err != nil {
-		return nil, fmt.Errorf("cannot retrieve on disk volume for %q: %v", partPath, err)
-	}
-
-	return onDiskVol, nil
-}
-
 // applyOnDiskStructureToLaidOut finds the on disk structure from a
 // partition node and takes the laid out information from laidOutVols
 // and inserts it there.
@@ -487,13 +498,18 @@ func deviceForMaybeEncryptedVolume(volStruct *gadget.VolumeStructure, encSetupDa
 
 // WriteContent writes gadget content to the devices specified in
 // onVolumes. It returns the resolved on disk volumes.
-func WriteContent(onVolumes map[string]*gadget.Volume, allLaidOutVols map[string]*gadget.LaidOutVolume, encSetupData *EncryptionSetupData, observer gadget.ContentObserver, perfTimings timings.Measurer) ([]*gadget.OnDiskVolume, error) {
+func WriteContent(onVolumes map[string]*gadget.Volume, allLaidOutVols map[string]*gadget.LaidOutVolume, encSetupData *EncryptionSetupData, kSnapInfo *KernelSnapInfo, observer gadget.ContentObserver, perfTimings timings.Measurer) ([]*gadget.OnDiskVolume, error) {
 	// TODO this taking onVolumes and allLaidOutVols is odd,
 	// we should try to avoid this when we have partial
 
 	var onDiskVols []*gadget.OnDiskVolume
 	for volName, vol := range onVolumes {
-		var onDiskVol *gadget.OnDiskVolume
+		onDiskVol, err := gadget.OnDiskVolumeFromGadgetVol(vol)
+		if err != nil {
+			return nil, err
+		}
+		onDiskVols = append(onDiskVols, onDiskVol)
+
 		for _, volStruct := range vol.Structure {
 			// TODO write mbr?
 			if volStruct.Role == "mbr" {
@@ -504,21 +520,6 @@ func WriteContent(onVolumes map[string]*gadget.Volume, allLaidOutVols map[string
 				continue
 			}
 
-			// TODO maybe some changes will be needed when we have
-			// encrypted partitions, as the device won't be directly
-			// associated with a disk.
-			partSysfsPath, err := sysfsPathForBlockDevice(volStruct.Device)
-			if err != nil {
-				return nil, err
-			}
-			// Volume needs to be resolved only once inside the loop
-			if onDiskVol == nil {
-				onDiskVol, err = onDiskVolumeFromPartitionSysfsPath(partSysfsPath)
-				if err != nil {
-					return nil, err
-				}
-				onDiskVols = append(onDiskVols, onDiskVol)
-			}
 			logger.Debugf("finding layout for %q", volStruct.Device)
 			// Obtain partition data and link with laid out information
 			// TODO: do we need to consider different
@@ -533,7 +534,7 @@ func WriteContent(onVolumes map[string]*gadget.Volume, allLaidOutVols map[string
 			device := deviceForMaybeEncryptedVolume(&volStruct, encSetupData)
 			logger.Debugf("writing content on partition %s", device)
 			partDisp := roleOrLabelOrName(laidOut.Role(), &laidOut.OnDiskStructure)
-			if err := writePartitionContent(laidOut, device, observer, partDisp, perfTimings); err != nil {
+			if err := writePartitionContent(laidOut, kSnapInfo, device, observer, partDisp, perfTimings); err != nil {
 				return nil, err
 			}
 		}
@@ -568,13 +569,7 @@ func MountVolumes(onVolumes map[string]*gadget.Volume, encSetupData *EncryptionS
 	numSeedPart := 0
 	unmount = func() (err error) {
 		for _, mntPt := range mountPoints {
-			errUnmount := sysUnmount(mntPt, 0)
-			if errUnmount != nil {
-				logger.Noticef("cannot unmount %q: %v (trying lazy unmount next)", mntPt, errUnmount)
-				// lazy umount on error, see LP:2025402
-				errUnmount = sysUnmount(mntPt, syscall.MNT_DETACH)
-				logger.Noticef("cannot lazy unmount %q: %v", mntPt, errUnmount)
-			}
+			errUnmount := unmountWithFallbackToLazy(mntPt, "mounting volumes")
 			// Make sure we do not set err to nil if it had already an error
 			if errUnmount != nil {
 				err = errUnmount
@@ -597,7 +592,7 @@ func MountVolumes(onVolumes map[string]*gadget.Volume, encSetupData *EncryptionS
 			// Device might have been encrypted
 			device := deviceForMaybeEncryptedVolume(&part, encSetupData)
 
-			if err := mountFilesystem(device, part.Filesystem, mntPt); err != nil {
+			if err := mountFilesystem(device, part.LinuxFilesystem(), mntPt); err != nil {
 				defer unmount()
 				return "", nil, fmt.Errorf("cannot mount %q at %q: %v", device, mntPt, err)
 			}
@@ -612,7 +607,7 @@ func MountVolumes(onVolumes map[string]*gadget.Volume, encSetupData *EncryptionS
 	return seedMntDir, unmount, nil
 }
 
-func SaveStorageTraits(model gadget.Model, allLaidOutVols map[string]*gadget.LaidOutVolume, encryptSetupData *EncryptionSetupData) error {
+func SaveStorageTraits(model gadget.Model, vols map[string]*gadget.Volume, encryptSetupData *EncryptionSetupData) error {
 	optsPerVol := map[string]*gadget.DiskVolumeValidationOptions{}
 	if encryptSetupData != nil {
 		for name, p := range encryptSetupData.parts {
@@ -625,7 +620,7 @@ func SaveStorageTraits(model gadget.Model, allLaidOutVols map[string]*gadget.Lai
 	}
 
 	// save the traits to ubuntu-data and ubuntu-save partitions
-	if err := saveStorageTraits(model, allLaidOutVols, optsPerVol, true); err != nil {
+	if err := saveStorageTraits(model, vols, optsPerVol, true); err != nil {
 		return err
 	}
 
@@ -633,20 +628,14 @@ func SaveStorageTraits(model gadget.Model, allLaidOutVols map[string]*gadget.Lai
 }
 
 func EncryptPartitions(onVolumes map[string]*gadget.Volume, encryptionType secboot.EncryptionType, model *asserts.Model, gadgetRoot, kernelRoot string, perfTimings timings.Measurer) (*EncryptionSetupData, error) {
-
-	// TODO for partial gadgets we should use the data from onVolumes instead of
-	// what using only what comes from gadget.yaml.
-	// we might not want to take onVolumes as such as input at that point
-	_, allLaidOutVols, err := gadget.LaidOutVolumesFromGadget(gadgetRoot, kernelRoot, model, encryptionType)
-	if err != nil {
-		return nil, fmt.Errorf("when encrypting partitions: cannot layout volumes: %v", err)
-	}
-
 	setupData := &EncryptionSetupData{
 		parts: make(map[string]partEncryptionData),
 	}
 	for volName, vol := range onVolumes {
-		var onDiskVol *gadget.OnDiskVolume
+		onDiskVol, err := gadget.OnDiskVolumeFromGadgetVol(vol)
+		if err != nil {
+			return nil, err
+		}
 		for _, volStruct := range vol.Structure {
 			// We will only encrypt save or data roles
 			if volStruct.Role != gadget.SystemSave && volStruct.Role != gadget.SystemData {
@@ -657,32 +646,22 @@ func EncryptPartitions(onVolumes map[string]*gadget.Volume, encryptionType secbo
 			}
 			device := volStruct.Device
 
-			partSysfsPath, err := sysfsPathForBlockDevice(device)
+			onDiskStruct, err := structureFromPartDevice(onDiskVol, device)
 			if err != nil {
-				return nil, err
-			}
-			// Volume needs to be resolved only once inside the loop
-			if onDiskVol == nil {
-				onDiskVol, err = onDiskVolumeFromPartitionSysfsPath(partSysfsPath)
-				if err != nil {
-					return nil, err
-				}
-			}
-			// Obtain partition data and link with laid out information
-			const creatingPart = true
-			laidOut, err := applyOnDiskStructureToLaidOut(onDiskVol, device, allLaidOutVols, volName, creatingPart)
-			if err != nil {
-				return nil, fmt.Errorf("cannot retrieve on disk info for %q: %v", device, err)
+				return nil, fmt.Errorf("cannot find partition %q: %v", device, err)
 			}
 
 			logger.Debugf("encrypting partition %s", device)
 
-			fsParams, encryptionKey, err :=
-				maybeEncryptPartition(laidOut, encryptionType, onDiskVol.SectorSize, perfTimings)
+			fsParams, encryptionKey, err := maybeEncryptPartition(
+				&gadget.OnDiskAndGadgetStructurePair{
+					DiskStructure:   onDiskStruct,
+					GadgetStructure: &volStruct},
+				encryptionType, onDiskVol.SectorSize, perfTimings)
 			if err != nil {
 				return nil, fmt.Errorf("cannot encrypt %q: %v", device, err)
 			}
-			setupData.parts[laidOut.Name()] = partEncryptionData{
+			setupData.parts[volStruct.Name] = partEncryptionData{
 				role:   volStruct.Role,
 				device: device,
 				// EncryptedDevice will be /dev/mapper/ubuntu-data, etc.
@@ -705,7 +684,7 @@ func KeysForRole(setupData *EncryptionSetupData) map[string]keys.EncryptionKey {
 	return keyForRole
 }
 
-func FactoryReset(model gadget.Model, gadgetRoot, kernelRoot, bootDevice string, options Options, observer gadget.ContentObserver, perfTimings timings.Measurer) (*InstalledSystemSideData, error) {
+func FactoryReset(model gadget.Model, gadgetRoot string, kernelSnapInfo *KernelSnapInfo, bootDevice string, options Options, observer gadget.ContentObserver, perfTimings timings.Measurer) (*InstalledSystemSideData, error) {
 	logger.Noticef("performing factory reset on an installed system")
 	logger.Noticef("        gadget data from: %v", gadgetRoot)
 	logger.Noticef("        encryption: %v", options.EncryptionType)
@@ -717,13 +696,16 @@ func FactoryReset(model gadget.Model, gadgetRoot, kernelRoot, bootDevice string,
 		return nil, fmt.Errorf("cannot run factory-reset mode on pre-UC20 system")
 	}
 
-	laidOutBootVol, allLaidOutVols, err := gadget.LaidOutVolumesFromGadget(gadgetRoot, kernelRoot, model, options.EncryptionType)
+	// Find boot volume
+	info, err := gadget.ReadInfoAndValidate(gadgetRoot, model, nil)
 	if err != nil {
-		return nil, fmt.Errorf("cannot layout volumes: %v", err)
+		return nil, err
 	}
-	// TODO: resolve content paths from gadget here
-
-	bootDevice, err = resolveBootDevice(bootDevice, laidOutBootVol.Volume)
+	bootVol, err := gadget.FindBootVolume(info.Volumes)
+	if err != nil {
+		return nil, err
+	}
+	bootDevice, err = resolveBootDevice(bootDevice, bootVol)
 	if err != nil {
 		return nil, err
 	}
@@ -733,7 +715,7 @@ func FactoryReset(model gadget.Model, gadgetRoot, kernelRoot, bootDevice string,
 		return nil, fmt.Errorf("cannot read %v partitions: %v", bootDevice, err)
 	}
 
-	layoutCompatOps := &gadget.VolumeCompatibilityOptions{
+	volCompatOps := &gadget.VolumeCompatibilityOptions{
 		AssumeCreatablePartitionsCreated: true,
 		ExpectedStructureEncryption:      map[string]gadget.StructureEncryptionParameters{},
 	}
@@ -746,38 +728,53 @@ func FactoryReset(model gadget.Model, gadgetRoot, kernelRoot, bootDevice string,
 			// XXX what about ICE?
 			return nil, fmt.Errorf("unsupported encryption type %v", options.EncryptionType)
 		}
-		for _, volStruct := range laidOutBootVol.LaidOutStructure {
-			if !roleNeedsEncryption(volStruct.Role()) {
+		for _, volStruct := range bootVol.Structure {
+			if !roleNeedsEncryption(volStruct.Role) {
 				continue
 			}
-			layoutCompatOps.ExpectedStructureEncryption[volStruct.Name()] = encryptionParam
+			volCompatOps.ExpectedStructureEncryption[volStruct.Name] = encryptionParam
 		}
 	}
 	// factory reset is done on a system that was once installed, so this
 	// should be always successful unless the partition table has changed
-	if _, err := gadget.EnsureVolumeCompatibility(laidOutBootVol.Volume, diskLayout, layoutCompatOps); err != nil {
+	yamlIdxToOnDistStruct, err := gadget.EnsureVolumeCompatibility(bootVol, diskLayout, volCompatOps)
+	if err != nil {
 		return nil, fmt.Errorf("gadget and system-boot device %v partition table not compatible: %v", bootDevice, err)
 	}
 
+	kernelInfo, err := kernel.ReadInfo(kernelSnapInfo.MountPoint)
+	if err != nil {
+		return nil, err
+	}
 	var keyForRole map[string]keys.EncryptionKey
 	deviceForRole := map[string]string{}
-
-	savePart := partitionsWithRolesAndContent(laidOutBootVol, diskLayout, []string{gadget.SystemSave})
-	hasSavePartition := len(savePart) != 0
-	if hasSavePartition {
-		deviceForRole[gadget.SystemSave] = savePart[0].Node
-	}
+	var hasSavePartition bool
 	rolesToReset := []string{gadget.SystemBoot, gadget.SystemData}
-	partsToReset := partitionsWithRolesAndContent(laidOutBootVol, diskLayout, rolesToReset)
-	for _, laidOut := range partsToReset {
+	for _, yamlIdx := range onDiskStructsSortedIdx(yamlIdxToOnDistStruct) {
+		onDiskStruct := yamlIdxToOnDistStruct[yamlIdx]
+		vs := bootVol.StructFromYamlIndex(yamlIdx)
+		if vs == nil {
+			continue
+		}
+		if vs.Role == gadget.SystemSave {
+			hasSavePartition = true
+			deviceForRole[gadget.SystemSave] = onDiskStruct.Node
+			continue
+		}
+		if !strutil.ListContains(rolesToReset, vs.Role) {
+			continue
+		}
 		logger.Noticef("resetting %v structure %v (size %v) role %v",
-			laidOut.Node, laidOut, laidOut.Size.IECString(), laidOut.Role())
+			onDiskStruct.Node, vs, onDiskStruct.Size.IECString(), vs.Role)
 
 		// keep track of the /dev/<partition> (actual raw
 		// device) for each role
-		deviceForRole[laidOut.Role()] = laidOut.Node
+		deviceForRole[vs.Role] = onDiskStruct.Node
 
-		fsDevice, encryptionKey, err := installOnePartition(&laidOut, options.EncryptionType,
+		fsDevice, encryptionKey, err := installOnePartition(
+			&gadget.OnDiskAndGadgetStructurePair{
+				DiskStructure: onDiskStruct, GadgetStructure: vs},
+			kernelInfo, kernelSnapInfo, gadgetRoot, options.EncryptionType,
 			diskLayout.SectorSize, observer, perfTimings)
 		if err != nil {
 			return nil, err
@@ -786,12 +783,12 @@ func FactoryReset(model gadget.Model, gadgetRoot, kernelRoot, bootDevice string,
 			if keyForRole == nil {
 				keyForRole = map[string]keys.EncryptionKey{}
 			}
-			keyForRole[laidOut.Role()] = encryptionKey
+			keyForRole[vs.Role] = encryptionKey
 		}
-		if options.Mount && laidOut.Label() != "" && laidOut.HasFilesystem() {
+		if options.Mount && vs.Label != "" && vs.HasFilesystem() {
 			// fs is taken from gadget, as on disk one might be displayed as
 			// crypto_LUKS, which is not useful for formatting.
-			if err := mountFilesystem(fsDevice, laidOut.VolumeStructure.Filesystem, getMntPointForPart(laidOut.VolumeStructure)); err != nil {
+			if err := mountFilesystem(fsDevice, vs.LinuxFilesystem(), getMntPointForPart(vs)); err != nil {
 				return nil, err
 			}
 		}
@@ -802,12 +799,12 @@ func FactoryReset(model gadget.Model, gadgetRoot, kernelRoot, bootDevice string,
 	optsPerVol := map[string]*gadget.DiskVolumeValidationOptions{
 		// this assumes that the encrypted partitions above are always only on the
 		// system-boot volume, this assumption may change
-		laidOutBootVol.Name: {
-			ExpectedStructureEncryption: layoutCompatOps.ExpectedStructureEncryption,
+		bootVol.Name: {
+			ExpectedStructureEncryption: volCompatOps.ExpectedStructureEncryption,
 		},
 	}
 	// save the traits to ubuntu-data host and optionally to ubuntu-save if it exists
-	if err := saveStorageTraits(model, allLaidOutVols, optsPerVol, hasSavePartition); err != nil {
+	if err := saveStorageTraits(model, info.Volumes, optsPerVol, hasSavePartition); err != nil {
 		return nil, err
 	}
 
@@ -816,3 +813,25 @@ func FactoryReset(model gadget.Model, gadgetRoot, kernelRoot, bootDevice string,
 		DeviceForRole: deviceForRole,
 	}, nil
 }
+
+// MatchDisksToGadgetVolumes matches gadget volumes with disks present
+// in the system, taking into account the provided compatibility
+// options. It returns a map of volume names to maps of gadget
+// structure yaml indices to real disk structures.
+func MatchDisksToGadgetVolumes(gVols map[string]*gadget.Volume,
+	volCompatOpts *gadget.VolumeCompatibilityOptions) (map[string]map[int]*gadget.OnDiskStructure, error) {
+	volToGadgetToDiskStruct := map[string]map[int]*gadget.OnDiskStructure{}
+	for name, vol := range gVols {
+		diskVolume, err := gadget.OnDiskVolumeFromGadgetVol(vol)
+		if err != nil {
+			return nil, err
+		}
+		gadgetToDiskMap, err := gadget.EnsureVolumeCompatibility(vol, diskVolume, volCompatOpts)
+		if err != nil {
+			return nil, err
+		}
+		volToGadgetToDiskStruct[name] = gadgetToDiskMap
+	}
+
+	return volToGadgetToDiskStruct, nil
+}
