@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016-2020 Canonical Ltd
+ * Copyright (C) 2016-2024 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -17,13 +17,13 @@
  *
  */
 
-// Package apparmor implements integration between snappy and
-// ubuntu-core-launcher around apparmor.
+// Package apparmor implements integration between snappy and snap-confine
+// around apparmor.
 //
 // Snappy creates apparmor profiles for each application (for each snap)
-// present in the system.  Upon each execution of ubuntu-core-launcher
-// application process is launched under the profile. Prior to that the profile
-// must be parsed, compiled and loaded into the kernel using the support tool
+// present in the system.  Upon each execution of snap-confine application
+// process is launched under the profile. Prior to that the profile must be
+// parsed, compiled and loaded into the kernel using the support tool
 // "apparmor_parser".
 //
 // Each apparmor profile contains a simple <header><content><footer> structure.
@@ -40,7 +40,6 @@ package apparmor
 import (
 	"bytes"
 	"fmt"
-	"io/ioutil"
 	"os"
 	"path"
 	"path/filepath"
@@ -55,6 +54,7 @@ import (
 	"github.com/snapcore/snapd/release"
 	apparmor_sandbox "github.com/snapcore/snapd/sandbox/apparmor"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/snapdenv"
 	"github.com/snapcore/snapd/strutil"
 	"github.com/snapcore/snapd/timings"
 )
@@ -162,10 +162,10 @@ func snapConfineFromSnapProfile(info *snap.Info) (dir, glob string, content map[
 	// We must test the ".real" suffix first, this is a workaround for
 	// https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=858004
 	vanillaProfilePath := filepath.Join(info.MountDir(), "/etc/apparmor.d/usr.lib.snapd.snap-confine.real")
-	vanillaProfileText, err := ioutil.ReadFile(vanillaProfilePath)
+	vanillaProfileText, err := os.ReadFile(vanillaProfilePath)
 	if os.IsNotExist(err) {
 		vanillaProfilePath = filepath.Join(info.MountDir(), "/etc/apparmor.d/usr.lib.snapd.snap-confine")
-		vanillaProfileText, err = ioutil.ReadFile(vanillaProfilePath)
+		vanillaProfileText, err = os.ReadFile(vanillaProfilePath)
 	}
 	if err != nil {
 		return "", "", nil, fmt.Errorf("cannot open apparmor profile for vanilla snap-confine: %s", err)
@@ -264,10 +264,10 @@ func (b *Backend) setupSnapConfineReexec(info *snap.Info) error {
 
 	changed, removed, errEnsure := osutil.EnsureDirState(dir, glob, content)
 	if len(changed) == 0 {
-		// XXX: because NFS workaround is handled separately the same correct
-		// snap-confine profile may need to be re-loaded. This is because the
-		// profile contains include directives and those load a second file
-		// that has changed outside of the scope of EnsureDirState.
+		// XXX: because remote file system workaround is handled separately the
+		// same correct snap-confine profile may need to be re-loaded. This is
+		// because the profile contains include directives and those load a
+		// second file that has changed outside of the scope of EnsureDirState.
 		//
 		// To counter that, always reload the profile by pretending it had
 		// changed.
@@ -310,7 +310,7 @@ func nsProfile(snapName string) string {
 // apps and hooks while the second profile describes the snap-update-ns profile
 // for the whole snap.
 func profileGlobs(snapName string) []string {
-	return []string{interfaces.SecurityTagGlob(snapName), nsProfile(snapName)}
+	return append(interfaces.SecurityTagGlobs(snapName), nsProfile(snapName))
 }
 
 // Determine if a profile filename is removable during core refresh/rollback.
@@ -344,18 +344,20 @@ type profilePathsResults struct {
 	removed   []string
 }
 
-func (b *Backend) prepareProfiles(snapInfo *snap.Info, opts interfaces.ConfinementOptions, repo *interfaces.Repository) (prof *profilePathsResults, err error) {
-	snapName := snapInfo.InstanceName()
-	spec, err := repo.SnapSpecification(b.Name(), snapName)
+func (b *Backend) prepareProfiles(appSet *interfaces.SnapAppSet, opts interfaces.ConfinementOptions, repo *interfaces.Repository) (prof *profilePathsResults, err error) {
+	snapName := appSet.InstanceName()
+	spec, err := repo.SnapSpecification(b.Name(), appSet, opts)
 	if err != nil {
 		return nil, fmt.Errorf("cannot obtain apparmor specification for snap %q: %s", snapName, err)
 	}
 
+	snapInfo := appSet.Info()
+
 	// Add snippets for parallel snap installation mapping
 	spec.(*Specification).AddOvername(snapInfo)
 
 	// Add snippets derived from the layout definition.
-	spec.(*Specification).AddLayout(snapInfo)
+	spec.(*Specification).AddLayout(appSet)
 
 	// Add additional mount layouts rules for the snap.
 	spec.(*Specification).AddExtraLayouts(snapInfo, opts.ExtraLayouts)
@@ -394,13 +396,15 @@ func (b *Backend) prepareProfiles(snapInfo *snap.Info, opts interfaces.Confineme
 	}
 
 	// Get the files that this snap should have
-	content := b.deriveContent(spec.(*Specification), snapInfo, opts)
+	content := b.deriveContent(spec.(*Specification), appSet, opts)
 
 	dir := dirs.SnapAppArmorDir
-	globs := profileGlobs(snapInfo.InstanceName())
 	if err := os.MkdirAll(dir, 0755); err != nil {
 		return nil, fmt.Errorf("cannot create directory for apparmor profiles %q: %s", dir, err)
 	}
+
+	globs := profileGlobs(snapInfo.InstanceName())
+
 	changed, removedPaths, errEnsure := osutil.EnsureDirStateGlobs(dir, globs, content)
 	// XXX: in the old code this error was reported late, after doing load/removeCached.
 	if errEnsure != nil {
@@ -438,12 +442,14 @@ func (b *Backend) prepareProfiles(snapInfo *snap.Info, opts interfaces.Confineme
 //
 // This method should be called after changing plug, slots, connections between
 // them or application present in the snap.
-func (b *Backend) Setup(snapInfo *snap.Info, opts interfaces.ConfinementOptions, repo *interfaces.Repository, tm timings.Measurer) error {
-	prof, err := b.prepareProfiles(snapInfo, opts, repo)
+func (b *Backend) Setup(appSet *interfaces.SnapAppSet, opts interfaces.ConfinementOptions, repo *interfaces.Repository, tm timings.Measurer) error {
+	prof, err := b.prepareProfiles(appSet, opts, repo)
 	if err != nil {
 		return err
 	}
 
+	snapInfo := appSet.Info()
+
 	// Load all changed profiles with a flag that asks apparmor to skip reading
 	// the cache (since we know those changed for sure).  This allows us to
 	// work despite time being wrong (e.g. in the past). For more details see
@@ -485,12 +491,12 @@ func (b *Backend) Setup(snapInfo *snap.Info, opts interfaces.ConfinementOptions,
 // collects and returns them all.
 //
 // This method is useful mainly for regenerating profiles.
-func (b *Backend) SetupMany(snaps []*snap.Info, confinement func(snapName string) interfaces.ConfinementOptions, repo *interfaces.Repository, tm timings.Measurer) []error {
+func (b *Backend) SetupMany(appSets []*interfaces.SnapAppSet, confinement func(snapName string) interfaces.ConfinementOptions, repo *interfaces.Repository, tm timings.Measurer) []error {
 	var allChangedPaths, allUnchangedPaths, allRemovedPaths []string
 	var fallback bool
-	for _, snapInfo := range snaps {
-		opts := confinement(snapInfo.InstanceName())
-		prof, err := b.prepareProfiles(snapInfo, opts, repo)
+	for _, set := range appSets {
+		opts := confinement(set.InstanceName())
+		prof, err := b.prepareProfiles(set, opts, repo)
 		if err != nil {
 			fallback = true
 			break
@@ -506,7 +512,7 @@ func (b *Backend) SetupMany(snaps []*snap.Info, confinement func(snapName string
 			aaFlags |= apparmor_sandbox.SkipKernelLoad
 		}
 		var errReloadChanged error
-		timings.Run(tm, "load-profiles[changed-many]", fmt.Sprintf("load changed security profiles of %d snaps", len(snaps)), func(nesttm timings.Measurer) {
+		timings.Run(tm, "load-profiles[changed-many]", fmt.Sprintf("load changed security profiles of %d snaps", len(appSets)), func(nesttm timings.Measurer) {
 			errReloadChanged = loadProfiles(allChangedPaths, apparmor_sandbox.CacheDir, aaFlags)
 		})
 
@@ -515,7 +521,7 @@ func (b *Backend) SetupMany(snaps []*snap.Info, confinement func(snapName string
 			aaFlags |= apparmor_sandbox.SkipKernelLoad
 		}
 		var errReloadOther error
-		timings.Run(tm, "load-profiles[unchanged-many]", fmt.Sprintf("load unchanged security profiles %d snaps", len(snaps)), func(nesttm timings.Measurer) {
+		timings.Run(tm, "load-profiles[unchanged-many]", fmt.Sprintf("load unchanged security profiles %d snaps", len(appSets)), func(nesttm timings.Measurer) {
 			errReloadOther = loadProfiles(allUnchangedPaths, apparmor_sandbox.CacheDir, aaFlags)
 		})
 
@@ -537,10 +543,11 @@ func (b *Backend) SetupMany(snaps []*snap.Info, confinement func(snapName string
 	var errors []error
 	// if an error was encountered when processing all profiles at once, re-try them one by one
 	if fallback {
-		for _, snapInfo := range snaps {
-			opts := confinement(snapInfo.InstanceName())
-			if err := b.Setup(snapInfo, opts, repo, tm); err != nil {
-				errors = append(errors, fmt.Errorf("cannot setup profiles for snap %q: %s", snapInfo.InstanceName(), err))
+		for _, set := range appSets {
+			instanceName := set.InstanceName()
+			opts := confinement(instanceName)
+			if err := b.Setup(set, opts, repo, tm); err != nil {
+				errors = append(errors, fmt.Errorf("cannot setup profiles for snap %q: %s", instanceName, err))
 			}
 		}
 	}
@@ -608,24 +615,16 @@ var (
 	coreRuntimePattern = regexp.MustCompile("^core([0-9][0-9])?$")
 )
 
-const (
-	attachPattern  = "(attach_disconnected,mediate_deleted)"
-	attachComplain = "(attach_disconnected,mediate_deleted,complain)"
-)
+func (b *Backend) deriveContent(spec *Specification, appSet *interfaces.SnapAppSet, opts interfaces.ConfinementOptions) (content map[string]osutil.FileState) {
+	runnables := appSet.Runnables()
+	content = make(map[string]osutil.FileState, len(runnables))
+	snapInfo := appSet.Info()
 
-func (b *Backend) deriveContent(spec *Specification, snapInfo *snap.Info, opts interfaces.ConfinementOptions) (content map[string]osutil.FileState) {
-	content = make(map[string]osutil.FileState, len(snapInfo.Apps)+len(snapInfo.Hooks)+1)
+	// Add profile for apps and hooks.
+	for _, r := range runnables {
+		b.addContent(r.SecurityTag, snapInfo, r.CommandName, opts, spec.SnippetForTag(r.SecurityTag), content, spec)
+	}
 
-	// Add profile for each app.
-	for _, appInfo := range snapInfo.Apps {
-		securityTag := appInfo.SecurityTag()
-		b.addContent(securityTag, snapInfo, appInfo.Name, opts, spec.SnippetForTag(securityTag), content, spec)
-	}
-	// Add profile for each hook.
-	for _, hookInfo := range snapInfo.Hooks {
-		securityTag := hookInfo.SecurityTag()
-		b.addContent(securityTag, snapInfo, "hook."+hookInfo.Name, opts, spec.SnippetForTag(securityTag), content, spec)
-	}
 	// Add profile for snap-update-ns if we have any apps or hooks.
 	// If we have neither then we don't have any need to create an executing environment.
 	// This applies to, for example, kernel snaps or gadget snaps (unless they have hooks).
@@ -653,6 +652,18 @@ func addUpdateNSProfile(snapInfo *snap.Info, snippets string, content map[string
 				snippets += strings.Replace(apparmor_sandbox.OverlayRootSnippet, "###UPPERDIR###", overlayRoot, -1)
 			}
 			return snippets
+		case "###INCLUDE_SYSTEM_TUNABLES_HOME_D_WITH_VENDORED_APPARMOR###":
+			// XXX: refactor this so that we don't have to duplicate this part.
+			// TODO: rewrite this whole mess with go templates.
+			features, _ := parserFeatures()
+			if strutil.ListContains(features, "snapd-internal") {
+				return `#include if exists "/etc/apparmor.d/tunables/home.d"`
+			}
+			return ""
+		default:
+			if snapdenv.Testing() || osutil.IsTestBinary() {
+				panic(fmt.Sprintf("cannot expand snippet for pattern %q", placeholder))
+			}
 		}
 		return ""
 	})
@@ -665,6 +676,9 @@ func addUpdateNSProfile(snapInfo *snap.Info, snippets string, content map[string
 	}
 }
 
+// Allow optional trailing ' ' after "###PROMPT###"
+var promptReplacer = regexp.MustCompile("###PROMPT### ?")
+
 func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName string, opts interfaces.ConfinementOptions, snippetForTag string, content map[string]osutil.FileState, spec *Specification) {
 	// If base is specified and it doesn't match the core snaps (not
 	// specifying a base should use the default core policy since in this
@@ -684,13 +698,6 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
 		policy = classicTemplate
 		ignoreSnippets = true
 	}
-	// If a snap is in devmode (or is using classic confinement) then make the
-	// profile non-enforcing where violations are logged but not denied.
-	// This is also done for classic so that no confinement applies. Just in
-	// case the profile we start with is not permissive enough.
-	if (opts.DevMode || opts.Classic) && !opts.JailMode {
-		policy = strings.Replace(policy, attachPattern, attachComplain, -1)
-	}
 	policy = templatePattern.ReplaceAllStringFunc(policy, func(placeholder string) string {
 		switch placeholder {
 		case "###DEVMODE_SNAP_CONFINE###":
@@ -783,7 +790,7 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
 				// initial seed change and continue on. This code will be
 				// removed/adapted before it is merged to the main branch,
 				// it is only meant to exist on the security release branch.
-				msg := fmt.Sprintf("neither snapd nor core snap available while preparing apparmor profile for devmode snap %s, panicing to restart snapd to continue seeding", snapInfo.InstanceName())
+				msg := fmt.Sprintf("neither snapd nor core snap available while preparing apparmor profile for devmode snap %s, panicking to restart snapd to continue seeding", snapInfo.InstanceName())
 				panic(msg)
 			}
 
@@ -855,6 +862,44 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
 			return templateVariables(snapInfo, securityTag, cmdName)
 		case "###PROFILEATTACH###":
 			return fmt.Sprintf("profile \"%s\"", securityTag)
+		case "###FLAGS###":
+			// default flags
+			flags := []string{"attach_disconnected", "mediate_deleted"}
+			if spec.Unconfined() == UnconfinedEnabled {
+				// need both parser and kernel support for unconfined
+				pfeatures, _ := parserFeatures()
+				kfeatures, _ := kernelFeatures()
+				if strutil.ListContains(pfeatures, "unconfined") &&
+					strutil.ListContains(kfeatures, "policy:unconfined_restrictions") {
+					flags = append(flags, "unconfined")
+				}
+			}
+			// If a snap is in devmode (or is using classic confinement) then make the
+			// profile non-enforcing where violations are logged but not denied.
+			// This is also done for classic so that no confinement applies. Just in
+			// case the profile we start with is not permissive enough.
+			if (opts.DevMode || opts.Classic) && !opts.JailMode {
+				if !strutil.ListContains(flags, "unconfined") {
+					// Profile modes unconfined and complain
+					// conflict with each other and are
+					// rejected by the parser, in any case
+					// this is fine since we already
+					// requested unconfined based on the
+					// spec and complain would no enforce
+					// any rules anyway.
+					flags = append(flags, "complain")
+				}
+			}
+			if len(flags) > 0 {
+				return "flags=(" + strings.Join(flags, ",") + ")"
+			} else {
+				return ""
+			}
+		case "###PYCACHEDENY###":
+			if spec.SuppressPycacheDeny() {
+				return ""
+			}
+			return pycacheDenySnippet
 		case "###CHANGEPROFILE_RULE###":
 			features, _ := parserFeatures()
 			if strutil.ListContains(features, "unsafe") {
@@ -873,12 +918,13 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
 				// ignoring all apparmor snippets as they may conflict with the
 				// super-broad template we are starting with.
 			} else {
-				// Check if NFS is mounted at or under $HOME. Because NFS is not
-				// transparent to apparmor we must alter the profile to counter that and
-				// allow access to SNAP_USER_* files.
+				// Check if a remote file system is mounted at or under $HOME.
+				// Because some file systems, like NFS, are not transparent to
+				// apparmor we must alter the profile to counter that and allow
+				// access to SNAP_USER_* files.
 				tagSnippets = snippetForTag
-				if nfs, _ := osutil.IsHomeUsingNFS(); nfs {
-					tagSnippets += apparmor_sandbox.NfsSnippet
+				if isRemote, _ := osutil.IsHomeUsingRemoteFS(); isRemote {
+					tagSnippets += apparmor_sandbox.RemoteFSSnippet
 				}
 
 				if overlayRoot, _ := isRootWritableOverlay(); overlayRoot != "" {
@@ -915,6 +961,13 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
 				}
 				tagSnippets = strings.Replace(tagSnippets, "###HOME_IX###", repl, -1)
 
+				// Use prompt prefix if prompting is supported and enabled
+				repl = ""
+				if spec.UsePromptPrefix() {
+					repl = "prompt "
+				}
+				tagSnippets = promptReplacer.ReplaceAllLiteralString(tagSnippets, repl)
+
 				// Conditionally add privilege dropping policy
 				if len(snapInfo.SystemUsernames) > 0 {
 					tagSnippets += privDropAndChownRules
@@ -922,6 +975,12 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
 			}
 
 			return tagSnippets
+		default:
+			if snapdenv.Testing() || osutil.IsTestBinary() {
+				panic(fmt.Sprintf("cannot expand snippet for pattern %q", placeholder))
+			} else {
+				logger.Noticef("WARNING: cannto expand snippet for pattern %q", placeholder)
+			}
 		}
 		return ""
 	})
@@ -933,8 +992,11 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
 }
 
 // NewSpecification returns a new, empty apparmor specification.
-func (b *Backend) NewSpecification() interfaces.Specification {
-	return &Specification{}
+func (b *Backend) NewSpecification(appSet *interfaces.SnapAppSet, opts interfaces.ConfinementOptions) interfaces.Specification {
+	return &Specification{
+		appSet:          appSet,
+		usePromptPrefix: opts.AppArmorPrompting,
+	}
 }
 
 // SandboxFeatures returns the list of apparmor features supported by the kernel.
