@@ -28,6 +28,7 @@ import (
 	"github.com/snapcore/snapd/asserts"
 	"github.com/snapcore/snapd/interfaces"
 	"github.com/snapcore/snapd/interfaces/builtin"
+	"github.com/snapcore/snapd/interfaces/ifacetest"
 	"github.com/snapcore/snapd/interfaces/policy"
 	"github.com/snapcore/snapd/release"
 	"github.com/snapcore/snapd/snap"
@@ -67,11 +68,13 @@ plugs:
   %s:
 `, iface)
 	}
-	slotSnap := snaptest.MockInfo(c, slotYaml, nil)
-	plugSnap := snaptest.MockInfo(c, plugYaml, nil)
+
+	slot, _ := ifacetest.MockConnectedSlot(c, slotYaml, nil, iface)
+	plug, _ := ifacetest.MockConnectedPlug(c, plugYaml, nil, iface)
+
 	return &policy.ConnectCandidate{
-		Plug:            interfaces.NewConnectedPlug(plugSnap.Plugs[iface], nil, nil),
-		Slot:            interfaces.NewConnectedSlot(slotSnap.Slots[iface], nil, nil),
+		Plug:            plug,
+		Slot:            slot,
 		BaseDeclaration: s.baseDecl,
 	}
 }
@@ -1005,42 +1008,43 @@ func (s *baseDeclSuite) TestPlugInstallation(c *C) {
 	all := builtin.Interfaces()
 
 	restricted := map[string]bool{
-		"block-devices":           true,
-		"classic-support":         true,
-		"desktop-launch":          true,
-		"dm-crypt":                true,
-		"docker-support":          true,
-		"greengrass-support":      true,
-		"gpio-control":            true,
-		"ion-memory-control":      true,
-		"kernel-firmware-control": true,
-		"kernel-module-control":   true,
-		"kernel-module-load":      true,
-		"kubernetes-support":      true,
-		"lxd-support":             true,
-		"microceph-support":       true,
-		"microstack-support":      true,
-		"mount-control":           true,
-		"multipass-support":       true,
-		"nvidia-drivers-support":  true,
-		"packagekit-control":      true,
-		"personal-files":          true,
-		"polkit":                  true,
-		"polkit-agent":            true,
-		"remoteproc":              true,
-		"sd-control":              true,
-		"shutdown":                true,
-		"snap-refresh-control":    true,
-		"snap-themes-control":     true,
-		"snap-refresh-observe":    true,
-		"snapd-control":           true,
-		"steam-support":           true,
-		"system-files":            true,
-		"tee":                     true,
-		"uinput":                  true,
-		"unity8":                  true,
-		"userns":                  true,
-		"xilinx-dma":              true,
+		"block-devices":                    true,
+		"classic-support":                  true,
+		"desktop-launch":                   true,
+		"dm-crypt":                         true,
+		"docker-support":                   true,
+		"greengrass-support":               true,
+		"gpio-control":                     true,
+		"ion-memory-control":               true,
+		"kernel-firmware-control":          true,
+		"kernel-module-control":            true,
+		"kernel-module-load":               true,
+		"kubernetes-support":               true,
+		"lxd-support":                      true,
+		"microceph-support":                true,
+		"microstack-support":               true,
+		"mount-control":                    true,
+		"multipass-support":                true,
+		"nvidia-drivers-support":           true,
+		"packagekit-control":               true,
+		"personal-files":                   true,
+		"polkit":                           true,
+		"polkit-agent":                     true,
+		"remoteproc":                       true,
+		"sd-control":                       true,
+		"shutdown":                         true,
+		"snap-interfaces-requests-control": true,
+		"snap-refresh-control":             true,
+		"snap-refresh-observe":             true,
+		"snap-themes-control":              true,
+		"snapd-control":                    true,
+		"steam-support":                    true,
+		"system-files":                     true,
+		"tee":                              true,
+		"uinput":                           true,
+		"unity8":                           true,
+		"userns":                           true,
+		"xilinx-dma":                       true,
 	}
 
 	for _, iface := range all {
@@ -1298,51 +1302,53 @@ func (s *baseDeclSuite) TestValidity(c *C) {
 	// given how the rules work this can be delicate,
 	// listed here to make sure that was a conscious decision
 	bothSides := map[string]bool{
-		"block-devices":           true,
-		"audio-playback":          true,
-		"classic-support":         true,
-		"core-support":            true,
-		"custom-device":           true,
-		"desktop-launch":          true,
-		"dm-crypt":                true,
-		"docker-support":          true,
-		"greengrass-support":      true,
-		"gpio-control":            true,
-		"ion-memory-control":      true,
-		"kernel-firmware-control": true,
-		"kernel-module-control":   true,
-		"kernel-module-load":      true,
-		"kubernetes-support":      true,
-		"lxd-support":             true,
-		"microceph-support":       true,
-		"microstack-support":      true,
-		"mount-control":           true,
-		"multipass-support":       true,
-		"nvidia-drivers-support":  true,
-		"packagekit-control":      true,
-		"personal-files":          true,
-		"pkcs11":                  true,
-		"posix-mq":                true,
-		"polkit":                  true,
-		"polkit-agent":            true,
-		"remoteproc":              true,
-		"qualcomm-ipc-router":     true,
-		"sd-control":              true,
-		"shutdown":                true,
-		"shared-memory":           true,
-		"snap-refresh-control":    true,
-		"snap-themes-control":     true,
-		"snap-refresh-observe":    true,
-		"snapd-control":           true,
-		"steam-support":           true,
-		"system-files":            true,
-		"tee":                     true,
-		"udisks2":                 true,
-		"uinput":                  true,
-		"unity8":                  true,
-		"userns":                  true,
-		"wayland":                 true,
-		"xilinx-dma":              true,
+		"block-devices":                    true,
+		"audio-playback":                   true,
+		"classic-support":                  true,
+		"core-support":                     true,
+		"custom-device":                    true,
+		"desktop-launch":                   true,
+		"dm-crypt":                         true,
+		"docker-support":                   true,
+		"greengrass-support":               true,
+		"gpio-control":                     true,
+		"ion-memory-control":               true,
+		"kernel-firmware-control":          true,
+		"kernel-module-control":            true,
+		"kernel-module-load":               true,
+		"kubernetes-support":               true,
+		"lxd-support":                      true,
+		"microceph-support":                true,
+		"microstack-support":               true,
+		"mount-control":                    true,
+		"multipass-support":                true,
+		"nvidia-drivers-support":           true,
+		"packagekit-control":               true,
+		"personal-files":                   true,
+		"pkcs11":                           true,
+		"posix-mq":                         true,
+		"polkit":                           true,
+		"polkit-agent":                     true,
+		"remoteproc":                       true,
+		"qualcomm-ipc-router":              true,
+		"sd-control":                       true,
+		"shutdown":                         true,
+		"shared-memory":                    true,
+		"snap-interfaces-requests-control": true,
+		"snap-refresh-control":             true,
+		"snap-refresh-observe":             true,
+		"snap-themes-control":              true,
+		"snapd-control":                    true,
+		"steam-support":                    true,
+		"system-files":                     true,
+		"tee":                              true,
+		"udisks2":                          true,
+		"uinput":                           true,
+		"unity8":                           true,
+		"userns":                           true,
+		"wayland":                          true,
+		"xilinx-dma":                       true,
+		"registry":                         true,
 	}
 
 	for _, iface := range all {
@@ -1699,28 +1705,28 @@ plugs:
 }
 
 func (s *baseDeclSuite) TestRawVolumeOverride(c *C) {
-	slotYaml := `name: slot-snap
+	const slotYaml = `name: slot-snap
 type: gadget
 version: 0
 slots:
   raw-volume:
     path: /dev/mmcblk0p1
 `
-	slotSnap := snaptest.MockInfo(c, slotYaml, nil)
+	slot, _ := ifacetest.MockConnectedSlot(c, slotYaml, nil, "raw-volume")
 	// mock a well-formed slot snap decl with SnapID
 	slotSnapDecl := s.mockSnapDecl(c, "slot-snap", "slotsnapidididididididididididid", "canonical", "")
 
-	plugYaml := `name: plug-snap
+	const plugYaml = `name: plug-snap
 version: 0
 plugs:
   raw-volume:
 `
-	plugSnap := snaptest.MockInfo(c, plugYaml, nil)
+	plug, _ := ifacetest.MockConnectedPlug(c, plugYaml, nil, "raw-volume")
 
 	// no plug-side declaration
 	cand := &policy.ConnectCandidate{
-		Plug:                interfaces.NewConnectedPlug(plugSnap.Plugs["raw-volume"], nil, nil),
-		Slot:                interfaces.NewConnectedSlot(slotSnap.Slots["raw-volume"], nil, nil),
+		Plug:                plug,
+		Slot:                slot,
 		SlotSnapDeclaration: slotSnapDecl,
 		BaseDeclaration:     s.baseDecl,
 	}
