@@ -30,10 +30,11 @@ import (
 	"github.com/snapcore/snapd/interfaces/udev"
 	"github.com/snapcore/snapd/release"
 	"github.com/snapcore/snapd/snap"
-	"github.com/snapcore/snapd/snap/snaptest"
+	"github.com/snapcore/snapd/testutil"
 )
 
 type specSuite struct {
+	testutil.BaseTest
 	iface    *ifacetest.TestInterface
 	spec     *udev.Specification
 	plugInfo *snap.PlugInfo
@@ -64,32 +65,40 @@ var _ = Suite(&specSuite{
 	},
 })
 
-func (s *specSuite) SetUpSuite(c *C) {
-	info1 := snaptest.MockInfo(c, `name: snap1
+func (s *specSuite) SetUpTest(c *C) {
+	s.BaseTest.SetUpTest(c)
+
+	dirs.SetRootDir(c.MkDir())
+	s.AddCleanup(func() { dirs.SetRootDir("") })
+
+	const plugYaml = `name: snap1
 version: 0
 plugs:
-    name:
-        interface: test
+  name:
+    interface: test
 apps:
-    foo:
-        command: bin/foo
+  foo:
+    command: bin/foo
+components:
+  comp:
+    type: test
+    hooks:
+      install:
 hooks:
-    configure:
-`, nil)
-	info2 := snaptest.MockInfo(c, `name: snap2
+  configure:
+`
+	const compYaml = "component: snap1+comp\ntype: test"
+	s.plug, s.plugInfo = ifacetest.MockConnectedPlugWithComponents(c, plugYaml, nil, []string{compYaml}, "name")
+
+	const slotYaml = `name: snap2
 version: 0
 slots:
-    name:
-        interface: test
-`, nil)
-	s.plugInfo = info1.Plugs["name"]
-	s.plug = interfaces.NewConnectedPlug(s.plugInfo, nil, nil)
-	s.slotInfo = info2.Slots["name"]
-	s.slot = interfaces.NewConnectedSlot(s.slotInfo, nil, nil)
-}
+  name:
+    interface: test
+`
+	s.slot, s.slotInfo = ifacetest.MockConnectedSlot(c, slotYaml, nil, "name")
 
-func (s *specSuite) SetUpTest(c *C) {
-	s.spec = &udev.Specification{}
+	s.spec = udev.NewSpecification(s.plug.AppSet())
 }
 
 func (s *specSuite) TestAddSnippte(c *C) {
@@ -121,15 +130,20 @@ func (s *specSuite) testTagDevice(c *C, helperDir string) {
 
 	c.Assert(s.spec.Snippets(), DeepEquals, []string{
 		`# iface-1
+kernel="voodoo", TAG+="snap_snap1__comp_hook_install"`,
+		`# iface-2
+kernel="hoodoo", TAG+="snap_snap1__comp_hook_install"`,
+		fmt.Sprintf(`TAG=="snap_snap1__comp_hook_install", SUBSYSTEM!="module", SUBSYSTEM!="subsystem", RUN+="%s/snap-device-helper $env{ACTION} snap_snap1__comp_hook_install $devpath $major:$minor"`, helperDir),
+		`# iface-1
 kernel="voodoo", TAG+="snap_snap1_foo"`,
 		`# iface-2
 kernel="hoodoo", TAG+="snap_snap1_foo"`,
-		fmt.Sprintf(`TAG=="snap_snap1_foo", RUN+="%s/snap-device-helper $env{ACTION} snap_snap1_foo $devpath $major:$minor"`, helperDir),
+		fmt.Sprintf(`TAG=="snap_snap1_foo", SUBSYSTEM!="module", SUBSYSTEM!="subsystem", RUN+="%s/snap-device-helper $env{ACTION} snap_snap1_foo $devpath $major:$minor"`, helperDir),
 		`# iface-1
 kernel="voodoo", TAG+="snap_snap1_hook_configure"`,
 		`# iface-2
 kernel="hoodoo", TAG+="snap_snap1_hook_configure"`,
-		fmt.Sprintf(`TAG=="snap_snap1_hook_configure", RUN+="%[1]s/snap-device-helper $env{ACTION} snap_snap1_hook_configure $devpath $major:$minor"`, helperDir),
+		fmt.Sprintf(`TAG=="snap_snap1_hook_configure", SUBSYSTEM!="module", SUBSYSTEM!="subsystem", RUN+="%[1]s/snap-device-helper $env{ACTION} snap_snap1_hook_configure $devpath $major:$minor"`, helperDir),
 	})
 }
 
@@ -153,12 +167,21 @@ func (s *specSuite) TestTagDeviceAltLibexecdir(c *C) {
 
 // The spec.Specification can be used through the interfaces.Specification interface
 func (s *specSuite) TestSpecificationIface(c *C) {
-	var r interfaces.Specification = s.spec
-	c.Assert(r.AddConnectedPlug(s.iface, s.plug, s.slot), IsNil)
-	c.Assert(r.AddConnectedSlot(s.iface, s.plug, s.slot), IsNil)
+	appSet, err := interfaces.NewSnapAppSet(s.plugInfo.Snap, nil)
+	c.Assert(err, IsNil)
+	spec := udev.NewSpecification(appSet)
+	var r interfaces.Specification = spec
 	c.Assert(r.AddPermanentPlug(s.iface, s.plugInfo), IsNil)
+	c.Assert(r.AddConnectedPlug(s.iface, s.plug, s.slot), IsNil)
+	c.Assert(spec.Snippets(), DeepEquals, []string{"connected-plug", "permanent-plug"})
+
+	appSet, err = interfaces.NewSnapAppSet(s.slotInfo.Snap, nil)
+	c.Assert(err, IsNil)
+	spec = udev.NewSpecification(appSet)
+	r = spec
+	c.Assert(r.AddConnectedSlot(s.iface, s.plug, s.slot), IsNil)
 	c.Assert(r.AddPermanentSlot(s.iface, s.slotInfo), IsNil)
-	c.Assert(s.spec.Snippets(), DeepEquals, []string{"connected-plug", "connected-slot", "permanent-plug", "permanent-slot"})
+	c.Assert(spec.Snippets(), DeepEquals, []string{"connected-slot", "permanent-slot"})
 }
 
 func (s *specSuite) TestControlsDeviceCgroup(c *C) {
