@@ -34,12 +34,15 @@ import (
 
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/overlord"
-	"github.com/snapcore/snapd/overlord/configstate/config"
 	"github.com/snapcore/snapd/overlord/hookstate"
 	"github.com/snapcore/snapd/overlord/hookstate/hooktest"
+	"github.com/snapcore/snapd/overlord/restart"
 	"github.com/snapcore/snapd/overlord/snapstate"
+	"github.com/snapcore/snapd/overlord/snapstate/sequence"
+	"github.com/snapcore/snapd/overlord/snapstate/snapstatetest"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/snap/naming"
 	"github.com/snapcore/snapd/snap/snaptest"
 	"github.com/snapcore/snapd/testutil"
 )
@@ -60,6 +63,10 @@ type baseHookManagerSuite struct {
 	command     *testutil.MockCmd
 }
 
+var (
+	settleTimeout = testutil.HostScaledTimeout(15 * time.Second)
+)
+
 func (s *baseHookManagerSuite) commonSetUpTest(c *C) {
 	s.BaseTest.SetUpTest(c)
 
@@ -70,6 +77,10 @@ func (s *baseHookManagerSuite) commonSetUpTest(c *C) {
 	dirs.SetRootDir(c.MkDir())
 	s.o = overlord.Mock()
 	s.state = s.o.State()
+	s.state.Lock()
+	_, err := restart.Manager(s.state, "boot-id-0", nil)
+	s.state.Unlock()
+	c.Assert(err, IsNil)
 	manager, err := hookstate.Manager(s.state, s.o.TaskRunner())
 	c.Assert(err, IsNil)
 	s.manager = manager
@@ -89,9 +100,6 @@ func (s *baseHookManagerSuite) commonSetUpTest(c *C) {
 		s.context = context
 		return s.mockHandler
 	})
-	s.AddCleanup(hookstate.MockErrtrackerReport(func(string, string, string, map[string]string) (string, error) {
-		return "", nil
-	}))
 }
 
 func (s *baseHookManagerSuite) commonTearDownTest(c *C) {
@@ -127,7 +135,7 @@ func (s *baseHookManagerSuite) setUpSnap(c *C, instanceName string, yaml string)
 	snaptest.MockSnapInstance(c, instanceName, yaml, sideInfo)
 	snapstate.Set(s.state, instanceName, &snapstate.SnapState{
 		Active:      true,
-		Sequence:    []*snap.SideInfo{sideInfo},
+		Sequence:    snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{sideInfo}),
 		Current:     snap.R(1),
 		InstanceKey: instanceKey,
 	})
@@ -174,7 +182,7 @@ func (s *hookManagerSuite) TearDownTest(c *C) {
 }
 
 func (s *hookManagerSuite) settle(c *C) {
-	err := s.o.Settle(5 * time.Second)
+	err := s.o.Settle(settleTimeout)
 	c.Assert(err, IsNil)
 }
 
@@ -187,7 +195,7 @@ func (s *hookManagerSuite) TestHookSetupJsonMarshal(c *C) {
 	hookSetup := &hookstate.HookSetup{Snap: "snap-name", Revision: snap.R(1), Hook: "hook-name"}
 	out, err := json.Marshal(hookSetup)
 	c.Assert(err, IsNil)
-	c.Check(string(out), Equals, "{\"snap\":\"snap-name\",\"revision\":\"1\",\"hook\":\"hook-name\"}")
+	c.Check(string(out), Equals, "{\"snap\":\"snap-name\",\"revision\":\"1\",\"hook\":\"hook-name\",\"component-revision\":\"unset\"}")
 }
 
 func (s *hookManagerSuite) TestHookSetupJsonUnmarshal(c *C) {
@@ -265,7 +273,9 @@ func (s *hookManagerSuite) TestHookTaskEnsure(c *C) {
 
 func (s *hookManagerSuite) TestHookTaskEnsureRestarting(c *C) {
 	// we do no start new hooks runs if we are restarting
-	s.state.RequestRestart(state.RestartDaemon)
+	s.state.Lock()
+	restart.MockPending(s.state, restart.RestartDaemon)
+	s.state.Unlock()
 
 	s.se.Ensure()
 	s.se.Wait()
@@ -436,6 +446,33 @@ func (s *hookManagerSuite) TestHookTaskHandleIgnoreErrorWorks(c *C) {
 	checkTaskLogContains(c, s.task, ".*ignoring failure in hook.*")
 }
 
+func (s *hookManagerSuite) TestHookTaskHandlesHookErrorAndIgnoresIt(c *C) {
+	// tell the mock handler to return 'true' from its Error() handler,
+	// indicating to the hookmgr to ignore the original hook error.
+	s.mockHandler.IgnoreOriginalErr = true
+
+	// Simulate hook error
+	cmd := testutil.MockCommand(
+		c, "snap", ">&2 echo 'hook failed at user request'; exit 1")
+	defer cmd.Restore()
+
+	s.se.Ensure()
+	s.se.Wait()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	c.Check(s.mockHandler.BeforeCalled, Equals, true)
+	c.Check(s.mockHandler.DoneCalled, Equals, false)
+	c.Check(s.mockHandler.ErrorCalled, Equals, true)
+
+	c.Check(s.task.Kind(), Equals, "run-hook")
+	c.Check(s.task.Status(), Equals, state.DoneStatus)
+	c.Check(s.change.Status(), Equals, state.DoneStatus)
+
+	c.Check(s.manager.NumRunningHooks(), Equals, 0)
+}
+
 func (s *hookManagerSuite) TestHookTaskEnforcesTimeout(c *C) {
 	var hooksup hookstate.HookSetup
 
@@ -888,67 +925,6 @@ func (s *hookManagerSuite) TestHookTaskRunsRightSnapCmd(c *C) {
 
 }
 
-func (s *hookManagerSuite) TestHookTaskHandlerReportsErrorIfRequested(c *C) {
-	s.state.Lock()
-	var hooksup hookstate.HookSetup
-	s.task.Get("hook-setup", &hooksup)
-	hooksup.TrackError = true
-	s.task.Set("hook-setup", &hooksup)
-	s.state.Unlock()
-
-	errtrackerCalled := false
-	hookstate.MockErrtrackerReport(func(snap, errmsg, dupSig string, extra map[string]string) (string, error) {
-		c.Check(snap, Equals, "test-snap")
-		c.Check(errmsg, Equals, "hook configure in snap \"test-snap\" failed: hook failed at user request")
-		c.Check(dupSig, Equals, "hook:test-snap:configure:exit status 1\nhook failed at user request\n")
-
-		errtrackerCalled = true
-		return "some-oopsid", nil
-	})
-
-	// Force the snap command to exit 1, and print something to stderr
-	cmd := testutil.MockCommand(
-		c, "snap", ">&2 echo 'hook failed at user request'; exit 1")
-	defer cmd.Restore()
-
-	s.se.Ensure()
-	s.se.Wait()
-
-	s.state.Lock()
-	defer s.state.Unlock()
-
-	c.Check(errtrackerCalled, Equals, true)
-}
-
-func (s *hookManagerSuite) TestHookTaskHandlerReportsErrorDisabled(c *C) {
-	s.state.Lock()
-	var hooksup hookstate.HookSetup
-	s.task.Get("hook-setup", &hooksup)
-	hooksup.TrackError = true
-	s.task.Set("hook-setup", &hooksup)
-
-	tr := config.NewTransaction(s.state)
-	tr.Set("core", "problem-reports.disabled", true)
-	tr.Commit()
-	s.state.Unlock()
-
-	hookstate.MockErrtrackerReport(func(snap, errmsg, dupSig string, extra map[string]string) (string, error) {
-		c.Fatalf("no error reports should be generated")
-		return "", nil
-	})
-
-	// Force the snap command to exit 1, and print something to stderr
-	cmd := testutil.MockCommand(
-		c, "snap", ">&2 echo 'hook failed at user request'; exit 1")
-	defer cmd.Restore()
-
-	s.se.Ensure()
-	s.se.Wait()
-
-	s.state.Lock()
-	defer s.state.Unlock()
-}
-
 func (s *hookManagerSuite) TestHookTasksForSameSnapAreSerialized(c *C) {
 	var Executing int32
 	var TotalExecutions int32
@@ -1016,8 +992,8 @@ func (h *MockConcurrentHandler) Done() error {
 	return nil
 }
 
-func (h *MockConcurrentHandler) Error(err error) error {
-	return nil
+func (h *MockConcurrentHandler) Error(err error) (bool, error) {
+	return false, nil
 }
 
 func NewMockConcurrentHandler(onDone func()) *MockConcurrentHandler {
@@ -1043,7 +1019,7 @@ func (s *hookManagerSuite) TestHookTasksForDifferentSnapsRunConcurrently(c *C) {
 	c.Assert(info.Hooks, HasLen, 1)
 	snapstate.Set(s.state, "test-snap-1", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{sideInfo},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{sideInfo}),
 		Current:  snap.R(1),
 	})
 
@@ -1051,7 +1027,7 @@ func (s *hookManagerSuite) TestHookTasksForDifferentSnapsRunConcurrently(c *C) {
 	snaptest.MockSnap(c, snapYaml2, sideInfo)
 	snapstate.Set(s.state, "test-snap-2", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{sideInfo},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{sideInfo}),
 		Current:  snap.R(1),
 	})
 
@@ -1332,3 +1308,148 @@ func (s *parallelInstancesHookManagerSuite) TestHookTaskEnsureHookRan(c *C) {
 
 	c.Check(s.manager.NumRunningHooks(), Equals, 0)
 }
+
+type componentHookManagerSuite struct {
+	baseHookManagerSuite
+}
+
+var _ = Suite(&componentHookManagerSuite{})
+
+func (s *baseHookManagerSuite) setUpComponent(c *C, instanceName string, componentName string, hookName string) {
+	hooksup := &hookstate.HookSetup{
+		Snap:              instanceName,
+		Hook:              hookName,
+		Revision:          snap.R(1),
+		ComponentRevision: snap.R(1),
+		Component:         componentName,
+	}
+
+	s.state.Lock()
+	defer s.state.Unlock()
+	s.task = hookstate.HookTask(s.state, "test-hook-task", hooksup, nil)
+
+	s.change = s.state.NewChange("run-test-hook", "...")
+	s.change.AddTask(s.task)
+
+	snapName, instanceKey := snap.SplitInstanceName(instanceName)
+
+	sideInfo := &snap.SideInfo{
+		RealName: snapName,
+		SnapID:   "some-snap-id",
+		Revision: snap.R(1),
+	}
+
+	componentSideInfo := &snap.ComponentSideInfo{
+		Component: naming.ComponentRef{
+			SnapName:      snapName,
+			ComponentName: componentName,
+		},
+		Revision: snap.R(1),
+	}
+
+	const componentYaml = `
+component: %s+%s
+type: test
+`
+
+	const snapYaml = `
+name: %s
+version: 1.0
+components:
+  %s:
+    type: test
+    hooks:
+      %s:
+`
+
+	snapInfo := snaptest.MockSnapInstance(c, instanceName, fmt.Sprintf(snapYaml, snapName, componentName, hookName), sideInfo)
+	snaptest.MockComponent(c, fmt.Sprintf(componentYaml, snapName, componentName), snapInfo, snap.ComponentSideInfo{
+		Revision: snap.R(1),
+	})
+
+	snapstate.Set(s.state, instanceName, &snapstate.SnapState{
+		Active: true,
+		Sequence: snapstatetest.NewSequenceFromRevisionSideInfos([]*sequence.RevisionSideState{{
+			Snap: sideInfo,
+			Components: []*sequence.ComponentState{{
+				SideInfo: componentSideInfo,
+				CompType: snap.TestComponent,
+			}},
+		}}),
+		Current:     snap.R(1),
+		InstanceKey: instanceKey,
+	})
+}
+
+func (s *componentHookManagerSuite) SetUpTest(c *C) {
+	s.commonSetUpTest(c)
+	s.mockHandler = hooktest.NewMockHandler()
+}
+
+func (s *componentHookManagerSuite) TestComponentHookTaskEnsure(c *C) {
+	s.setUpComponent(c, "test-snap", "test-component", "install")
+
+	s.se.Ensure()
+	s.se.Wait()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	c.Check(s.command.Calls(), DeepEquals, [][]string{{
+		"snap", "run", "--hook", "install", "-r", "1", "test-snap+test-component",
+	}})
+
+	c.Check(s.task.Kind(), Equals, "run-hook")
+	c.Check(s.task.Status(), Equals, state.DoneStatus)
+	c.Check(s.change.Status(), Equals, state.DoneStatus)
+
+	c.Check(s.manager.NumRunningHooks(), Equals, 0)
+}
+
+func (s *componentHookManagerSuite) TestComponentHookTaskEnsureInstance(c *C) {
+	s.setUpComponent(c, "test-snap_instance", "test-component", "install")
+
+	s.se.Ensure()
+	s.se.Wait()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	c.Check(s.command.Calls(), DeepEquals, [][]string{{
+		"snap", "run", "--hook", "install", "-r", "1", "test-snap_instance+test-component",
+	}})
+
+	fmt.Println(s.change.Err())
+
+	c.Check(s.task.Kind(), Equals, "run-hook")
+	c.Check(s.task.Status(), Equals, state.DoneStatus)
+	c.Check(s.change.Status(), Equals, state.DoneStatus)
+
+	c.Check(s.manager.NumRunningHooks(), Equals, 0)
+}
+
+func (s *componentHookManagerSuite) TestComponentHookWithoutHookIsError(c *C) {
+	s.setUpComponent(c, "test-snap", "test-component", "install")
+
+	s.state.Lock()
+
+	var hooksup hookstate.HookSetup
+	err := s.task.Get("hook-setup", &hooksup)
+	c.Assert(err, IsNil)
+
+	hooksup.Hook = "missing-hook"
+	s.task.Set("hook-setup", &hooksup)
+
+	s.state.Unlock()
+
+	s.se.Ensure()
+	s.se.Wait()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	c.Check(s.task.Kind(), Equals, "run-hook")
+	c.Check(s.task.Status(), Equals, state.ErrorStatus)
+	c.Check(s.change.Status(), Equals, state.ErrorStatus)
+	checkTaskLogContains(c, s.task, `.*component "test-snap\+test-component" has no "missing-hook" hook.*`)
+}
