@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2016-2018 Canonical Ltd
+ * Copyright (C) 2016-2024 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -21,9 +21,13 @@ import (
 	"time"
 
 	"github.com/snapcore/snapd/interfaces"
+	"github.com/snapcore/snapd/overlord/ifacestate/apparmorprompting"
+	"github.com/snapcore/snapd/overlord/ifacestate/schema"
 	"github.com/snapcore/snapd/overlord/ifacestate/udevmonitor"
+	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/testutil"
 	"github.com/snapcore/snapd/timings"
 )
 
@@ -57,11 +61,23 @@ var (
 	AllocHotplugSeq              = allocHotplugSeq
 	AddHotplugSeqWaitTask        = addHotplugSeqWaitTask
 	AddHotplugSlot               = addHotplugSlot
+	HasActiveConnection          = hasActiveConnection
 
 	BatchConnectTasks                = batchConnectTasks
 	FirstTaskAfterBootWhenPreseeding = firstTaskAfterBootWhenPreseeding
 )
 
+func NewInterfaceManagerWithAppArmorPrompting(useAppArmorPrompting bool) *InterfaceManager {
+	m := &InterfaceManager{
+		useAppArmorPrompting: useAppArmorPrompting,
+	}
+	return m
+}
+
+func (m *InterfaceManager) BuildConfinementOptions(st *state.State, snapInfo *snap.Info, flags snapstate.Flags) (interfaces.ConfinementOptions, error) {
+	return m.buildConfinementOptions(st, snapInfo, flags)
+}
+
 type ConnectOpts = connectOpts
 
 func NewConnectOptsWithAutoSet() connectOpts {
@@ -86,6 +102,12 @@ func MockRemoveStaleConnections(f func(st *state.State) error) (restore func())
 	return func() { removeStaleConnections = old }
 }
 
+func MockSnapdAppArmorServiceIsDisabled(f func() bool) (restore func()) {
+	r := testutil.Backup(&snapdAppArmorServiceIsDisabled)
+	snapdAppArmorServiceIsDisabled = f
+	return r
+}
+
 func MockContentLinkRetryTimeout(d time.Duration) (restore func()) {
 	old := contentLinkRetryTimeout
 	contentLinkRetryTimeout = d
@@ -106,6 +128,18 @@ func MockCreateUDevMonitor(new func(udevmonitor.DeviceAddedFunc, udevmonitor.Dev
 	}
 }
 
+func MockCreateInterfacesRequestsManager(new func(s *state.State) (*apparmorprompting.InterfacesRequestsManager, error)) (restore func()) {
+	return testutil.Mock(&createInterfacesRequestsManager, new)
+}
+
+func MockInterfacesRequestsManagerStop(new func(m *apparmorprompting.InterfacesRequestsManager) error) (restore func()) {
+	return testutil.Mock(&interfacesRequestsManagerStop, new)
+}
+
+func MockAssessAppArmorPrompting(new func(m *InterfaceManager) bool) (restore func()) {
+	return testutil.Mock(&assessAppArmorPrompting, new)
+}
+
 func MockUDevInitRetryTimeout(t time.Duration) (restore func()) {
 	old := udevInitRetryTimeout
 	udevInitRetryTimeout = t
@@ -116,19 +150,19 @@ func MockUDevInitRetryTimeout(t time.Duration) (restore func()) {
 
 // UpperCaseConnState returns a canned connection state map.
 // This allows us to keep connState private and still write some tests for it.
-func UpperCaseConnState() map[string]*connState {
-	return map[string]*connState{
+func UpperCaseConnState() map[string]*schema.ConnState {
+	return map[string]*schema.ConnState{
 		"APP:network CORE:network": {Auto: true, Interface: "network"},
 	}
 }
 
-func UpdateConnectionInConnState(conns map[string]*connState, conn *interfaces.Connection, autoConnect, byGadget, undesired, hotplugGone bool) {
+func UpdateConnectionInConnState(conns map[string]*schema.ConnState, conn *interfaces.Connection, autoConnect, byGadget, undesired, hotplugGone bool) {
 	connRef := &interfaces.ConnRef{
 		PlugRef: *conn.Plug.Ref(),
 		SlotRef: *conn.Slot.Ref(),
 	}
 
-	conns[connRef.ID()] = &connState{
+	conns[connRef.ID()] = &schema.ConnState{
 		Interface:        conn.Interface(),
 		StaticPlugAttrs:  conn.Plug.StaticAttrs(),
 		DynamicPlugAttrs: conn.Plug.DynamicAttrs(),
@@ -141,7 +175,7 @@ func UpdateConnectionInConnState(conns map[string]*connState, conn *interfaces.C
 	}
 }
 
-func GetConnStateAttrs(conns map[string]*connState, connID string) (plugStatic, plugDynamic, slotStatic, SlotDynamic map[string]interface{}, ok bool) {
+func GetConnStateAttrs(conns map[string]*schema.ConnState, connID string) (plugStatic, plugDynamic, slotStatic, SlotDynamic map[string]interface{}, ok bool) {
 	conn, ok := conns[connID]
 	if !ok {
 		return nil, nil, nil, nil, false
@@ -155,23 +189,31 @@ func (m *IdentityMapper) SystemSnapName() string {
 }
 
 // MockProfilesNeedRegeneration mocks the function checking if profiles need regeneration.
-func MockProfilesNeedRegeneration(fn func() bool) func() {
-	old := profilesNeedRegeneration
-	profilesNeedRegeneration = fn
-	return func() { profilesNeedRegeneration = old }
+func MockProfilesNeedRegeneration(fn func(m *InterfaceManager) bool) func() {
+	old := profilesNeedRegenerationImpl
+	profilesNeedRegenerationImpl = fn
+	return func() { profilesNeedRegenerationImpl = old }
 }
 
 // MockWriteSystemKey mocks the function responsible for writing the system key.
-func MockWriteSystemKey(fn func() error) func() {
+func MockWriteSystemKey(fn func(extraData interfaces.SystemKeyExtraData) error) func() {
 	old := writeSystemKey
 	writeSystemKey = fn
 	return func() { writeSystemKey = old }
 }
 
+func MockSnapstateFinishRestart(f func(task *state.Task, snapsup *snapstate.SnapSetup) error) (restore func()) {
+	old := snapstateFinishRestart
+	snapstateFinishRestart = f
+	return func() {
+		snapstateFinishRestart = old
+	}
+}
+
 func (m *InterfaceManager) TransitionConnectionsCoreMigration(st *state.State, oldName, newName string) error {
 	return m.transitionConnectionsCoreMigration(st, oldName, newName)
 }
 
-func (m *InterfaceManager) SetupSecurityByBackend(task *state.Task, snaps []*snap.Info, opts []interfaces.ConfinementOptions, tm timings.Measurer) error {
-	return m.setupSecurityByBackend(task, snaps, opts, tm)
+func (m *InterfaceManager) SetupSecurityByBackend(task *state.Task, appSets []*interfaces.SnapAppSet, opts []interfaces.ConfinementOptions, tm timings.Measurer) error {
+	return m.setupSecurityByBackend(task, appSets, opts, tm)
 }
