@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016 Canonical Ltd
+ * Copyright (C) 2016-2024 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -20,7 +20,9 @@
 package ifacestate
 
 import (
+	"errors"
 	"fmt"
+	"path"
 	"reflect"
 	"sort"
 	"strings"
@@ -28,24 +30,70 @@ import (
 
 	"gopkg.in/tomb.v2"
 
+	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/i18n"
 	"github.com/snapcore/snapd/interfaces"
 	"github.com/snapcore/snapd/interfaces/hotplug"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/overlord/hookstate"
+	"github.com/snapcore/snapd/overlord/ifacestate/schema"
+	"github.com/snapcore/snapd/overlord/servicestate"
 	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/snap/quota"
 	"github.com/snapcore/snapd/timings"
 )
 
-// confinementOptions returns interfaces.ConfinementOptions from snapstate.Flags.
-func confinementOptions(flags snapstate.Flags) interfaces.ConfinementOptions {
-	return interfaces.ConfinementOptions{
-		DevMode:  flags.DevMode,
-		JailMode: flags.JailMode,
-		Classic:  flags.Classic,
+var snapstateFinishRestart = snapstate.FinishRestart
+
+// journalQuotaLayout returns the necessary journal quota mount layouts
+// to mimick what systemd does for services with log namespaces.
+func journalQuotaLayout(quotaGroup *quota.Group) []snap.Layout {
+	if quotaGroup.JournalLimit == nil {
+		return nil
 	}
+
+	// bind mount the journal namespace folder on top of the journal folder
+	// /run/systemd/journal.<ns> -> /run/systemd/journal
+	layouts := []snap.Layout{{
+		Bind: path.Join(dirs.SnapSystemdRunDir, fmt.Sprintf("journal.%s", quotaGroup.JournalNamespaceName())),
+		Path: path.Join(dirs.SnapSystemdRunDir, "journal"),
+		Mode: 0755,
+	}}
+	return layouts
+}
+
+// getExtraLayouts helper function to dynamically calculate the extra mount layouts for
+// a snap instance. These are the layouts which can change during the lifetime of a snap
+// like for instance mimicking systemd journal namespace mount layouts.
+func getExtraLayouts(st *state.State, snapInfo *snap.Info) ([]snap.Layout, error) {
+	snapOpts, err := servicestate.SnapServiceOptions(st, snapInfo, nil)
+	if err != nil {
+		return nil, err
+	}
+
+	var extraLayouts []snap.Layout
+	if snapOpts.QuotaGroup != nil {
+		extraLayouts = append(extraLayouts, journalQuotaLayout(snapOpts.QuotaGroup)...)
+	}
+
+	return extraLayouts, nil
+}
+
+func (m *InterfaceManager) buildConfinementOptions(st *state.State, snapInfo *snap.Info, flags snapstate.Flags) (interfaces.ConfinementOptions, error) {
+	extraLayouts, err := getExtraLayouts(st, snapInfo)
+	if err != nil {
+		return interfaces.ConfinementOptions{}, fmt.Errorf("cannot get extra mount layouts of snap %q: %s", snapInfo.InstanceName(), err)
+	}
+
+	return interfaces.ConfinementOptions{
+		DevMode:           flags.DevMode,
+		JailMode:          flags.JailMode,
+		Classic:           flags.Classic,
+		ExtraLayouts:      extraLayouts,
+		AppArmorPrompting: m.useAppArmorPrompting,
+	}, nil
 }
 
 func (m *InterfaceManager) setupAffectedSnaps(task *state.Task, affectingSnap string, affectedSnaps []string, tm timings.Measurer) error {
@@ -69,8 +117,17 @@ func (m *InterfaceManager) setupAffectedSnaps(task *state.Task, affectingSnap st
 		if err := addImplicitSlots(st, affectedSnapInfo); err != nil {
 			return err
 		}
-		opts := confinementOptions(snapst.Flags)
-		if err := m.setupSnapSecurity(task, affectedSnapInfo, opts, tm); err != nil {
+
+		affectedAppSet, err := appSetForSnapRevision(st, affectedSnapInfo)
+		if err != nil {
+			return fmt.Errorf("building app set for snap %q: %v", affectingSnap, err)
+		}
+
+		opts, err := m.buildConfinementOptions(st, affectedSnapInfo, snapst.Flags)
+		if err != nil {
+			return err
+		}
+		if err := m.setupSnapSecurity(task, affectedAppSet, opts, tm); err != nil {
 			return err
 		}
 	}
@@ -104,7 +161,7 @@ func (m *InterfaceManager) doSetupProfiles(task *state.Task, tomb *tomb.Tomb) er
 	// This code is just here to deal with old state that may still
 	// have the 2nd setup-profiles with this flag set.
 	var corePhase2 bool
-	if err := task.Get("core-phase-2", &corePhase2); err != nil && err != state.ErrNoState {
+	if err := task.Get("core-phase-2", &corePhase2); err != nil && !errors.Is(err, state.ErrNoState) {
 		return err
 	}
 	if corePhase2 {
@@ -112,18 +169,69 @@ func (m *InterfaceManager) doSetupProfiles(task *state.Task, tomb *tomb.Tomb) er
 		return nil
 	}
 
-	opts := confinementOptions(snapsup.Flags)
-	return m.setupProfilesForSnap(task, tomb, snapInfo, opts, perfTimings)
-}
-
-func (m *InterfaceManager) setupProfilesForSnap(task *state.Task, _ *tomb.Tomb, snapInfo *snap.Info, opts interfaces.ConfinementOptions, tm timings.Measurer) error {
-	st := task.State()
+	opts, err := m.buildConfinementOptions(task.State(), snapInfo, snapsup.Flags)
+	if err != nil {
+		return err
+	}
 
 	if err := addImplicitSlots(task.State(), snapInfo); err != nil {
 		return err
 	}
 
-	snapName := snapInfo.InstanceName()
+	// this app set is derived from the current task, which will include any
+	// components that are already installed, with the addition of any new
+	// components that are getting setup up by this task
+	appSet, err := appSetForTask(task, snapInfo)
+	if err != nil {
+		return err
+	}
+
+	if err := m.setupProfilesForAppSet(task, appSet, opts, perfTimings); err != nil {
+		return err
+	}
+	return setPendingProfilesSideInfo(task.State(), snapsup.InstanceName(), appSet)
+}
+
+// setupPendingProfilesSideInfo helps updating information about any
+// revision for which security profiles are set up while the snap is
+// not yet active.
+func setPendingProfilesSideInfo(st *state.State, instanceName string, appSet *interfaces.SnapAppSet) error {
+	var snapst snapstate.SnapState
+	if err := snapstate.Get(st, instanceName, &snapst); err != nil && !errors.Is(err, state.ErrNoState) {
+		return err
+	}
+	if !snapst.IsInstalled() {
+		// not yet visible to the rest of the system, nothing to do here
+		return nil
+	}
+	if snapst.Active {
+		// nothing is pending
+		return nil
+	}
+
+	if appSet != nil {
+		csis := make([]*snap.ComponentSideInfo, 0, len(appSet.Components()))
+		for _, ci := range appSet.Components() {
+			csis = append(csis, &ci.ComponentSideInfo)
+		}
+
+		snapst.PendingSecurity = &snapstate.PendingSecurityState{
+			SideInfo:   &appSet.Info().SideInfo,
+			Components: csis,
+		}
+	} else {
+		snapst.PendingSecurity = &snapstate.PendingSecurityState{}
+	}
+
+	snapstate.Set(st, instanceName, &snapst)
+	return nil
+}
+
+func (m *InterfaceManager) setupProfilesForAppSet(task *state.Task, appSet *interfaces.SnapAppSet, opts interfaces.ConfinementOptions, tm timings.Measurer) error {
+	st := task.State()
+
+	snapInfo := appSet.Info()
+	snapName := appSet.InstanceName()
 
 	// The snap may have been updated so perform the following operation to
 	// ensure that we are always working on the correct state:
@@ -144,7 +252,7 @@ func (m *InterfaceManager) setupProfilesForSnap(task *state.Task, _ *tomb.Tomb,
 	if err := m.repo.RemoveSnap(snapName); err != nil {
 		return err
 	}
-	if err := m.repo.AddSnap(snapInfo); err != nil {
+	if err := m.repo.AddAppSet(appSet); err != nil {
 		return err
 	}
 	if len(snapInfo.BadInterfaces) > 0 {
@@ -181,13 +289,15 @@ func (m *InterfaceManager) setupProfilesForSnap(task *state.Task, _ *tomb.Tomb,
 	sort.Strings(affectedNames)
 	affectedNames = append([]string{snapName}, affectedNames...)
 
-	// Obtain snap.Info for each affected snap, skipping those that cannot be
-	// found and compute the confinement options that apply to it.
-	affectedSnaps := make([]*snap.Info, 0, len(affectedSet))
+	// Obtain interfaces.SnapAppSet for each affected snap, skipping those that
+	// cannot be found and compute the confinement options that apply to it.
+	affectedSnapSets := make([]*interfaces.SnapAppSet, 0, len(affectedSet))
 	confinementOpts := make([]interfaces.ConfinementOptions, 0, len(affectedSet))
+
 	// For the snap being setup we know exactly what was requested.
-	affectedSnaps = append(affectedSnaps, snapInfo)
+	affectedSnapSets = append(affectedSnapSets, appSet)
 	confinementOpts = append(confinementOpts, opts)
+
 	// For remaining snaps we need to interrogate the state.
 	for _, name := range affectedNames[1:] {
 		var snapst snapstate.SnapState
@@ -202,11 +312,22 @@ func (m *InterfaceManager) setupProfilesForSnap(task *state.Task, _ *tomb.Tomb,
 		if err := addImplicitSlots(st, snapInfo); err != nil {
 			return err
 		}
-		affectedSnaps = append(affectedSnaps, snapInfo)
-		confinementOpts = append(confinementOpts, confinementOptions(snapst.Flags))
+
+		appSet, err := appSetForSnapRevision(st, snapInfo)
+		if err != nil {
+			return fmt.Errorf("building app set for snap %q: %v", name, err)
+		}
+
+		opts, err := m.buildConfinementOptions(st, snapInfo, snapst.Flags)
+		if err != nil {
+			return err
+		}
+
+		affectedSnapSets = append(affectedSnapSets, appSet)
+		confinementOpts = append(confinementOpts, opts)
 	}
 
-	return m.setupSecurityByBackend(task, affectedSnaps, confinementOpts, tm)
+	return m.setupSecurityByBackend(task, affectedSnapSets, confinementOpts, tm)
 }
 
 func (m *InterfaceManager) doRemoveProfiles(task *state.Task, tomb *tomb.Tomb) error {
@@ -224,7 +345,12 @@ func (m *InterfaceManager) doRemoveProfiles(task *state.Task, tomb *tomb.Tomb) e
 	}
 	snapName := snapSetup.InstanceName()
 
-	return m.removeProfilesForSnap(task, tomb, snapName, perfTimings)
+	if err := m.removeProfilesForSnap(task, tomb, snapName, perfTimings); err != nil {
+		return err
+	}
+
+	// no pending profiles on disk
+	return setPendingProfilesSideInfo(task.State(), snapName, nil)
 }
 
 func (m *InterfaceManager) removeProfilesForSnap(task *state.Task, _ *tomb.Tomb, snapName string, tm timings.Measurer) error {
@@ -263,7 +389,7 @@ func (m *InterfaceManager) undoSetupProfiles(task *state.Task, tomb *tomb.Tomb)
 	defer perfTimings.Save(st)
 
 	var corePhase2 bool
-	if err := task.Get("core-phase-2", &corePhase2); err != nil && err != state.ErrNoState {
+	if err := task.Get("core-phase-2", &corePhase2); err != nil && !errors.Is(err, state.ErrNoState) {
 		return err
 	}
 	if corePhase2 {
@@ -281,7 +407,7 @@ func (m *InterfaceManager) undoSetupProfiles(task *state.Task, tomb *tomb.Tomb)
 	// about the snap, if there is one.
 	var snapst snapstate.SnapState
 	err = snapstate.Get(st, snapName, &snapst)
-	if err != nil && err != state.ErrNoState {
+	if err != nil && !errors.Is(err, state.ErrNoState) {
 		return err
 	}
 	sideInfo := snapst.CurrentSideInfo()
@@ -294,8 +420,27 @@ func (m *InterfaceManager) undoSetupProfiles(task *state.Task, tomb *tomb.Tomb)
 		if err != nil {
 			return err
 		}
-		opts := confinementOptions(snapst.Flags)
-		return m.setupProfilesForSnap(task, tomb, snapInfo, opts, perfTimings)
+		opts, err := m.buildConfinementOptions(task.State(), snapInfo, snapst.Flags)
+		if err != nil {
+			return err
+		}
+
+		if err := addImplicitSlots(st, snapInfo); err != nil {
+			return err
+		}
+
+		// this app set is derived from the currently installed revision of the
+		// snap (not the revision that we are reverting from). it only includes
+		// components that were installed with that revision.
+		appSet, err := appSetForSnapRevision(st, snapInfo)
+		if err != nil {
+			return err
+		}
+
+		if err := m.setupProfilesForAppSet(task, appSet, opts, perfTimings); err != nil {
+			return err
+		}
+		return setPendingProfilesSideInfo(task.State(), snapName, appSet)
 	}
 }
 
@@ -313,18 +458,18 @@ func (m *InterfaceManager) doDiscardConns(task *state.Task, _ *tomb.Tomb) error
 
 	var snapst snapstate.SnapState
 	err = snapstate.Get(st, instanceName, &snapst)
-	if err != nil && err != state.ErrNoState {
+	if err != nil && !errors.Is(err, state.ErrNoState) {
 		return err
 	}
 
-	if err == nil && len(snapst.Sequence) != 0 {
+	if err == nil && len(snapst.Sequence.Revisions) != 0 {
 		return fmt.Errorf("cannot discard connections for snap %q while it is present", instanceName)
 	}
 	conns, err := getConns(st)
 	if err != nil {
 		return err
 	}
-	removed := make(map[string]*connState)
+	removed := make(map[string]*schema.ConnState)
 	for id := range conns {
 		connRef, err := interfaces.ParseConnRef(id)
 		if err != nil {
@@ -345,9 +490,9 @@ func (m *InterfaceManager) undoDiscardConns(task *state.Task, _ *tomb.Tomb) erro
 	st.Lock()
 	defer st.Unlock()
 
-	var removed map[string]*connState
+	var removed map[string]*schema.ConnState
 	err := task.Get("removed", &removed)
-	if err != nil && err != state.ErrNoState {
+	if err != nil && !errors.Is(err, state.ErrNoState) {
 		return err
 	}
 
@@ -365,10 +510,10 @@ func (m *InterfaceManager) undoDiscardConns(task *state.Task, _ *tomb.Tomb) erro
 }
 
 func getDynamicHookAttributes(task *state.Task) (plugAttrs, slotAttrs map[string]interface{}, err error) {
-	if err = task.Get("plug-dynamic", &plugAttrs); err != nil && err != state.ErrNoState {
+	if err = task.Get("plug-dynamic", &plugAttrs); err != nil && !errors.Is(err, state.ErrNoState) {
 		return nil, nil, err
 	}
-	if err = task.Get("slot-dynamic", &slotAttrs); err != nil && err != state.ErrNoState {
+	if err = task.Get("slot-dynamic", &slotAttrs); err != nil && !errors.Is(err, state.ErrNoState) {
 		return nil, nil, err
 	}
 	if plugAttrs == nil {
@@ -400,15 +545,15 @@ func (m *InterfaceManager) doConnect(task *state.Task, _ *tomb.Tomb) (err error)
 	}
 
 	var autoConnect bool
-	if err := task.Get("auto", &autoConnect); err != nil && err != state.ErrNoState {
+	if err := task.Get("auto", &autoConnect); err != nil && !errors.Is(err, state.ErrNoState) {
 		return err
 	}
 	var byGadget bool
-	if err := task.Get("by-gadget", &byGadget); err != nil && err != state.ErrNoState {
+	if err := task.Get("by-gadget", &byGadget); err != nil && !errors.Is(err, state.ErrNoState) {
 		return err
 	}
 	var delayedSetupProfiles bool
-	if err := task.Get("delayed-setup-profiles", &delayedSetupProfiles); err != nil && err != state.ErrNoState {
+	if err := task.Get("delayed-setup-profiles", &delayedSetupProfiles); err != nil && !errors.Is(err, state.ErrNoState) {
 		return err
 	}
 
@@ -426,7 +571,7 @@ func (m *InterfaceManager) doConnect(task *state.Task, _ *tomb.Tomb) (err error)
 
 	var plugSnapst snapstate.SnapState
 	if err := snapstate.Get(st, plugRef.Snap, &plugSnapst); err != nil {
-		if autoConnect && err == state.ErrNoState {
+		if autoConnect && errors.Is(err, state.ErrNoState) {
 			// conflict logic should prevent this
 			return fmt.Errorf("internal error: snap %q is no longer available for auto-connecting", plugRef.Snap)
 		}
@@ -435,7 +580,7 @@ func (m *InterfaceManager) doConnect(task *state.Task, _ *tomb.Tomb) (err error)
 
 	var slotSnapst snapstate.SnapState
 	if err := snapstate.Get(st, slotRef.Snap, &slotSnapst); err != nil {
-		if autoConnect && err == state.ErrNoState {
+		if autoConnect && errors.Is(err, state.ErrNoState) {
 			// conflict logic should prevent this
 			return fmt.Errorf("internal error: snap %q is no longer available for auto-connecting", slotRef.Snap)
 		}
@@ -448,12 +593,22 @@ func (m *InterfaceManager) doConnect(task *state.Task, _ *tomb.Tomb) (err error)
 		return fmt.Errorf("snap %q has no %q plug", connRef.PlugRef.Snap, connRef.PlugRef.Name)
 	}
 
+	plugAppSet, err := appSetForSnapRevision(st, plug.Snap)
+	if err != nil {
+		return fmt.Errorf("building app set for snap %q: %v", plug.Snap.InstanceName(), err)
+	}
+
 	slot := m.repo.Slot(connRef.SlotRef.Snap, connRef.SlotRef.Name)
 	if slot == nil {
 		// conflict logic should prevent this
 		return fmt.Errorf("snap %q has no %q slot", connRef.SlotRef.Snap, connRef.SlotRef.Name)
 	}
 
+	slotAppSet, err := appSetForSnapRevision(st, slot.Snap)
+	if err != nil {
+		return fmt.Errorf("building app set for snap %q: %v", slot.Snap.InstanceName(), err)
+	}
+
 	// attributes are always present, even if there are no hooks (they're initialized by Connect).
 	plugDynamicAttrs, slotDynamicAttrs, err := getDynamicHookAttributes(task)
 	if err != nil {
@@ -466,7 +621,7 @@ func (m *InterfaceManager) doConnect(task *state.Task, _ *tomb.Tomb) (err error)
 	// policy "connection" rules, other auto-connections obey the
 	// "auto-connection" rules
 	if autoConnect && !byGadget {
-		autochecker, err := newAutoConnectChecker(st, task, m.repo, deviceCtx)
+		autochecker, err := newAutoConnectChecker(st, m.repo, deviceCtx)
 		if err != nil {
 			return err
 		}
@@ -496,13 +651,27 @@ func (m *InterfaceManager) doConnect(task *state.Task, _ *tomb.Tomb) (err error)
 	}()
 
 	if !delayedSetupProfiles {
-		slotOpts := confinementOptions(slotSnapst.Flags)
-		if err := m.setupSnapSecurity(task, slot.Snap, slotOpts, perfTimings); err != nil {
+		slotSnapInfo, err := slotSnapst.CurrentInfo()
+		if err != nil {
+			return err
+		}
+		slotOpts, err := m.buildConfinementOptions(st, slotSnapInfo, slotSnapst.Flags)
+		if err != nil {
+			return err
+		}
+		if err := m.setupSnapSecurity(task, slotAppSet, slotOpts, perfTimings); err != nil {
 			return err
 		}
 
-		plugOpts := confinementOptions(plugSnapst.Flags)
-		if err := m.setupSnapSecurity(task, plug.Snap, plugOpts, perfTimings); err != nil {
+		plugSnapInfo, err := plugSnapst.CurrentInfo()
+		if err != nil {
+			return err
+		}
+		plugOpts, err := m.buildConfinementOptions(st, plugSnapInfo, plugSnapst.Flags)
+		if err != nil {
+			return err
+		}
+		if err := m.setupSnapSecurity(task, plugAppSet, plugOpts, perfTimings); err != nil {
 			return err
 		}
 	} else {
@@ -518,7 +687,7 @@ func (m *InterfaceManager) doConnect(task *state.Task, _ *tomb.Tomb) (err error)
 		task.Set("old-conn", old)
 	}
 
-	conns[connRef.ID()] = &connState{
+	conns[connRef.ID()] = &schema.ConnState{
 		Interface:        conn.Interface(),
 		StaticPlugAttrs:  conn.Plug.StaticAttrs(),
 		DynamicPlugAttrs: conn.Plug.DynamicAttrs(),
@@ -558,7 +727,7 @@ func (m *InterfaceManager) doDisconnect(task *state.Task, _ *tomb.Tomb) error {
 
 	// forget flag can be passed with snap disconnect --forget
 	var forget bool
-	if err := task.Get("forget", &forget); err != nil && err != state.ErrNoState {
+	if err := task.Get("forget", &forget); err != nil && !errors.Is(err, state.ErrNoState) {
 		return fmt.Errorf("internal error: cannot read 'forget' flag: %s", err)
 	}
 
@@ -566,7 +735,7 @@ func (m *InterfaceManager) doDisconnect(task *state.Task, _ *tomb.Tomb) error {
 	for _, instanceName := range []string{plugRef.Snap, slotRef.Snap} {
 		var snapst snapstate.SnapState
 		if err := snapstate.Get(st, instanceName, &snapst); err != nil {
-			if err == state.ErrNoState {
+			if errors.Is(err, state.ErrNoState) {
 				task.Logf("skipping disconnect operation for connection %s %s, snap %q doesn't exist", plugRef, slotRef, instanceName)
 				return nil
 			}
@@ -602,8 +771,17 @@ func (m *InterfaceManager) doDisconnect(task *state.Task, _ *tomb.Tomb) error {
 		if err != nil {
 			return err
 		}
-		opts := confinementOptions(snapst.Flags)
-		if err := m.setupSnapSecurity(task, snapInfo, opts, perfTimings); err != nil {
+
+		appSet, err := appSetForSnapRevision(st, snapInfo)
+		if err != nil {
+			return fmt.Errorf("building app set for snap %q: %v", snapInfo.InstanceName(), err)
+		}
+
+		opts, err := m.buildConfinementOptions(st, snapInfo, snapst.Flags)
+		if err != nil {
+			return err
+		}
+		if err := m.setupSnapSecurity(task, appSet, opts, perfTimings); err != nil {
 			return err
 		}
 	}
@@ -611,14 +789,14 @@ func (m *InterfaceManager) doDisconnect(task *state.Task, _ *tomb.Tomb) error {
 	// "auto-disconnect" flag indicates it's a disconnect triggered automatically as part of snap removal;
 	// such disconnects should not set undesired flag and instead just remove the connection.
 	var autoDisconnect bool
-	if err := task.Get("auto-disconnect", &autoDisconnect); err != nil && err != state.ErrNoState {
+	if err := task.Get("auto-disconnect", &autoDisconnect); err != nil && !errors.Is(err, state.ErrNoState) {
 		return fmt.Errorf("internal error: failed to read 'auto-disconnect' flag: %s", err)
 	}
 
 	// "by-hotplug" flag indicates it's a disconnect triggered by hotplug remove event;
 	// we want to keep information of the connection and just mark it as hotplug-gone.
 	var byHotplug bool
-	if err := task.Get("by-hotplug", &byHotplug); err != nil && err != state.ErrNoState {
+	if err := task.Get("by-hotplug", &byHotplug); err != nil && !errors.Is(err, state.ErrNoState) {
 		return fmt.Errorf("internal error: cannot read 'by-hotplug' flag: %s", err)
 	}
 
@@ -651,9 +829,9 @@ func (m *InterfaceManager) undoDisconnect(task *state.Task, _ *tomb.Tomb) error
 	perfTimings := state.TimingsForTask(task)
 	defer perfTimings.Save(st)
 
-	var oldconn connState
+	var oldconn schema.ConnState
 	err := task.Get("old-conn", &oldconn)
-	if err == state.ErrNoState {
+	if errors.Is(err, state.ErrNoState) {
 		return nil
 	}
 	if err != nil {
@@ -661,7 +839,7 @@ func (m *InterfaceManager) undoDisconnect(task *state.Task, _ *tomb.Tomb) error
 	}
 
 	var forget bool
-	if err := task.Get("forget", &forget); err != nil && err != state.ErrNoState {
+	if err := task.Get("forget", &forget); err != nil && !errors.Is(err, state.ErrNoState) {
 		return fmt.Errorf("internal error: cannot read 'forget' flag: %s", err)
 	}
 
@@ -688,6 +866,7 @@ func (m *InterfaceManager) undoDisconnect(task *state.Task, _ *tomb.Tomb) error
 
 	plug := m.repo.Plug(connRef.PlugRef.Snap, connRef.PlugRef.Name)
 	slot := m.repo.Slot(connRef.SlotRef.Snap, connRef.SlotRef.Name)
+
 	if forget && (plug == nil || slot == nil) {
 		// we were trying to forget an inactive connection that was
 		// referring to a non-existing plug or slot; just restore it
@@ -703,17 +882,42 @@ func (m *InterfaceManager) undoDisconnect(task *state.Task, _ *tomb.Tomb) error
 		return fmt.Errorf("snap %q has no %q slot", connRef.SlotRef.Snap, connRef.SlotRef.Name)
 	}
 
+	plugAppSet, err := appSetForSnapRevision(st, plug.Snap)
+	if err != nil {
+		return fmt.Errorf("building app set for snap %q: %v", plug.Snap.InstanceName(), err)
+	}
+
+	slotAppSet, err := appSetForSnapRevision(st, slot.Snap)
+	if err != nil {
+		return fmt.Errorf("building app set for snap %q: %v", slot.Snap.InstanceName(), err)
+	}
+
 	_, err = m.repo.Connect(connRef, nil, oldconn.DynamicPlugAttrs, nil, oldconn.DynamicSlotAttrs, nil)
 	if err != nil {
 		return err
 	}
 
-	slotOpts := confinementOptions(slotSnapst.Flags)
-	if err := m.setupSnapSecurity(task, slot.Snap, slotOpts, perfTimings); err != nil {
+	slotSnapInfo, err := slotSnapst.CurrentInfo()
+	if err != nil {
 		return err
 	}
-	plugOpts := confinementOptions(plugSnapst.Flags)
-	if err := m.setupSnapSecurity(task, plug.Snap, plugOpts, perfTimings); err != nil {
+	slotOpts, err := m.buildConfinementOptions(st, slotSnapInfo, slotSnapst.Flags)
+	if err != nil {
+		return err
+	}
+	if err := m.setupSnapSecurity(task, slotAppSet, slotOpts, perfTimings); err != nil {
+		return err
+	}
+
+	plugSnapInfo, err := plugSnapst.CurrentInfo()
+	if err != nil {
+		return err
+	}
+	plugOpts, err := m.buildConfinementOptions(st, plugSnapInfo, plugSnapst.Flags)
+	if err != nil {
+		return err
+	}
+	if err := m.setupSnapSecurity(task, plugAppSet, plugOpts, perfTimings); err != nil {
 		return err
 	}
 
@@ -741,9 +945,9 @@ func (m *InterfaceManager) undoConnect(task *state.Task, _ *tomb.Tomb) error {
 		return err
 	}
 
-	var old connState
+	var old schema.ConnState
 	err = task.Get("old-conn", &old)
-	if err != nil && err != state.ErrNoState {
+	if err != nil && !errors.Is(err, state.ErrNoState) {
 		return err
 	}
 	if err == nil {
@@ -758,7 +962,7 @@ func (m *InterfaceManager) undoConnect(task *state.Task, _ *tomb.Tomb) error {
 	}
 
 	var delayedSetupProfiles bool
-	if err := task.Get("delayed-setup-profiles", &delayedSetupProfiles); err != nil && err != state.ErrNoState {
+	if err := task.Get("delayed-setup-profiles", &delayedSetupProfiles); err != nil && !errors.Is(err, state.ErrNoState) {
 		return err
 	}
 	if delayedSetupProfiles {
@@ -770,14 +974,25 @@ func (m *InterfaceManager) undoConnect(task *state.Task, _ *tomb.Tomb) error {
 	if plug == nil {
 		return fmt.Errorf("internal error: snap %q has no %q plug", connRef.PlugRef.Snap, connRef.PlugRef.Name)
 	}
+
+	plugAppSet, err := appSetForSnapRevision(st, plug.Snap)
+	if err != nil {
+		return fmt.Errorf("building app set for snap %q: %v", plug.Snap.InstanceName(), err)
+	}
+
 	slot := m.repo.Slot(connRef.SlotRef.Snap, connRef.SlotRef.Name)
 	if slot == nil {
 		return fmt.Errorf("internal error: snap %q has no %q slot", connRef.SlotRef.Snap, connRef.SlotRef.Name)
 	}
 
+	slotAppSet, err := appSetForSnapRevision(st, slot.Snap)
+	if err != nil {
+		return fmt.Errorf("building app set for snap %q: %v", slot.Snap.InstanceName(), err)
+	}
+
 	var plugSnapst snapstate.SnapState
 	err = snapstate.Get(st, plugRef.Snap, &plugSnapst)
-	if err == state.ErrNoState {
+	if errors.Is(err, state.ErrNoState) {
 		return fmt.Errorf("internal error: snap %q is no longer available", plugRef.Snap)
 	}
 	if err != nil {
@@ -785,18 +1000,34 @@ func (m *InterfaceManager) undoConnect(task *state.Task, _ *tomb.Tomb) error {
 	}
 	var slotSnapst snapstate.SnapState
 	err = snapstate.Get(st, slotRef.Snap, &slotSnapst)
-	if err == state.ErrNoState {
+	if errors.Is(err, state.ErrNoState) {
 		return fmt.Errorf("internal error: snap %q is no longer available", slotRef.Snap)
 	}
 	if err != nil {
 		return err
 	}
-	slotOpts := confinementOptions(slotSnapst.Flags)
-	if err := m.setupSnapSecurity(task, slot.Snap, slotOpts, perfTimings); err != nil {
+
+	slotSnapInfo, err := slotSnapst.CurrentInfo()
+	if err != nil {
 		return err
 	}
-	plugOpts := confinementOptions(plugSnapst.Flags)
-	if err := m.setupSnapSecurity(task, plug.Snap, plugOpts, perfTimings); err != nil {
+	slotOpts, err := m.buildConfinementOptions(st, slotSnapInfo, slotSnapst.Flags)
+	if err != nil {
+		return err
+	}
+	if err := m.setupSnapSecurity(task, slotAppSet, slotOpts, perfTimings); err != nil {
+		return err
+	}
+
+	plugSnapInfo, err := plugSnapst.CurrentInfo()
+	if err != nil {
+		return err
+	}
+	plugOpts, err := m.buildConfinementOptions(st, plugSnapInfo, plugSnapst.Flags)
+	if err != nil {
+		return err
+	}
+	if err := m.setupSnapSecurity(task, plugAppSet, plugOpts, perfTimings); err != nil {
 		return err
 	}
 
@@ -815,7 +1046,7 @@ func obsoleteCorePhase2SetupProfiles(kind string, task *state.Task) (bool, error
 	}
 
 	var corePhase2 bool
-	if err := task.Get("core-phase-2", &corePhase2); err != nil && err != state.ErrNoState {
+	if err := task.Get("core-phase-2", &corePhase2); err != nil && !errors.Is(err, state.ErrNoState) {
 		return false, err
 	}
 	return corePhase2, nil
@@ -829,6 +1060,12 @@ func checkAutoconnectConflicts(st *state.State, autoconnectTask *state.Task, plu
 
 		k := task.Kind()
 		if k == "connect" || k == "disconnect" {
+			// if the task depends on the auto-connect in some way, then we assume that it is safe to go ahead
+			// as it was scheduled by that exact task, and we should not block for that.
+			if inSameChangeWaitChain(autoconnectTask, task) {
+				continue
+			}
+
 			// retry if we found another connect/disconnect affecting same snap; note we can only encounter
 			// connects/disconnects created by doAutoDisconnect / doAutoConnect here as manual interface ops
 			// are rejected by conflict check logic in snapstate.
@@ -876,6 +1113,13 @@ func checkAutoconnectConflicts(st *state.State, autoconnectTask *state.Task, plu
 			if k == "discard-snap" && autoconnectTask.Change() != nil && autoconnectTask.Change().ID() == task.Change().ID() {
 				continue
 			}
+
+			// setup-profiles will be scheduled when we schedule the connect hooks during pre-seed, and they are postponed until
+			// after pre-seeding. They will still cause a conflict here, so take that into account.
+			if k == "setup-profiles" && inSameChangeWaitChain(autoconnectTask, task) {
+				continue
+			}
+
 			// if snap is getting removed, we will retry but the snap will be gone and auto-connect becomes no-op
 			// if snap is getting installed/refreshed - temporary conflict, retry later
 			return &state.Retry{After: connectRetryTimeout, Reason: fmt.Sprintf("conflicting snap %s with task %q", otherSnapName, k)}
@@ -1120,13 +1364,13 @@ func (m *InterfaceManager) doAutoConnect(task *state.Task, _ *tomb.Tomb) error {
 	// if this is the case we can only proceed once the restart
 	// has happened or we may not have all the interfaces of the
 	// new core/base snap.
-	if err := snapstate.FinishRestart(task, snapsup); err != nil {
+	if err := snapstateFinishRestart(task, snapsup); err != nil {
 		return err
 	}
 
 	snapName := snapsup.InstanceName()
 
-	autochecker, err := newAutoConnectChecker(st, task, m.repo, deviceCtx)
+	autochecker, err := newAutoConnectChecker(st, m.repo, deviceCtx)
 	if err != nil {
 		return err
 	}
@@ -1144,7 +1388,7 @@ func (m *InterfaceManager) doAutoConnect(task *state.Task, _ *tomb.Tomb) error {
 	// snaps with not ready link-snap|setup-profiles tasks
 	defaultProviders := snap.DefaultContentProviders(m.repo.Plugs(snapName))
 	for _, chg := range st.Changes() {
-		if chg.Status().Ready() {
+		if chg.IsReady() {
 			continue
 		}
 		for _, t := range chg.Tasks() {
@@ -1197,7 +1441,7 @@ func (m *InterfaceManager) doAutoConnect(task *state.Task, _ *tomb.Tomb) error {
 	cannotAutoConnectLog := func(plug *snap.PlugInfo, candRefs []string) string {
 		return fmt.Sprintf("cannot auto-connect plug %s, candidates found: %s", plug, strings.Join(candRefs, ", "))
 	}
-	if err := autochecker.addAutoConnections(newconns, plugs, nil, conns, cannotAutoConnectLog, conflictError); err != nil {
+	if err := autochecker.addAutoConnections(task, newconns, plugs, nil, conns, cannotAutoConnectLog, conflictError); err != nil {
 		return err
 	}
 	// Auto-connect all the slots
@@ -1210,7 +1454,7 @@ func (m *InterfaceManager) doAutoConnect(task *state.Task, _ *tomb.Tomb) error {
 		cannotAutoConnectLog := func(plug *snap.PlugInfo, candRefs []string) string {
 			return fmt.Sprintf("cannot auto-connect slot %s to plug %s, candidates found: %s", slot, plug, strings.Join(candRefs, ", "))
 		}
-		if err := autochecker.addAutoConnections(newconns, candidates, filterForSlot(slot), conns, cannotAutoConnectLog, conflictError); err != nil {
+		if err := autochecker.addAutoConnections(task, newconns, candidates, filterForSlot(slot), conns, cannotAutoConnectLog, conflictError); err != nil {
 			return err
 		}
 	}
@@ -1447,7 +1691,7 @@ func (m *InterfaceManager) doHotplugConnect(task *state.Task, _ *tomb.Tomb) erro
 		conn := conns[id]
 		// device was not unplugged, this is the case if snapd is restarted and we enumerate devices.
 		// note, the situation where device was not unplugged but has changed is handled
-		// by hotlugDeviceAdded handler - updateDevice.
+		// by hotplugDeviceAdded handler - updateDevice.
 		if !conn.HotplugGone || conn.Undesired {
 			continue
 		}
@@ -1466,7 +1710,7 @@ func (m *InterfaceManager) doHotplugConnect(task *state.Task, _ *tomb.Tomb) erro
 	}
 
 	// find new auto-connections
-	autochecker, err := newAutoConnectChecker(st, task, m.repo, deviceCtx)
+	autochecker, err := newAutoConnectChecker(st, m.repo, deviceCtx)
 	if err != nil {
 		return err
 	}
@@ -1479,7 +1723,7 @@ func (m *InterfaceManager) doHotplugConnect(task *state.Task, _ *tomb.Tomb) erro
 	cannotAutoConnectLog := func(plug *snap.PlugInfo, candRefs []string) string {
 		return fmt.Sprintf("cannot auto-connect hotplug slot %s to plug %s, candidates found: %s", slot, plug, strings.Join(candRefs, ", "))
 	}
-	if err := autochecker.addAutoConnections(newconns, candidates, filterForSlot(slot), conns, cannotAutoConnectLog, conflictError); err != nil {
+	if err := autochecker.addAutoConnections(task, newconns, candidates, filterForSlot(slot), conns, cannotAutoConnectLog, conflictError); err != nil {
 		return err
 	}
 
@@ -1756,7 +2000,7 @@ func (m *InterfaceManager) doHotplugSeqWait(task *state.Task, _ *tomb.Tomb) erro
 	}
 
 	for _, otherChg := range st.Changes() {
-		if otherChg.Status().Ready() || otherChg.ID() == chg.ID() {
+		if otherChg.IsReady() || otherChg.ID() == chg.ID() {
 			continue
 		}
 
