@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2021-2023 Canonical Ltd
+ * Copyright (C) 2021-2024 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -23,7 +23,6 @@ import (
 	"bytes"
 	"crypto"
 	"fmt"
-	"io/ioutil"
 	"os"
 	"os/exec"
 	"path/filepath"
@@ -448,7 +447,7 @@ func (s *installSuite) TestEncryptionSupportInfoForceUnencrypted(c *C) {
 		} else {
 			err := os.MkdirAll(filepath.Dir(forceUnencryptedPath), 0755)
 			c.Assert(err, IsNil)
-			err = ioutil.WriteFile(forceUnencryptedPath, nil, 0644)
+			err = os.WriteFile(forceUnencryptedPath, nil, 0644)
 			c.Assert(err, IsNil)
 		}
 
@@ -814,7 +813,7 @@ func (s *installSuite) mockBootloader(c *C, trustedAssets bool, managedAssets bo
 	if trustedAssets || managedAssets {
 		tab := bootloadertest.Mock("trusted", bootloaderRootdir).WithTrustedAssets()
 		if trustedAssets {
-			tab.TrustedAssetsList = []string{"trusted-asset"}
+			tab.TrustedAssetsMap = map[string]string{"trusted-asset": "trusted-asset"}
 		}
 		if managedAssets {
 			tab.ManagedAssetsList = []string{"managed-asset"}
@@ -824,7 +823,7 @@ func (s *installSuite) mockBootloader(c *C, trustedAssets bool, managedAssets bo
 
 		err := os.MkdirAll(boot.InitramfsUbuntuSeedDir, 0755)
 		c.Assert(err, IsNil)
-		err = ioutil.WriteFile(filepath.Join(boot.InitramfsUbuntuSeedDir, "trusted-asset"), nil, 0644)
+		err = os.WriteFile(filepath.Join(boot.InitramfsUbuntuSeedDir, "trusted-asset"), nil, 0644)
 		c.Assert(err, IsNil)
 	} else {
 		bl := bootloadertest.Mock("mock", bootloaderRootdir)
@@ -890,6 +889,10 @@ func (s *installSuite) TestPrepareEncryptedSystemData(c *C) {
 	c.Assert(err, IsNil)
 	c.Assert(to, NotNil)
 
+	// We are required to call ObserveExistingTrustedRecoveryAssets on trusted observers
+	err = to.ObserveExistingTrustedRecoveryAssets(boot.InitramfsUbuntuSeedDir)
+	c.Assert(err, IsNil)
+
 	keyForRole := map[string]keys.EncryptionKey{
 		gadget.SystemData: dataEncryptionKey,
 		gadget.SystemSave: saveKey,
@@ -898,13 +901,13 @@ func (s *installSuite) TestPrepareEncryptedSystemData(c *C) {
 	c.Assert(err, IsNil)
 
 	c.Check(filepath.Join(filepath.Join(dirs.GlobalRootDir, "/run/mnt/ubuntu-data/system-data/var/lib/snapd/device/fde"), "ubuntu-save.key"), testutil.FileEquals, []byte(saveKey))
-	marker, err := ioutil.ReadFile(filepath.Join(filepath.Join(dirs.GlobalRootDir, "/run/mnt/ubuntu-data/system-data/var/lib/snapd/device/fde"), "marker"))
+	marker, err := os.ReadFile(filepath.Join(filepath.Join(dirs.GlobalRootDir, "/run/mnt/ubuntu-data/system-data/var/lib/snapd/device/fde"), "marker"))
 	c.Assert(err, IsNil)
 	c.Check(marker, HasLen, 32)
 	c.Check(filepath.Join(boot.InstallHostFDESaveDir, "marker"), testutil.FileEquals, marker)
 
 	// the assets cache was written to
-	l, err := ioutil.ReadDir(filepath.Join(dirs.SnapBootAssetsDir, "trusted"))
+	l, err := os.ReadDir(filepath.Join(dirs.SnapBootAssetsDir, "trusted"))
 	c.Assert(err, IsNil)
 	c.Assert(l, HasLen, 1)
 }
@@ -970,7 +973,7 @@ func (s *installSuite) TestPrepareRunSystemDataSupportsCloudInitInDangerous(c *C
 	err := os.MkdirAll(cloudCfg, 0755)
 	c.Assert(err, IsNil)
 	for _, mockCfg := range []string{"foo.cfg", "bar.cfg"} {
-		err = ioutil.WriteFile(filepath.Join(cloudCfg, mockCfg), []byte(fmt.Sprintf("%s config", mockCfg)), 0644)
+		err = os.WriteFile(filepath.Join(cloudCfg, mockCfg), []byte(fmt.Sprintf("%s config", mockCfg)), 0644)
 		c.Assert(err, IsNil)
 	}
 
@@ -997,7 +1000,7 @@ func (s *installSuite) TestPrepareRunSystemDataSupportsCloudInitGadgetAndSeedCon
 	err := os.MkdirAll(cloudCfg, 0755)
 	c.Assert(err, IsNil)
 	for _, mockCfg := range []string{"foo.cfg", "bar.cfg"} {
-		err = ioutil.WriteFile(filepath.Join(cloudCfg, mockCfg), []byte(fmt.Sprintf("%s config", mockCfg)), 0644)
+		err = os.WriteFile(filepath.Join(cloudCfg, mockCfg), []byte(fmt.Sprintf("%s config", mockCfg)), 0644)
 		c.Assert(err, IsNil)
 	}
 
@@ -1007,7 +1010,7 @@ func (s *installSuite) TestPrepareRunSystemDataSupportsCloudInitGadgetAndSeedCon
 	})
 
 	// we also have gadget cloud init too
-	err = ioutil.WriteFile(filepath.Join(gadgetDir, "cloud.conf"), nil, 0644)
+	err = os.WriteFile(filepath.Join(gadgetDir, "cloud.conf"), nil, 0644)
 	c.Assert(err, IsNil)
 
 	err = install.PrepareRunSystemData(mockModel, gadgetDir, s.perfTimings)
@@ -1030,7 +1033,7 @@ func (s *installSuite) TestPrepareRunSystemDataSupportsCloudInitBothGadgetAndUbu
 	err := os.MkdirAll(cloudCfg, 0755)
 	c.Assert(err, IsNil)
 	for _, mockCfg := range []string{"foo.cfg", "bar.cfg"} {
-		err = ioutil.WriteFile(filepath.Join(cloudCfg, mockCfg), []byte(fmt.Sprintf("%s config", mockCfg)), 0644)
+		err = os.WriteFile(filepath.Join(cloudCfg, mockCfg), []byte(fmt.Sprintf("%s config", mockCfg)), 0644)
 		c.Assert(err, IsNil)
 	}
 
@@ -1038,7 +1041,7 @@ func (s *installSuite) TestPrepareRunSystemDataSupportsCloudInitBothGadgetAndUbu
 	mockModel := s.mockModel(nil)
 
 	// we also have gadget cloud init too
-	err = ioutil.WriteFile(filepath.Join(gadgetDir, "cloud.conf"), nil, 0644)
+	err = os.WriteFile(filepath.Join(gadgetDir, "cloud.conf"), nil, 0644)
 	c.Assert(err, IsNil)
 
 	err = install.PrepareRunSystemData(mockModel, gadgetDir, s.perfTimings)
@@ -1083,7 +1086,7 @@ func (s *installSuite) TestPrepareRunSystemDataSecuredGadgetCloudConfCloudInit(c
 	})
 
 	// pretend we have a cloud.conf from the gadget
-	err := ioutil.WriteFile(filepath.Join(gadgetDir, "cloud.conf"), nil, 0644)
+	err := os.WriteFile(filepath.Join(gadgetDir, "cloud.conf"), nil, 0644)
 	c.Assert(err, IsNil)
 
 	err = install.PrepareRunSystemData(mockModel, gadgetDir, s.perfTimings)
@@ -1104,7 +1107,7 @@ func (s *installSuite) TestPrepareRunSystemDataSecuredNoUbuntuSeedCloudInit(c *C
 	err := os.MkdirAll(cloudCfg, 0755)
 	c.Assert(err, IsNil)
 	for _, mockCfg := range []string{"foo.cfg", "bar.cfg"} {
-		err = ioutil.WriteFile(filepath.Join(cloudCfg, mockCfg), []byte(fmt.Sprintf("%s config", mockCfg)), 0644)
+		err = os.WriteFile(filepath.Join(cloudCfg, mockCfg), []byte(fmt.Sprintf("%s config", mockCfg)), 0644)
 		c.Assert(err, IsNil)
 	}
 
@@ -1136,7 +1139,7 @@ func (s *installSuite) TestPrepareRunSystemDataWritesTimesyncdClockHappy(c *C) {
 
 	clockTsInSrc := filepath.Join(dirs.GlobalRootDir, "/var/lib/systemd/timesync/clock")
 	c.Assert(os.MkdirAll(filepath.Dir(clockTsInSrc), 0755), IsNil)
-	c.Assert(ioutil.WriteFile(clockTsInSrc, nil, 0644), IsNil)
+	c.Assert(os.WriteFile(clockTsInSrc, nil, 0644), IsNil)
 	// a month old timestamp file
 	c.Assert(os.Chtimes(clockTsInSrc, now.AddDate(0, -1, 0), now.AddDate(0, -1, 0)), IsNil)
 
@@ -1164,7 +1167,7 @@ func (s *installSuite) TestPrepareRunSystemDataWritesTimesyncdClockErr(c *C) {
 
 	clockTsInSrc := filepath.Join(dirs.GlobalRootDir, "/var/lib/systemd/timesync/clock")
 	c.Assert(os.MkdirAll(filepath.Dir(clockTsInSrc), 0755), IsNil)
-	c.Assert(ioutil.WriteFile(clockTsInSrc, nil, 0644), IsNil)
+	c.Assert(os.WriteFile(clockTsInSrc, nil, 0644), IsNil)
 
 	timesyncDirInDst := filepath.Join(filepath.Join(dirs.GlobalRootDir, "/run/mnt/ubuntu-data/system-data"), "/var/lib/systemd/timesync/")
 	c.Assert(os.MkdirAll(timesyncDirInDst, 0755), IsNil)
@@ -1259,7 +1262,7 @@ func (s *installSuite) TestApplyPreseededData(c *C) {
 	model := s.setupCore20Seed(c)
 
 	c.Assert(os.MkdirAll(writableDir, 0755), IsNil)
-	c.Assert(ioutil.WriteFile(preseedArtifact, nil, 0644), IsNil)
+	c.Assert(os.WriteFile(preseedArtifact, nil, 0644), IsNil)
 	c.Assert(os.MkdirAll(filepath.Join(dirs.SnapSeedDir, "snaps"), 0755), IsNil)
 	c.Assert(os.MkdirAll(dirs.SnapBlobDir, 0755), IsNil)
 
@@ -1310,7 +1313,7 @@ func (s *installSuite) TestApplyPreseededData(c *C) {
 		{"pc/1", "pc_1.snap"},
 		{"optional20-a/x1", "optional20-a_x1.snap"},
 	} {
-		c.Assert(osutil.FileExists(filepath.Join(writableDir, "/snap", seedSnap.name)), Equals, true, &dumpDirContents{c, writableDir})
+		c.Assert(osutil.FileExists(filepath.Join(writableDir, dirs.StripRootDir(dirs.SnapMountDir), seedSnap.name)), Equals, true, &dumpDirContents{c, writableDir})
 		c.Assert(osutil.FileExists(filepath.Join(writableDir, dirs.SnapBlobDir, seedSnap.blob)), Equals, true, &dumpDirContents{c, writableDir})
 	}
 
@@ -1347,7 +1350,7 @@ func (s *installSuite) TestApplyPreseededDataAssertionMissing(c *C) {
 	s.setupCore20Seed(c)
 
 	c.Assert(os.MkdirAll(writableDir, 0755), IsNil)
-	c.Assert(ioutil.WriteFile(preseedArtifact, nil, 0644), IsNil)
+	c.Assert(os.WriteFile(preseedArtifact, nil, 0644), IsNil)
 	c.Assert(os.MkdirAll(filepath.Join(dirs.SnapSeedDir, "snaps"), 0755), IsNil)
 	c.Assert(os.MkdirAll(dirs.SnapBlobDir, 0755), IsNil)
 
@@ -1363,7 +1366,7 @@ func (s *installSuite) TestApplyPreseededDataAssertionMissing(c *C) {
 
 	preseedAsPath := filepath.Join(ubuntuSeedDir, "systems", sysLabel, "preseed")
 	// empty "preseed" assertion file
-	c.Assert(ioutil.WriteFile(preseedAsPath, nil, 0644), IsNil)
+	c.Assert(os.WriteFile(preseedAsPath, nil, 0644), IsNil)
 
 	err = install.ApplyPreseededData(preseedSeed, writableDir)
 	c.Assert(err, ErrorMatches, `system preseed assertion file must contain a preseed assertion`)
@@ -1375,8 +1378,8 @@ func (s *installSuite) TestApplyPreseededDataSnapMismatch(c *C) {
 
 	snapPath1 := filepath.Join(dirs.GlobalRootDir, "essential-snap_1.snap")
 	snapPath2 := filepath.Join(dirs.GlobalRootDir, "mode-snap_3.snap")
-	c.Assert(ioutil.WriteFile(snapPath1, nil, 0644), IsNil)
-	c.Assert(ioutil.WriteFile(snapPath2, nil, 0644), IsNil)
+	c.Assert(os.WriteFile(snapPath1, nil, 0644), IsNil)
+	c.Assert(os.WriteFile(snapPath2, nil, 0644), IsNil)
 
 	ubuntuSeedDir := filepath.Join(dirs.GlobalRootDir, "run/mnt/ubuntu-seed")
 	sysLabel := "20220105"
@@ -1384,7 +1387,7 @@ func (s *installSuite) TestApplyPreseededDataSnapMismatch(c *C) {
 	preseedArtifact := filepath.Join(ubuntuSeedDir, "systems", sysLabel, "preseed.tgz")
 	c.Assert(os.MkdirAll(filepath.Join(ubuntuSeedDir, "systems", sysLabel), 0755), IsNil)
 	c.Assert(os.MkdirAll(writableDir, 0755), IsNil)
-	c.Assert(ioutil.WriteFile(preseedArtifact, nil, 0644), IsNil)
+	c.Assert(os.WriteFile(preseedArtifact, nil, 0644), IsNil)
 
 	model := s.mockModel(map[string]interface{}{
 		"grade": "dangerous",
@@ -1460,7 +1463,7 @@ func (s *installSuite) TestApplyPreseededDataWrongDigest(c *C) {
 	defer mockTarCmd.Restore()
 
 	snapPath1 := filepath.Join(dirs.GlobalRootDir, "essential-snap_1.snap")
-	c.Assert(ioutil.WriteFile(snapPath1, nil, 0644), IsNil)
+	c.Assert(os.WriteFile(snapPath1, nil, 0644), IsNil)
 
 	ubuntuSeedDir := filepath.Join(dirs.GlobalRootDir, "run/mnt/ubuntu-seed")
 	sysLabel := "20220105"
@@ -1468,7 +1471,7 @@ func (s *installSuite) TestApplyPreseededDataWrongDigest(c *C) {
 	preseedArtifact := filepath.Join(ubuntuSeedDir, "systems", sysLabel, "preseed.tgz")
 	c.Assert(os.MkdirAll(filepath.Join(ubuntuSeedDir, "systems", sysLabel), 0755), IsNil)
 	c.Assert(os.MkdirAll(writableDir, 0755), IsNil)
-	c.Assert(ioutil.WriteFile(preseedArtifact, nil, 0644), IsNil)
+	c.Assert(os.WriteFile(preseedArtifact, nil, 0644), IsNil)
 
 	model := s.mockModel(map[string]interface{}{
 		"grade": "dangerous",
@@ -1538,11 +1541,11 @@ func (*fakeSeed) LoadEssentialMeta(essentialTypes []snap.Type, tm timings.Measur
 	return nil
 }
 
-func (*fakeSeed) LoadEssentialMetaWithSnapHandler([]snap.Type, seed.SnapHandler, timings.Measurer) error {
+func (*fakeSeed) LoadEssentialMetaWithSnapHandler([]snap.Type, seed.ContainerHandler, timings.Measurer) error {
 	return nil
 }
 
-func (*fakeSeed) LoadMeta(string, seed.SnapHandler, timings.Measurer) error {
+func (*fakeSeed) LoadMeta(string, seed.ContainerHandler, timings.Measurer) error {
 	return nil
 }
 
