@@ -20,9 +20,11 @@
 package snapstate_test
 
 import (
+	"errors"
 	"fmt"
 	"os"
 	"path/filepath"
+	"reflect"
 	"time"
 
 	. "gopkg.in/check.v1"
@@ -34,8 +36,11 @@ import (
 	"github.com/snapcore/snapd/overlord/assertstate"
 	"github.com/snapcore/snapd/overlord/configstate/config"
 	"github.com/snapcore/snapd/overlord/snapstate"
+	"github.com/snapcore/snapd/overlord/snapstate/sequence"
+	"github.com/snapcore/snapd/overlord/snapstate/snapstatetest"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/snap/naming"
 	"github.com/snapcore/snapd/testutil"
 )
 
@@ -45,9 +50,9 @@ func (s *snapmgrTestSuite) TestRemoveTasks(c *C) {
 
 	snapstate.Set(s.state, "foo", &snapstate.SnapState{
 		Active: true,
-		Sequence: []*snap.SideInfo{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
 			{RealName: "foo", Revision: snap.R(11)},
-		},
+		}),
 		Current:  snap.R(11),
 		SnapType: "app",
 	})
@@ -69,9 +74,9 @@ func (s *snapmgrTestSuite) TestRemoveTasksAutoSnapshotDisabled(c *C) {
 
 	snapstate.Set(s.state, "foo", &snapstate.SnapState{
 		Active: true,
-		Sequence: []*snap.SideInfo{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
 			{RealName: "foo", Revision: snap.R(11)},
-		},
+		}),
 		Current:  snap.R(11),
 		SnapType: "app",
 	})
@@ -97,9 +102,9 @@ func (s *snapmgrTestSuite) TestRemoveTasksAutoSnapshotDisabledByPurgeFlag(c *C)
 
 	snapstate.Set(s.state, "foo", &snapstate.SnapState{
 		Active: true,
-		Sequence: []*snap.SideInfo{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
 			{RealName: "foo", Revision: snap.R(11)},
-		},
+		}),
 		Current:  snap.R(11),
 		SnapType: "app",
 	})
@@ -125,10 +130,10 @@ func (s *snapmgrTestSuite) TestRemoveHookNotExecutedIfNotLastRevison(c *C) {
 
 	snapstate.Set(s.state, "foo", &snapstate.SnapState{
 		Active: true,
-		Sequence: []*snap.SideInfo{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
 			{RealName: "foo", Revision: snap.R(11)},
 			{RealName: "foo", Revision: snap.R(12)},
-		},
+		}),
 		Current: snap.R(12),
 	})
 
@@ -146,7 +151,7 @@ func (s *snapmgrTestSuite) TestRemoveConflict(c *C) {
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{{RealName: "some-snap", Revision: snap.R(11)}},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{{RealName: "some-snap", Revision: snap.R(11)}}),
 		Current:  snap.R(11),
 	})
 
@@ -190,7 +195,7 @@ func (s *snapmgrTestSuite) testRemoveDiskSpaceCheck(c *C, featureFlag, automatic
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{{RealName: "some-snap", Revision: snap.R(11)}},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{{RealName: "some-snap", Revision: snap.R(11)}}),
 		Current:  snap.R(11),
 		SnapType: "app",
 	})
@@ -243,7 +248,7 @@ func (s *snapmgrTestSuite) TestRemoveRunThrough(c *C) {
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 		Current:  si.Revision,
 		SnapType: "app",
 	})
@@ -253,7 +258,6 @@ func (s *snapmgrTestSuite) TestRemoveRunThrough(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	defer s.se.Stop()
 	s.settle(c)
 
 	expected := fakeOps{
@@ -385,14 +389,14 @@ func (s *snapmgrTestSuite) TestParallelInstanceRemoveRunThrough(c *C) {
 	// pretend we have both a regular snap and a parallel instance
 	snapstate.Set(s.state, "some-snap_instance", &snapstate.SnapState{
 		Active:      true,
-		Sequence:    []*snap.SideInfo{&si},
+		Sequence:    snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 		Current:     si.Revision,
 		SnapType:    "app",
 		InstanceKey: "instance",
 	})
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 		Current:  si.Revision,
 		SnapType: "app",
 	})
@@ -438,7 +442,7 @@ func (s *snapmgrTestSuite) TestParallelInstanceRemoveRunThrough(c *C) {
 		{
 			op:             "remove-snap-data-dir",
 			name:           "some-snap_instance",
-			path:           filepath.Join(dirs.SnapDataDir, "some-snap"),
+			path:           filepath.Join(dirs.SnapDataDir, "some-snap_instance"),
 			otherInstances: true,
 		},
 		{
@@ -461,7 +465,7 @@ func (s *snapmgrTestSuite) TestParallelInstanceRemoveRunThrough(c *C) {
 		{
 			op:             "remove-snap-dir",
 			name:           "some-snap_instance",
-			path:           filepath.Join(dirs.SnapMountDir, "some-snap"),
+			path:           filepath.Join(dirs.SnapMountDir, "some-snap_instance"),
 			otherInstances: true,
 		},
 	}
@@ -538,14 +542,14 @@ func (s *snapmgrTestSuite) TestParallelInstanceRemoveRunThroughOtherInstances(c
 	// pretend we have both a regular snap and a parallel instance
 	snapstate.Set(s.state, "some-snap_instance", &snapstate.SnapState{
 		Active:      true,
-		Sequence:    []*snap.SideInfo{&si},
+		Sequence:    snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 		Current:     si.Revision,
 		SnapType:    "app",
 		InstanceKey: "instance",
 	})
 	snapstate.Set(s.state, "some-snap_other", &snapstate.SnapState{
 		Active:      true,
-		Sequence:    []*snap.SideInfo{&si},
+		Sequence:    snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 		Current:     si.Revision,
 		SnapType:    "app",
 		InstanceKey: "other",
@@ -592,7 +596,7 @@ func (s *snapmgrTestSuite) TestParallelInstanceRemoveRunThroughOtherInstances(c
 		{
 			op:             "remove-snap-data-dir",
 			name:           "some-snap_instance",
-			path:           filepath.Join(dirs.SnapDataDir, "some-snap"),
+			path:           filepath.Join(dirs.SnapDataDir, "some-snap_instance"),
 			otherInstances: true,
 		},
 		{
@@ -615,7 +619,7 @@ func (s *snapmgrTestSuite) TestParallelInstanceRemoveRunThroughOtherInstances(c
 		{
 			op:             "remove-snap-dir",
 			name:           "some-snap_instance",
-			path:           filepath.Join(dirs.SnapMountDir, "some-snap"),
+			path:           filepath.Join(dirs.SnapMountDir, "some-snap_instance"),
 			otherInstances: true,
 		},
 	}
@@ -658,7 +662,7 @@ func (s *snapmgrTestSuite) TestRemoveWithManyRevisionsRunThrough(c *C) {
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si5, &si3, &si7},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si5, &si3, &si7}),
 		Current:  si7.Revision,
 		SnapType: "app",
 	})
@@ -668,7 +672,6 @@ func (s *snapmgrTestSuite) TestRemoveWithManyRevisionsRunThrough(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	defer s.se.Stop()
 	s.settle(c)
 
 	expected := fakeOps{
@@ -831,7 +834,7 @@ func (s *snapmgrTestSuite) TestRemoveOneRevisionRunThrough(c *C) {
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si5, &si3, &si7},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si5, &si3, &si7}),
 		Current:  si7.Revision,
 		SnapType: "app",
 	})
@@ -841,7 +844,6 @@ func (s *snapmgrTestSuite) TestRemoveOneRevisionRunThrough(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	defer s.se.Stop()
 	s.settle(c)
 
 	c.Check(len(s.fakeBackend.ops), Equals, 2)
@@ -884,7 +886,7 @@ func (s *snapmgrTestSuite) TestRemoveOneRevisionRunThrough(c *C) {
 	var snapst snapstate.SnapState
 	err = snapstate.Get(s.state, "some-snap", &snapst)
 	c.Assert(err, IsNil)
-	c.Check(snapst.Sequence, HasLen, 2)
+	c.Check(snapst.Sequence.Revisions, HasLen, 2)
 }
 
 func (s *snapmgrTestSuite) TestRemoveOneRevisionDropsRevertStatus(c *C) {
@@ -908,7 +910,7 @@ func (s *snapmgrTestSuite) TestRemoveOneRevisionDropsRevertStatus(c *C) {
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si5, &si3, &si7},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si5, &si3, &si7}),
 		Current:  si7.Revision,
 		RevertStatus: map[int]snapstate.RevertStatus{
 			3: snapstate.NotBlocked,
@@ -922,14 +924,13 @@ func (s *snapmgrTestSuite) TestRemoveOneRevisionDropsRevertStatus(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	defer s.se.Stop()
 	s.settle(c)
 
 	// verify snaps in the system state
 	var snapst snapstate.SnapState
 	err = snapstate.Get(s.state, "some-snap", &snapst)
 	c.Assert(err, IsNil)
-	c.Check(snapst.Sequence, HasLen, 2)
+	c.Check(snapst.Sequence.Revisions, HasLen, 2)
 	// revert status of revision 3 got dropped
 	c.Check(snapst.RevertStatus, DeepEquals, map[int]snapstate.RevertStatus{
 		5: snapstate.NotBlocked,
@@ -947,7 +948,7 @@ func (s *snapmgrTestSuite) TestRemoveLastRevisionRunThrough(c *C) {
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   false,
-		Sequence: []*snap.SideInfo{&si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 		Current:  si.Revision,
 		SnapType: "app",
 	})
@@ -957,7 +958,6 @@ func (s *snapmgrTestSuite) TestRemoveLastRevisionRunThrough(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	defer s.se.Stop()
 	s.settle(c)
 
 	c.Check(len(s.fakeBackend.ops), Equals, 10)
@@ -1057,7 +1057,7 @@ func (s *snapmgrTestSuite) TestRemoveCurrentActiveRevisionRefused(c *C) {
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 		Current:  si.Revision,
 		SnapType: "app",
 	})
@@ -1078,7 +1078,7 @@ func (s *snapmgrTestSuite) TestRemoveCurrentRevisionOfSeveralRefused(c *C) {
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si, &si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si, &si}),
 		Current:  si.Revision,
 		SnapType: "app",
 	})
@@ -1099,7 +1099,7 @@ func (s *snapmgrTestSuite) TestRemoveMissingRevisionRefused(c *C) {
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 		Current:  si.Revision,
 		SnapType: "app",
 	})
@@ -1120,7 +1120,7 @@ func (s *snapmgrTestSuite) TestRemoveRefused(c *C) {
 
 	snapstate.Set(s.state, "brand-gadget", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 		Current:  si.Revision,
 		SnapType: "gadget",
 	})
@@ -1141,7 +1141,7 @@ func (s *snapmgrTestSuite) TestRemoveRefusedLastRevision(c *C) {
 
 	snapstate.Set(s.state, "brand-gadget", &snapstate.SnapState{
 		Active:   false,
-		Sequence: []*snap.SideInfo{&si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 		Current:  si.Revision,
 		SnapType: "gadget",
 	})
@@ -1162,14 +1162,14 @@ func (s *snapmgrTestSuite) TestRemoveDeletesConfigOnLastRevision(c *C) {
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 		Current:  si.Revision,
 		SnapType: "app",
 	})
 
 	snapstate.Set(s.state, "another-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 		Current:  si.Revision,
 		SnapType: "app",
 	})
@@ -1193,7 +1193,6 @@ func (s *snapmgrTestSuite) TestRemoveDeletesConfigOnLastRevision(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	defer s.se.Stop()
 	s.settle(c)
 
 	// verify snaps in the system state
@@ -1226,7 +1225,7 @@ func (s *snapmgrTestSuite) TestRemoveDoesntDeleteConfigIfNotLastRevision(c *C) {
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si1, &si2},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si1, &si2}),
 		Current:  si2.Revision,
 		SnapType: "app",
 	})
@@ -1244,7 +1243,6 @@ func (s *snapmgrTestSuite) TestRemoveDoesntDeleteConfigIfNotLastRevision(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	defer s.se.Stop()
 	s.settle(c)
 
 	// verify snaps in the system state
@@ -1263,16 +1261,16 @@ func (s *snapmgrTestSuite) TestRemoveMany(c *C) {
 
 	snapstate.Set(s.state, "one", &snapstate.SnapState{
 		Active: true,
-		Sequence: []*snap.SideInfo{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
 			{RealName: "one", SnapID: "one-id", Revision: snap.R(1)},
-		},
+		}),
 		Current: snap.R(1),
 	})
 	snapstate.Set(s.state, "two", &snapstate.SnapState{
 		Active: true,
-		Sequence: []*snap.SideInfo{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
 			{RealName: "two", SnapID: "two-id", Revision: snap.R(1)},
-		},
+		}),
 		Current: snap.R(1),
 	})
 
@@ -1353,17 +1351,17 @@ func (s *snapmgrTestSuite) testRemoveManyDiskSpaceCheck(c *C, featureFlag, autom
 	snapstate.Set(s.state, "one", &snapstate.SnapState{
 		Active:   true,
 		SnapType: "app",
-		Sequence: []*snap.SideInfo{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
 			{RealName: "one", SnapID: "one-id", Revision: snap.R(1)},
-		},
+		}),
 		Current: snap.R(1),
 	})
 	snapstate.Set(s.state, "two", &snapstate.SnapState{
 		Active:   true,
 		SnapType: "app",
-		Sequence: []*snap.SideInfo{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
 			{RealName: "two", SnapID: "two-id", Revision: snap.R(1)},
-		},
+		}),
 		Current: snap.R(1),
 	})
 
@@ -1492,7 +1490,7 @@ func makeTestSnaps(c *C, st *state.State) {
 
 	snapstate.Set(st, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si1, &si2},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si1, &si2}),
 		Current:  si1.Revision,
 		SnapType: "app",
 	})
@@ -1522,7 +1520,6 @@ func (s *snapmgrTestSuite) TestRemoveManyUndoRestoresCurrent(c *C) {
 	// are still present.
 	injectError(c, chg, "clear-snap", snap.Revision{N: 1})
 
-	defer s.se.Stop()
 	s.settle(c)
 
 	c.Assert(chg.Status(), Equals, state.ErrorStatus)
@@ -1532,8 +1529,8 @@ func (s *snapmgrTestSuite) TestRemoveManyUndoRestoresCurrent(c *C) {
 	c.Assert(snapstate.Get(s.state, "some-snap", &snapst), IsNil)
 	c.Check(snapst.Active, Equals, true)
 	c.Check(snapst.Current, Equals, snap.Revision{N: 1})
-	c.Assert(snapst.Sequence, HasLen, 1)
-	c.Check(snapst.Sequence[0].Revision, Equals, snap.Revision{N: 1})
+	c.Assert(snapst.Sequence.Revisions, HasLen, 1)
+	c.Check(snapst.Sequence.Revisions[0].Snap.Revision, Equals, snap.Revision{N: 1})
 
 	expected := fakeOps{
 		{
@@ -1600,7 +1597,6 @@ func (s *snapmgrTestSuite) TestRemoveManyUndoLeavesInactiveSnapAfterDataIsLost(c
 	// current rev 1), before discarding the snap completely.
 	injectError(c, chg, "discard-snap", snap.Revision{N: 1})
 
-	defer s.se.Stop()
 	s.settle(c)
 
 	c.Assert(chg.Status(), Equals, state.ErrorStatus)
@@ -1613,8 +1609,8 @@ func (s *snapmgrTestSuite) TestRemoveManyUndoLeavesInactiveSnapAfterDataIsLost(c
 	// after its data was removed.
 	c.Check(snapst.Active, Equals, false)
 	c.Check(snapst.Current, Equals, snap.Revision{N: 1})
-	c.Assert(snapst.Sequence, HasLen, 1)
-	c.Check(snapst.Sequence[0].Revision, Equals, snap.Revision{N: 1})
+	c.Assert(snapst.Sequence.Revisions, HasLen, 1)
+	c.Check(snapst.Sequence.Revisions[0].Snap.Revision, Equals, snap.Revision{N: 1})
 
 	expected := fakeOps{
 		{
@@ -1696,7 +1692,7 @@ func (s *snapmgrTestSuite) TestRemovePrunesRefreshGatingDataOnLastRevision(c *C)
 		t := time.Now()
 		snapstate.Set(s.state, sn, &snapstate.SnapState{
 			Active:          true,
-			Sequence:        []*snap.SideInfo{&si},
+			Sequence:        snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 			Current:         si.Revision,
 			SnapType:        "app",
 			LastRefreshTime: &t,
@@ -1704,8 +1700,22 @@ func (s *snapmgrTestSuite) TestRemovePrunesRefreshGatingDataOnLastRevision(c *C)
 	}
 
 	rc := map[string]*snapstate.RefreshCandidate{
-		"some-snap": {},
-		"foo-snap":  {},
+		"some-snap": {
+			SnapSetup: snapstate.SnapSetup{
+				SideInfo: &snap.SideInfo{
+					RealName: "some-snap",
+					Revision: snap.R(7),
+				},
+			},
+		},
+		"foo-snap": {
+			SnapSetup: snapstate.SnapSetup{
+				SideInfo: &snap.SideInfo{
+					RealName: "foo-snap",
+					Revision: snap.R(7),
+				},
+			},
+		},
 	}
 	st.Set("refresh-candidates", rc)
 
@@ -1716,9 +1726,9 @@ func (s *snapmgrTestSuite) TestRemovePrunesRefreshGatingDataOnLastRevision(c *C)
 
 	held, err := snapstate.HeldSnaps(st, snapstate.HoldAutoRefresh)
 	c.Assert(err, IsNil)
-	c.Check(held, DeepEquals, map[string]bool{
-		"foo-snap":  true,
-		"some-snap": true,
+	c.Check(held, DeepEquals, map[string][]string{
+		"foo-snap":  {"some-snap"},
+		"some-snap": {"another-snap"},
 	})
 
 	chg := st.NewChange("remove", "remove a snap")
@@ -1726,7 +1736,6 @@ func (s *snapmgrTestSuite) TestRemovePrunesRefreshGatingDataOnLastRevision(c *C)
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	defer s.se.Stop()
 	s.settle(c)
 
 	// verify snaps in the system state
@@ -1754,16 +1763,23 @@ func (s *snapmgrTestSuite) TestRemoveKeepsGatingDataIfNotLastRevision(c *C) {
 	t := time.Now()
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active: true,
-		Sequence: []*snap.SideInfo{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
 			{RealName: "some-snap", Revision: snap.R(11)},
 			{RealName: "some-snap", Revision: snap.R(12)},
-		},
+		}),
 		Current:         snap.R(11),
 		SnapType:        "app",
 		LastRefreshTime: &t,
 	})
 
-	rc := map[string]*snapstate.RefreshCandidate{"some-snap": {}}
+	rc := map[string]*snapstate.RefreshCandidate{
+		"some-snap": {
+			SnapSetup: snapstate.SnapSetup{
+				SideInfo: &snap.SideInfo{
+					RealName: "some-snap",
+					Revision: snap.R(11),
+				},
+			}}}
 	st.Set("refresh-candidates", rc)
 
 	_, err := snapstate.HoldRefresh(st, snapstate.HoldAutoRefresh, "some-snap", 0, "some-snap")
@@ -1771,14 +1787,13 @@ func (s *snapmgrTestSuite) TestRemoveKeepsGatingDataIfNotLastRevision(c *C) {
 
 	held, err := snapstate.HeldSnaps(st, snapstate.HoldAutoRefresh)
 	c.Assert(err, IsNil)
-	c.Check(held, DeepEquals, map[string]bool{"some-snap": true})
+	c.Check(held, DeepEquals, map[string][]string{"some-snap": {"some-snap"}})
 
 	chg := st.NewChange("remove", "remove a snap")
 	ts, err := snapstate.Remove(st, "some-snap", snap.R(12), nil)
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	defer s.se.Stop()
 	s.settle(c)
 
 	// verify snap in the system state
@@ -1788,7 +1803,7 @@ func (s *snapmgrTestSuite) TestRemoveKeepsGatingDataIfNotLastRevision(c *C) {
 	// held snaps are intact
 	held, err = snapstate.HeldSnaps(st, snapstate.HoldAutoRefresh)
 	c.Assert(err, IsNil)
-	c.Check(held, DeepEquals, map[string]bool{"some-snap": true})
+	c.Check(held, DeepEquals, map[string][]string{"some-snap": {"some-snap"}})
 
 	// refresh-candidates are intact
 	var candidates map[string]*snapstate.RefreshCandidate
@@ -1820,7 +1835,7 @@ func (s *validationSetsSuite) removeSnapReferencedByValidationSet(c *C, presence
 	}
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si}),
 		Current:  si.Revision,
 		SnapType: "app",
 	})
@@ -1886,7 +1901,7 @@ func (s *validationSetsSuite) TestRemoveSnapRequiredByValidationSetAtSpecificRev
 	}
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si1, &si2, &si3},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si1, &si2, &si3}),
 		Current:  si1.Revision,
 		SnapType: "app",
 	})
@@ -1937,7 +1952,7 @@ func (s *validationSetsSuite) TestRemoveSnapRequiredByValidationSetAtSpecificRev
 	// rev 1 is active
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{&si1, &si2},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{&si1, &si2}),
 		Current:  si1.Revision,
 		SnapType: "app",
 	})
@@ -1981,21 +1996,21 @@ func (s *snapmgrTestSuite) TestRemoveDeduplicatesSnapNames(c *C) {
 	defer s.state.Unlock()
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
-		Sequence: []*snap.SideInfo{{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{{
 			RealName: "some-snap",
 			SnapID:   "some-snap-id",
 			Revision: snap.R(1),
-		}},
+		}}),
 		Current: snap.R(1),
 		Active:  true,
 	})
 
 	snapstate.Set(s.state, "some-base", &snapstate.SnapState{
-		Sequence: []*snap.SideInfo{{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{{
 			RealName: "some-base",
 			SnapID:   "some-base-id",
 			Revision: snap.R(1),
-		}},
+		}}),
 		Current: snap.R(1),
 		Active:  true,
 	})
@@ -2011,18 +2026,18 @@ func (s *snapmgrTestSuite) TestRemoveManyWithPurge(c *C) {
 	defer s.state.Unlock()
 
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
-		Sequence: []*snap.SideInfo{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
 			{RealName: "some-snap", SnapID: "some-snap-id", Revision: snap.R(1)},
-		},
+		}),
 		Current:  snap.R(1),
 		SnapType: "app",
 		Active:   true,
 	})
 
 	snapstate.Set(s.state, "foo", &snapstate.SnapState{
-		Sequence: []*snap.SideInfo{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
 			{RealName: "foo", SnapID: "foo-id", Revision: snap.R(11)},
-		},
+		}),
 		Current:  snap.R(11),
 		SnapType: "app",
 		Active:   true,
@@ -2048,3 +2063,354 @@ func (s *snapmgrTestSuite) TestRemoveManyWithPurge(c *C) {
 	}
 
 }
+
+func (s *snapmgrTestSuite) TestRemoveWithCompsTasks(c *C) {
+	const snapName = "snap1"
+	const comp1name = "comp1"
+	const comp2name = "comp2"
+
+	cref1 := naming.NewComponentRef(snapName, comp1name)
+	cref2 := naming.NewComponentRef(snapName, comp2name)
+
+	s.AddCleanup(snapstate.MockReadComponentInfo(func(compMntDir string,
+		snapInfo *snap.Info, csi *snap.ComponentSideInfo) (*snap.ComponentInfo, error) {
+		switch csi.Component.ComponentName {
+		case comp1name:
+			return &snap.ComponentInfo{
+				Component:         cref1,
+				Type:              snap.TestComponent,
+				ComponentSideInfo: *csi,
+			}, nil
+		case comp2name:
+			return &snap.ComponentInfo{
+				Component:         cref2,
+				Type:              snap.TestComponent,
+				ComponentSideInfo: *csi,
+			}, nil
+		}
+		return nil, errors.New("unexpected component")
+	}))
+
+	s.AddCleanup(snapstate.MockSnapReadInfo(func(name string, si *snap.SideInfo) (*snap.Info, error) {
+		info := &snap.Info{
+			SuggestedName: name,
+			SideInfo:      *si,
+			SnapType:      snap.TypeApp,
+			Components: map[string]*snap.Component{
+				comp1name: {Name: comp1name, Type: snap.TestComponent},
+				comp2name: {Name: comp2name, Type: snap.TestComponent},
+			},
+		}
+		info.Apps = map[string]*snap.AppInfo{
+			"app": {Snap: info, Name: "app"},
+		}
+		return info, nil
+	}))
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	siRev1 := &snap.SideInfo{RealName: snapName, Revision: snap.R(1)}
+	siRev2 := &snap.SideInfo{RealName: snapName, Revision: snap.R(2)}
+	comp1si := snap.NewComponentSideInfo(cref1, snap.R(11))
+	comp11si := snap.NewComponentSideInfo(cref1, snap.R(111))
+	comp2si := snap.NewComponentSideInfo(cref2, snap.R(22))
+	comp22si := snap.NewComponentSideInfo(cref2, snap.R(222))
+	seq := snapstatetest.NewSequenceFromRevisionSideInfos(
+		[]*sequence.RevisionSideState{
+			sequence.NewRevisionSideState(siRev1,
+				[]*sequence.ComponentState{
+					sequence.NewComponentState(
+						comp1si, snap.TestComponent),
+					sequence.NewComponentState(
+						comp2si, snap.TestComponent),
+				}),
+			sequence.NewRevisionSideState(siRev2,
+				[]*sequence.ComponentState{
+					sequence.NewComponentState(
+						comp11si, snap.TestComponent),
+					sequence.NewComponentState(
+						comp22si, snap.TestComponent),
+				}),
+		})
+	snapstate.Set(s.state, snapName, &snapstate.SnapState{
+		Active:   true,
+		Sequence: seq,
+		Current:  snap.R(1),
+		SnapType: "app",
+	})
+
+	ts, err := snapstate.Remove(s.state, snapName, snap.R(0), nil)
+	c.Assert(err, IsNil)
+
+	c.Assert(s.state.TaskCount(), Equals, len(ts.Tasks()))
+	c.Assert(taskKinds(ts.Tasks()), DeepEquals, []string{
+		"stop-snap-services",
+		"run-hook[remove]",
+		"auto-disconnect",
+		"save-snapshot",
+		"remove-aliases",
+		"unlink-snap",
+		"remove-profiles",
+		"clear-snap",
+		"unlink-component",
+		"discard-component",
+		"unlink-component",
+		"discard-component",
+		"discard-snap",
+		"clear-snap",
+		"unlink-component",
+		"discard-component",
+		"unlink-component",
+		"discard-component",
+		"discard-snap",
+	})
+	verifyStopReason(c, ts, "remove")
+
+	// Run the created tasks
+	chg := s.state.NewChange("remove", "remove a snap")
+	c.Assert(err, IsNil)
+	chg.AddAll(ts)
+
+	s.settle(c)
+
+	expected := fakeOps{
+		{
+			op:    "auto-disconnect:Doing",
+			name:  "snap1",
+			revno: snap.R(1),
+		},
+		{
+			op:   "remove-snap-aliases",
+			name: "snap1",
+		},
+		{
+			op:   "unlink-snap",
+			path: filepath.Join(dirs.SnapMountDir, "snap1/1"),
+		},
+		{
+			op:    "remove-profiles:Doing",
+			name:  "snap1",
+			revno: snap.R(1),
+		},
+		{
+			op:   "remove-snap-data",
+			path: filepath.Join(dirs.SnapMountDir, "snap1/2"),
+		},
+		{
+			op:   "unlink-component",
+			path: filepath.Join(dirs.SnapMountDir, "snap1/components/mnt/comp1/111"),
+		},
+		{
+			op:                "undo-setup-component",
+			containerName:     "snap1+comp1",
+			containerFileName: "snap1+comp1_111.comp",
+		},
+		{
+			op:                "remove-component-dir",
+			containerName:     "snap1+comp1",
+			containerFileName: "snap1+comp1_111.comp",
+		},
+		{
+			op:   "unlink-component",
+			path: filepath.Join(dirs.SnapMountDir, "snap1/components/mnt/comp2/222"),
+		},
+		{
+			op:                "undo-setup-component",
+			containerName:     "snap1+comp2",
+			containerFileName: "snap1+comp2_222.comp",
+		},
+		{
+			op:                "remove-component-dir",
+			containerName:     "snap1+comp2",
+			containerFileName: "snap1+comp2_222.comp",
+		},
+		{
+			op:    "remove-snap-files",
+			path:  filepath.Join(dirs.SnapMountDir, "snap1/2"),
+			stype: "app",
+		},
+		{
+			op:   "remove-snap-data",
+			path: filepath.Join(dirs.SnapMountDir, "snap1/1"),
+		},
+		{
+			op:   "remove-snap-common-data",
+			path: filepath.Join(dirs.SnapMountDir, "snap1/1"),
+		},
+		{
+			op:   "remove-snap-save-data",
+			path: filepath.Join(dirs.SnapDataSaveDir, "snap1"),
+		},
+		{
+			op:   "remove-snap-data-dir",
+			name: "snap1",
+			path: filepath.Join(dirs.SnapDataDir, "snap1"),
+		},
+		{
+			op:   "unlink-component",
+			path: filepath.Join(dirs.SnapMountDir, "snap1/components/mnt/comp1/11"),
+		},
+		{
+			op:                "undo-setup-component",
+			containerName:     "snap1+comp1",
+			containerFileName: "snap1+comp1_11.comp",
+		},
+		{
+			containerName:     "snap1+comp1",
+			containerFileName: "snap1+comp1_11.comp",
+			op:                "remove-component-dir",
+		},
+		{
+			op:   "unlink-component",
+			path: filepath.Join(dirs.SnapMountDir, "snap1/components/mnt/comp2/22"),
+		},
+		{
+			op:                "undo-setup-component",
+			containerName:     "snap1+comp2",
+			containerFileName: "snap1+comp2_22.comp",
+		},
+		{
+			op:                "remove-component-dir",
+			containerName:     "snap1+comp2",
+			containerFileName: "snap1+comp2_22.comp",
+		},
+		{
+			op:    "remove-snap-files",
+			path:  filepath.Join(dirs.SnapMountDir, "snap1/1"),
+			stype: "app",
+		},
+		{
+			op:   "remove-snap-mount-units",
+			name: "snap1",
+		},
+		{
+			op:   "discard-namespace",
+			name: "snap1",
+		},
+		{
+			op:   "remove-inhibit-lock",
+			name: "snap1",
+		},
+		{
+			op:   "remove-snap-dir",
+			name: "snap1",
+			path: filepath.Join(dirs.SnapMountDir, "snap1"),
+		},
+	}
+	// start with an easier-to-read error if this fails:
+	c.Check(len(s.fakeBackend.ops), Equals, len(expected))
+	c.Check(s.fakeBackend.ops[0:5], DeepEquals, expected[0:5])
+	c.Check(s.fakeBackend.ops[11:16], DeepEquals, expected[11:16])
+	c.Check(s.fakeBackend.ops[22:], DeepEquals, expected[22:])
+	// Check component tasks, that can run in parallel so the order is not
+	// deterministic, but still needs to follow an order per component.
+	checkComps := func(ops, exp1, exp2 fakeOps) {
+		var op1idx, op2idx int
+		for _, op := range ops {
+			switch {
+			case op1idx < 3 && reflect.DeepEqual(exp1[op1idx], op):
+				op1idx++
+			case op2idx < 3 && reflect.DeepEqual(exp2[op2idx], op):
+				op2idx++
+			default:
+				c.Error("expected op not found", op)
+			}
+		}
+	}
+	checkComps(s.fakeBackend.ops[5:11], expected[5:8], expected[8:11])
+	checkComps(s.fakeBackend.ops[16:22], expected[16:19], expected[19:22])
+}
+
+func (s *snapmgrTestSuite) TestRemoveWithTerminate(c *C) {
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	snapstate.Set(s.state, "foo", &snapstate.SnapState{
+		Active: true,
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
+			{RealName: "foo", Revision: snap.R(11)},
+		}),
+		Current:  snap.R(11),
+		SnapType: "app",
+	})
+
+	ts, err := snapstate.Remove(s.state, "foo", snap.R(0), &snapstate.RemoveFlags{Terminate: true})
+	c.Assert(err, IsNil)
+
+	c.Assert(taskKinds(ts.Tasks()), DeepEquals, []string{
+		"stop-snap-services",
+		"run-hook[remove]",
+		"auto-disconnect",
+		"kill-snap-apps",
+		"save-snapshot",
+		"remove-aliases",
+		"unlink-snap",
+		"remove-profiles",
+		"clear-snap",
+		"discard-snap",
+	})
+}
+
+func (s *snapmgrTestSuite) TestRemoveWithTerminateAndRevisionSet(c *C) {
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	snapstate.Set(s.state, "foo", &snapstate.SnapState{
+		Active: true,
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
+			{RealName: "foo", Revision: snap.R(11)},
+			{RealName: "foo", Revision: snap.R(10)},
+		}),
+		Current:  snap.R(11),
+		SnapType: "app",
+	})
+
+	_, err := snapstate.Remove(s.state, "foo", snap.R(10), &snapstate.RemoveFlags{Terminate: true})
+	c.Assert(err, ErrorMatches, "cannot terminate running apps unless all revisions are removed")
+}
+
+func (s *snapmgrTestSuite) TestRemoveManyWithTerminate(c *C) {
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
+			{RealName: "some-snap", SnapID: "some-snap-id", Revision: snap.R(1)},
+		}),
+		Current:  snap.R(1),
+		SnapType: "app",
+		Active:   true,
+	})
+
+	snapstate.Set(s.state, "foo", &snapstate.SnapState{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
+			{RealName: "foo", SnapID: "foo-id", Revision: snap.R(11)},
+		}),
+		Current:  snap.R(11),
+		SnapType: "app",
+		Active:   true,
+	})
+
+	removed, tss, err := snapstate.RemoveMany(s.state, []string{"some-snap", "foo"}, &snapstate.RemoveFlags{Terminate: true})
+	c.Check(removed, DeepEquals, []string{"some-snap", "foo"})
+	c.Check(tss, NotNil)
+	c.Check(err, IsNil)
+
+	// tasks to take snapshot aren't generated
+	for _, ts := range tss {
+		c.Assert(taskKinds(ts.Tasks()), DeepEquals, []string{
+			"stop-snap-services",
+			"run-hook[remove]",
+			"auto-disconnect",
+			"kill-snap-apps",
+			"save-snapshot",
+			"remove-aliases",
+			"unlink-snap",
+			"remove-profiles",
+			"clear-snap",
+			"discard-snap",
+		})
+	}
+
+}
