@@ -23,46 +23,82 @@
 : "${NESTED_DISK_LOGICAL_BLOCK_SIZE:=512}"
 
 nested_wait_for_ssh() {
-    # TODO:UC20: the retry count should be lowered to something more reasonable.
-    local retry=800
-    local wait=1
+    local retry=${1:-800}
+    local wait=${2:-1}
 
-    until remote.exec "true"; do
-        retry=$(( retry - 1 ))
-        if [ $retry -le 0 ]; then
-            echo "Timed out waiting for command 'true' to succeed. Aborting!"
+    until remote.exec "true" &>/dev/null; do
+        if [ "$retry" -le 0 ]; then
             return 1
         fi
+        retry=$(( retry - 1 ))
         sleep "$wait"
     done
 }
 
 nested_wait_for_no_ssh() {
-    local retry=200
-    local wait=1
+    local retry=${1:-200}
+    local wait=${2:-1}
 
-    while remote.exec "true"; do
-        retry=$(( retry - 1 ))
-        if [ $retry -le 0 ]; then
-            echo "Timed out waiting for command 'true' to fail. Aborting!"
+    while remote.exec "true" &>/dev/null; do
+        if [ "$retry" -le 0 ]; then
             return 1
         fi
+        retry=$(( retry - 1 ))
         sleep "$wait"
     done
 }
 
+nested_wait_vm_ready() {
+    echo "Waiting the vm is ready to be used"
+    local retry=${1:-120}
+
+    local serial_log="$NESTED_LOGS_DIR"/serial.log
+    while true; do
+        # Check the timeout is reached
+        if [ "$retry" -le 0 ]; then
+            echo "Timed out waiting for vm ready. Aborting!"
+            return 1
+        fi
+        retry=$(( retry - 1 ))
+
+        # Check the vm is active
+        if ! systemctl is-active "$NESTED_VM"; then
+            echo "Unit $NESTED_VM is not active. Aborting!"
+            return 1
+        fi
+
+        # Check if ssh connection can be established, and return if it is possible
+        if nested_wait_for_ssh 1 1; then
+            echo "SSH connection ready"
+            return
+        fi
+
+        # Check no infinite loops during boot
+        if nested_is_core_ge 20; then
+            test "$(grep -c -E "Command line:.*snapd_recovery_mode=install" "$serial_log")" -le 1
+            test "$(grep -c -E "Command line:.*snapd_recovery_mode=run" "$serial_log")" -le 1
+        else
+            test "$(grep -c -E "Command line:.*BOOT_IMAGE=\(loop\)/kernel.img" "$serial_log")" -le 1
+        fi
+
+        sleep 3
+    done
+
+    nested_check_unit_stays_active "$NESTED_VM" 2 1
+}
+
 nested_wait_for_snap_command() {
     # In this function the remote retry command cannot be used because it could
     # be executed before the tool is deployed.
-    local retry=200
-    local wait=1
+    local retry=${1:-200}
+    local wait=${2:-1}
 
-    while ! remote.exec "command -v snap"; do
-        retry=$(( retry - 1 ))
-        if [ $retry -le 0 ]; then
+    while ! remote.exec "command -v snap" &>/dev/null; do
+        if [ "$retry" -le 0 ]; then
             echo "Timed out waiting for command 'command -v snap' to success. Aborting!"
             return 1
         fi
+        retry=$(( retry - 1 ))
         sleep "$wait"
     done
 }
@@ -110,8 +146,8 @@ nested_uc20_transition_to_system_mode() {
     local recovery_system="$1"
     local mode="$2"
 
-    if ! nested_is_core_20_system && ! nested_is_core_22_system; then
-        echo "Transition can be done just on uc20 and uc22 systems, exiting..."
+    if nested_is_core_le 18; then
+        echo "Transition can be done just on uc20+ systems, exiting..."
         exit 1
     fi
 
@@ -130,17 +166,23 @@ nested_uc20_transition_to_system_mode() {
 }
 
 nested_prepare_ssh() {
-    remote.exec "sudo adduser --uid 12345 --extrausers --quiet --disabled-password --gecos '' test"
-    remote.exec "echo test:ubuntu | sudo chpasswd"
+    if nested_is_core_ge 24; then
+        remote.exec "sudo useradd --uid 12345 --create-home --extrausers test"
+        remote.exec "sudo useradd --create-home --extrausers external"
+    else 
+        remote.exec "sudo adduser --uid 12345 --extrausers --quiet --disabled-password --gecos '' test"
+        remote.exec "sudo adduser --extrausers --quiet --disabled-password --gecos '' external"
+    fi
+    
+    remote.exec "echo test:ubuntu123 | sudo chpasswd"
     remote.exec "echo 'test ALL=(ALL) NOPASSWD:ALL' | sudo tee /etc/sudoers.d/create-user-test"
     # Check we can connect with the new test user and make sudo
-    remote.exec --user test --pass ubuntu "sudo true"
+    remote.exec --user test --pass ubuntu123 "sudo true"
 
-    remote.exec "sudo adduser --extrausers --quiet --disabled-password --gecos '' external"
-    remote.exec "echo external:ubuntu | sudo chpasswd"
+    remote.exec "echo external:ubuntu123 | sudo chpasswd"
     remote.exec "echo 'external ALL=(ALL) NOPASSWD:ALL' | sudo tee /etc/sudoers.d/create-user-external"
     # Check we can connect with the new external user and make sudo
-    remote.exec --user external --pass ubuntu "sudo true"
+    remote.exec --user external --pass ubuntu123 "sudo true"
 }
 
 
@@ -287,6 +329,30 @@ nested_is_classic_system() {
     test "$NESTED_TYPE" = "classic"
 }
 
+nested_is_core_ge() {
+    local VERSION=$1
+    os.query is-ubuntu-ge "${VERSION}.04"
+}
+
+nested_is_core_gt() {
+    local VERSION=$1
+    os.query is-ubuntu-gt "${VERSION}.04"
+}
+
+nested_is_core_le() {
+    local VERSION=$1
+    os.query is-ubuntu-le "${VERSION}.04"
+}
+
+nested_is_core_lt() {
+    local VERSION=$1
+    os.query is-ubuntu-lt "${VERSION}.04"
+}
+
+nested_is_core_24_system() {
+    os.query is-noble
+}
+
 nested_is_core_22_system() {
     os.query is-jammy
 }
@@ -316,13 +382,13 @@ nested_refresh_to_new_core() {
             remote.exec "snap info core" | grep -E "^tracking: +latest/${NEW_CHANNEL}"
         fi
 
-        if nested_is_core_18_system || nested_is_core_20_system || nested_is_core_22_system; then
+        if nested_is_core_ge 18; then
             remote.exec "sudo snap refresh snapd --${NEW_CHANNEL}"
             remote.exec "snap info snapd" | grep -E "^tracking: +latest/${NEW_CHANNEL}"
         else
             CHANGE_ID=$(remote.exec "sudo snap refresh core --${NEW_CHANNEL} --no-wait")
-            nested_wait_for_no_ssh
-            nested_wait_for_ssh
+            nested_wait_for_no_ssh 200 1
+            nested_wait_for_ssh 300 1
             # wait for the refresh to be done before checking, if we check too
             # quickly then operations on the core snap like reverting, etc. may
             # fail because it will have refresh-snap change in progress
@@ -334,8 +400,10 @@ nested_refresh_to_new_core() {
 
 nested_get_snakeoil_key() {
     local KEYNAME="PkKek-1-snakeoil"
-    wget -q https://raw.githubusercontent.com/snapcore/pc-amd64-gadget/20/snakeoil/$KEYNAME.key
-    wget -q https://raw.githubusercontent.com/snapcore/pc-amd64-gadget/20/snakeoil/$KEYNAME.pem
+    local VERSION
+    VERSION="$(nested_get_version)"
+    wget -q https://raw.githubusercontent.com/snapcore/pc-amd64-gadget/"$VERSION"/snakeoil/"$KEYNAME".key
+    wget -q https://raw.githubusercontent.com/snapcore/pc-amd64-gadget/"$VERSION"/snakeoil/"$KEYNAME".pem
     echo "$KEYNAME"
 }
 
@@ -358,9 +426,19 @@ nested_secboot_sign_gadget() {
     local GADGET_DIR="$1"
     local KEY="$2"
     local CERT="$3"
+    if [ -f "$GADGET_DIR/fb.efi" ]; then
+        nested_secboot_sign_file "$GADGET_DIR/fb.efi" "$KEY" "$CERT"
+    fi
     nested_secboot_sign_file "$GADGET_DIR/shim.efi.signed" "$KEY" "$CERT"
 }
 
+nested_secboot_sign_kernel() {
+    local KERNEL_DIR="$1"
+    local KEY="$2"
+    local CERT="$3"
+    nested_secboot_sign_file "$KERNEL_DIR/kernel.efi" "$KEY" "$CERT"
+}
+
 nested_prepare_env() {
     mkdir -p "$NESTED_IMAGES_DIR"
     mkdir -p "$NESTED_RUNTIME_DIR"
@@ -456,6 +534,8 @@ nested_get_version() {
         echo "20"
     elif nested_is_core_22_system; then
         echo "22"
+    elif nested_is_core_24_system; then
+        echo "24"
     fi
 }
 
@@ -483,6 +563,9 @@ nested_get_model() {
         ubuntu-22.04-arm-64)
             echo "$TESTSLIB/assertions/nested-22-arm64.model"
             ;;
+         ubuntu-24.04-64)
+            echo "$TESTSLIB/assertions/nested-24-amd64.model"
+            ;;
         *)
             echo "unsupported system"
             exit 1
@@ -512,23 +595,33 @@ nested_prepare_snapd() {
         echo "Repacking snapd snap"
         local snap_name output_name snap_id
         if nested_is_core_16_system; then
-            snap_name="core"
-            output_name="core-from-snapd-deb.snap"
-            snap_id="99T7MUlRhtI3U0QFgl5mXXESAiSwt776"
+            if [ ! -f "$NESTED_ASSETS_DIR/core-from-snapd-deb.snap" ]; then
+                "$TESTSTOOLS"/snaps-state repack_snapd_deb_into_snap core "$NESTED_ASSETS_DIR"
+                cp "$NESTED_ASSETS_DIR/core-from-snapd-deb.snap" "$(nested_get_extra_snaps_path)/core-from-snapd-deb.snap"
+            fi
+            # sign the snapd snap with fakestore if requested
+            if [ "$NESTED_SIGN_SNAPS_FAKESTORE" = "true" ]; then
+                "$TESTSTOOLS"/store-state make-snap-installable --noack "$NESTED_FAKESTORE_BLOB_DIR" "$(nested_get_extra_snaps_path)/core-from-snapd-deb.snap" "99T7MUlRhtI3U0QFgl5mXXESAiSwt776"
+            fi
         else
-            snap_name="snapd"
-            output_name="snapd-from-deb.snap"
-            snap_id="PMrrV4ml8uWuEUDBT8dSGnKUYbevVhc4"
-        fi
-
-        if [ ! -f "$NESTED_ASSETS_DIR/$output_name" ]; then
-            "$TESTSTOOLS"/snaps-state repack_snapd_deb_into_snap "$snap_name" "$NESTED_ASSETS_DIR"
-        fi
-        cp "$NESTED_ASSETS_DIR/$output_name" "$(nested_get_extra_snaps_path)/$output_name"
-
-        # sign the snapd snap with fakestore if requested
-        if [ "$NESTED_SIGN_SNAPS_FAKESTORE" = "true" ]; then
-            "$TESTSTOOLS"/store-state make-snap-installable --noack "$NESTED_FAKESTORE_BLOB_DIR" "$(nested_get_extra_snaps_path)/$output_name" "$snap_id"
+            for f in "${NESTED_ASSETS_DIR}"/snapd_*.snap; do
+                snap_name="$(basename "${f}")"
+                break
+            done
+            if [ ! -f "${NESTED_ASSETS_DIR}/${snap_name}" ]; then
+                # shellcheck source=tests/lib/prepare.sh
+                . "$TESTSLIB"/prepare.sh
+                build_snapd_snap "$NESTED_ASSETS_DIR"
+                for f in "${NESTED_ASSETS_DIR}"/snapd_*.snap; do
+                    snap_name="$(basename "${f}")"
+                    break
+                done
+                cp "${NESTED_ASSETS_DIR}/${snap_name}" "$(nested_get_extra_snaps_path)/"
+            fi
+            # sign the snapd snap with fakestore if requested
+            if [ "$NESTED_SIGN_SNAPS_FAKESTORE" = "true" ]; then
+                "$TESTSTOOLS"/store-state make-snap-installable --noack "$NESTED_FAKESTORE_BLOB_DIR" "$(nested_get_extra_snaps_path)/${snap_name}" "PMrrV4ml8uWuEUDBT8dSGnKUYbevVhc4"
+            fi
         fi
     fi
 }
@@ -543,12 +636,12 @@ nested_prepare_kernel() {
         version="$(nested_get_version)"
 
         if [ ! -f "$NESTED_ASSETS_DIR/$output_name" ]; then
-            if nested_is_core_16_system || nested_is_core_18_system; then
+            if nested_is_core_le 18; then
                 kernel_snap=pc-kernel-new.snap
                 repack_kernel_snap "$kernel_snap"
 
-            elif nested_is_core_20_system || nested_is_core_22_system; then
-                snap download --basename=pc-kernel --channel="$version/edge" pc-kernel
+            elif nested_is_core_ge 20; then
+                snap download --basename=pc-kernel --channel="$version/${NESTED_KERNEL_CHANNEL}" pc-kernel
 
                 # set the unix bump time if the NESTED_* var is set,
                 # otherwise leave it empty
@@ -558,7 +651,11 @@ nested_prepare_kernel() {
                     epochBumpTime="--epoch-bump-time=$epochBumpTime"
                 fi
 
-                uc20_build_initramfs_kernel_snap "pc-kernel.snap" "$NESTED_ASSETS_DIR" "$epochBumpTime"
+                if nested_is_core_24_system; then
+                    uc24_build_initramfs_kernel_snap "pc-kernel.snap" "$NESTED_ASSETS_DIR" "$epochBumpTime"
+                else
+                    uc20_build_initramfs_kernel_snap "pc-kernel.snap" "$NESTED_ASSETS_DIR" "$epochBumpTime"
+                fi
                 rm -f "pc-kernel.snap" "pc-kernel.assert"
 
                 # Prepare the pc kernel snap
@@ -578,7 +675,7 @@ nested_prepare_kernel() {
 
 nested_prepare_gadget() {
     if [ "$NESTED_REPACK_GADGET_SNAP" = "true" ]; then
-        if nested_is_core_20_system || nested_is_core_22_system; then
+        if nested_is_core_ge 20; then
             # Prepare the pc gadget snap (unless provided by extra-snaps)
             local snap_id version gadget_snap
             version="$(nested_get_version)"
@@ -601,7 +698,7 @@ nested_prepare_gadget() {
             snakeoil_key="$PWD/$key_name.key"
             snakeoil_cert="$PWD/$key_name.pem"
 
-            snap download --basename=pc --channel="$version/edge" pc
+            snap download --basename=pc --channel="$version/${NESTED_GADGET_CHANNEL}" pc
             unsquashfs -d pc-gadget pc.snap
             nested_secboot_sign_gadget pc-gadget "$snakeoil_key" "$snakeoil_cert"
             case "${NESTED_UBUNTU_SAVE:-}" in
@@ -642,7 +739,7 @@ EOF
                 echo "$GADGET_EXTRA_CMDLINE" > pc-gadget/cmdline.extra
             fi
 
-            # pack it
+            # pack the gadget
             snap pack pc-gadget/ "$NESTED_ASSETS_DIR"
 
             gadget_snap=$(ls "$NESTED_ASSETS_DIR"/pc_*.snap)
@@ -674,6 +771,9 @@ nested_prepare_base() {
         elif nested_is_core_22_system; then
             snap_name="core22"
             snap_id="amcUKQILKXHHTlmSa7NMdnXSx02dNeeT"
+        elif nested_is_core_24_system; then
+            snap_name="core24"
+            snap_id="dwTAh7MZZ01zyriOZErqd1JynQLiOGvM"
         fi
         output_name="${snap_name}.snap"
 
@@ -703,6 +803,36 @@ nested_prepare_base() {
     fi 
 }
 
+nested_prepare_essential_snaps() {
+    # shellcheck source=tests/lib/prepare.sh
+    . "$TESTSLIB"/prepare.sh
+    # shellcheck source=tests/lib/snaps.sh
+    . "$TESTSLIB"/snaps.sh
+
+    nested_prepare_snapd
+    nested_prepare_kernel
+    nested_prepare_gadget
+    nested_prepare_base
+}
+
+nested_configure_default_user() {
+    local IMAGE_NAME
+    IMAGE_NAME="$(nested_get_image_name core)"
+    # Configure the user for the vm
+    if [ "$NESTED_USE_CLOUD_INIT" = "true" ]; then
+        if nested_is_core_ge 20; then
+            nested_configure_cloud_init_on_core20_vm "$NESTED_IMAGES_DIR/$IMAGE_NAME"
+        else
+            nested_configure_cloud_init_on_core_vm "$NESTED_IMAGES_DIR/$IMAGE_NAME"
+        fi
+    else
+        nested_create_assertions_disk
+    fi
+
+    # Save a copy of the image
+    cp -v "$NESTED_IMAGES_DIR/$IMAGE_NAME" "$NESTED_IMAGES_DIR/$IMAGE_NAME.pristine"
+}
+
 nested_create_core_vm() {
     # shellcheck source=tests/lib/prepare.sh
     . "$TESTSLIB"/prepare.sh
@@ -769,13 +899,13 @@ nested_create_core_vm() {
             fi
             # ubuntu-image creates sparse image files
             # shellcheck disable=SC2086
-            "$UBUNTU_IMAGE" snap --image-size 10G --validation=enforce \
+            SNAPD_DEBUG=1 "$UBUNTU_IMAGE" snap --image-size 10G --validation=enforce \
                "$NESTED_MODEL" \
                 $UBUNTU_IMAGE_CHANNEL_ARG \
                 "${UBUNTU_IMAGE_PRESEED_ARGS[@]:-}" \
                 --output-dir "$NESTED_IMAGES_DIR" \
                 --sector-size "${NESTED_DISK_LOGICAL_BLOCK_SIZE}" \
-                $EXTRA_SNAPS
+                $EXTRA_SNAPS |& tee "$NESTED_LOGS_DIR/ubuntu-image.log"
 
             # ubuntu-image dropped the --output parameter, so we have to rename
             # the image ourselves, the images are named after volumes listed in
@@ -792,19 +922,7 @@ nested_create_core_vm() {
         fi
     fi
 
-    # Configure the user for the vm
-    if [ "$NESTED_USE_CLOUD_INIT" = "true" ]; then
-        if nested_is_core_20_system || nested_is_core_22_system; then
-            nested_configure_cloud_init_on_core20_vm "$NESTED_IMAGES_DIR/$IMAGE_NAME"
-        else
-            nested_configure_cloud_init_on_core_vm "$NESTED_IMAGES_DIR/$IMAGE_NAME"
-        fi
-    else
-        nested_create_assertions_disk
-    fi
-
-    # Save a copy of the image
-    cp -v "$NESTED_IMAGES_DIR/$IMAGE_NAME" "$NESTED_IMAGES_DIR/$IMAGE_NAME.pristine"
+    nested_configure_default_user
 }
 
 nested_configure_cloud_init_on_core_vm() {
@@ -871,7 +989,7 @@ nested_create_cloud_init_config() {
    list: |
     user1:ubuntu
    expire: False
-  datasource_list: [ NoCloud ]
+  datasource_list: [ NoCloud, None ]
   datasource:
     NoCloud:
      userdata_raw: |
@@ -884,7 +1002,7 @@ nested_create_cloud_init_uc20_config() {
     local CONFIG_PATH=$1
     cat << 'EOF' > "$CONFIG_PATH"
 #cloud-config
-datasource_list: [NoCloud]
+datasource_list: [ None ]
 users:
   - name: user1
     sudo: "ALL=(ALL) NOPASSWD:ALL"
@@ -988,14 +1106,14 @@ nested_start_core_vm_unit() {
     # the caller can override PARAM_MEM
     local PARAM_MEM PARAM_SMP
     if [ "$SPREAD_BACKEND" = "google-nested" ] || [ "$SPREAD_BACKEND" = "google-nested-arm" ]; then
-        PARAM_MEM="${NESTED_PARAM_MEM:--m 4096}"
-        PARAM_SMP="-smp 2"
+        PARAM_MEM="-m ${NESTED_MEM:-4096}"
+        PARAM_SMP="-smp ${NESTED_CPUS:-2}"
     elif [ "$SPREAD_BACKEND" = "google-nested-dev" ]; then
-        PARAM_MEM="${NESTED_PARAM_MEM:--m 8192}"
-        PARAM_SMP="-smp 4"
+        PARAM_MEM="-m ${NESTED_MEM:-8192}"
+        PARAM_SMP="-smp ${NESTED_CPUS:-4}"
     elif [ "$SPREAD_BACKEND" = "qemu-nested" ]; then
-        PARAM_MEM="${NESTED_PARAM_MEM:--m 2048}"
-        PARAM_SMP="-smp 1"
+        PARAM_MEM="-m ${NESTED_MEM:-2048}"
+        PARAM_SMP="-smp ${NESTED_CPUS:-1}"
     else
         echo "unknown spread backend $SPREAD_BACKEND"
         exit 1
@@ -1081,7 +1199,7 @@ nested_start_core_vm_unit() {
         # storage to
         PARAM_ASSERTIONS="-drive if=none,id=stick,format=raw,file=$NESTED_ASSETS_DIR/assertions.disk,cache=none,format=raw -device nec-usb-xhci,id=xhci -device usb-storage,bus=xhci.0,removable=true,drive=stick"
     fi
-    if nested_is_core_20_system || nested_is_core_22_system; then
+    if nested_is_core_ge 20; then
         # use a bundle EFI bios by default
         if os.query is-arm; then
             PARAM_BIOS="-bios /usr/share/AAVMF/AAVMF_CODE.fd"
@@ -1089,35 +1207,34 @@ nested_start_core_vm_unit() {
             PARAM_BIOS="-bios /usr/share/ovmf/OVMF.fd"
         fi
         local OVMF_CODE OVMF_VARS
-        OVMF_CODE="secboot"
-        OVMF_VARS="ms"
+        OVMF_CODE=""
+        OVMF_VARS=""
 
-        if nested_is_core_22_system; then
+        if nested_is_core_ge 22; then
             wget -q https://storage.googleapis.com/snapd-spread-tests/dependencies/OVMF_CODE.secboot.fd
             mv OVMF_CODE.secboot.fd /usr/share/OVMF/OVMF_CODE.secboot.fd
             wget -q https://storage.googleapis.com/snapd-spread-tests/dependencies/OVMF_VARS.snakeoil.fd
             mv OVMF_VARS.snakeoil.fd /usr/share/OVMF/OVMF_VARS.snakeoil.fd
-        fi
-        # In this case the kernel.efi is unsigned and signed with snaleoil certs
-        if [ "$NESTED_BUILD_SNAPD_FROM_CURRENT" = "true" ]; then
-            OVMF_VARS="snakeoil"
+            wget -q https://storage.googleapis.com/snapd-spread-tests/dependencies/OVMF_VARS.ms.fd
+            mv OVMF_VARS.ms.fd /usr/share/OVMF/OVMF_VARS.ms.fd
         fi
 
-        if [ "${NESTED_ENABLE_OVMF:-}" = "true" ]; then
-            if os.query is-arm; then
-                PARAM_BIOS="-bios /usr/share/AAVMF/AAVMF_CODE.fd"
-            else
-                PARAM_BIOS="-bios /usr/share/OVMF/OVMF_CODE.fd"
-            fi
+        # In this case the kernel.efi is unsigned and signed with snaleoil certs
+        if [ "$NESTED_FORCE_MS_KEYS" != "true" ] && [ "$NESTED_BUILD_SNAPD_FROM_CURRENT" = "true" ]; then
+            OVMF_VARS="snakeoil"
+        else
+            OVMF_VARS="ms"
         fi
+
         if nested_is_secure_boot_enabled; then
+            OVMF_CODE="secboot"
             if os.query is-arm; then
                 cp -f "/usr/share/AAVMF/AAVMF_VARS.fd" "$NESTED_ASSETS_DIR/AAVMF_VARS.fd"
                 PARAM_BIOS="-drive file=/usr/share/AAVMF/AAVMF_CODE.fd,if=pflash,format=raw,unit=0,readonly=on -drive file=$NESTED_ASSETS_DIR/AAVMF_VARS.fd,if=pflash,format=raw"
             else
-                cp -f "/usr/share/OVMF/OVMF_VARS.$OVMF_VARS.fd" "$NESTED_ASSETS_DIR/OVMF_VARS.$OVMF_VARS.fd"
-                PARAM_BIOS="-drive file=/usr/share/OVMF/OVMF_CODE.$OVMF_CODE.fd,if=pflash,format=raw,unit=0,readonly=on -drive file=$NESTED_ASSETS_DIR/OVMF_VARS.$OVMF_VARS.fd,if=pflash,format=raw"
-                PARAM_MACHINE="-machine q35${ATTR_KVM} -global ICH9-LPC.disable_s3=1"            
+                cp -f "/usr/share/OVMF/OVMF_VARS.${OVMF_VARS}.fd" "$NESTED_ASSETS_DIR/OVMF_VARS.${OVMF_VARS}.fd"
+                PARAM_BIOS="-drive file=/usr/share/OVMF/OVMF_CODE.${OVMF_CODE}.fd,if=pflash,format=raw,unit=0,readonly=on -drive file=$NESTED_ASSETS_DIR/OVMF_VARS.${OVMF_VARS}.fd,if=pflash,format=raw"
+                PARAM_MACHINE="-machine q35${ATTR_KVM} -global ICH9-LPC.disable_s3=1"
             fi
         fi
 
@@ -1126,7 +1243,7 @@ nested_start_core_vm_unit() {
                 if [ -z "$NESTED_TPM_NO_RESTART" ]; then
                     # reset the tpm state
                     snap stop test-snapd-swtpm > /dev/null
-                    rm /var/snap/test-snapd-swtpm/current/tpm2-00.permall
+                    rm /var/snap/test-snapd-swtpm/current/tpm2-00.permall || true
                     snap start test-snapd-swtpm > /dev/null
                 fi
             else
@@ -1185,10 +1302,13 @@ nested_start_core_vm_unit() {
     EXPECT_SHUTDOWN=${NESTED_EXPECT_SHUTDOWN:-}
 
     if [ "$EXPECT_SHUTDOWN" != "1" ]; then
-        # Wait until ssh is ready
-        nested_wait_for_ssh
+        # Wait until the vm is ready to receive connections
+        if ! nested_wait_vm_ready 120; then
+            echo "failed to wait for the vm becomes ready to receive connections"
+            return 1
+        fi
         # Wait for the snap command to be available
-        nested_wait_for_snap_command
+        nested_wait_for_snap_command 120 1
         # Wait for snap seeding to be done
         # retry this wait command up to 3 times since we sometimes see races 
         # where the snap command appears, then immediately disappears and then 
@@ -1206,7 +1326,15 @@ nested_start_core_vm_unit() {
         nested_prepare_tools
         # Wait for cloud init to be done if the system is using cloud-init
         if [ "$NESTED_USE_CLOUD_INIT" = true ]; then
-            remote.exec "retry --wait 1 -n 5 sh -c 'cloud-init status --wait'"
+            if ! remote.exec "retry --wait 1 -n 5 sh -c 'cloud-init status --wait'"; then
+                # Error 2 means 'recoverable error', ignore that case
+                ret=0
+                remote.exec "cloud-init status" || ret=$?
+                if [ "$ret" -ne 0 ] && [ "$ret" -ne 2 ]; then
+                    echo "cloud-init finished with error $ret"
+                    exit 1
+                fi
+            fi
         fi
     fi
 }
@@ -1274,7 +1402,7 @@ nested_shutdown() {
     # least can't hurt anything
     remote.exec "sync"
     remote.exec "sudo shutdown now" || true
-    nested_wait_for_no_ssh
+    nested_wait_for_no_ssh 120 1
     nested_force_stop_vm
     tests.systemd wait-for-service -n 30 --wait 1 --state inactive "$NESTED_VM"
     sync
@@ -1284,7 +1412,7 @@ nested_start() {
     nested_save_serial_log
     nested_force_start_vm
     tests.systemd wait-for-service -n 30 --wait 1 --state active "$NESTED_VM"
-    nested_wait_for_ssh
+    nested_wait_for_ssh 300 1
     nested_prepare_tools
 }
 
@@ -1334,13 +1462,14 @@ nested_start_classic_vm() {
     PARAM_SMP="-smp 1"
     # use only 2G of RAM for qemu-nested
     if [ "$SPREAD_BACKEND" = "google-nested" ]; then
-        PARAM_MEM="${NESTED_PARAM_MEM:--m 4096}"
-        PARAM_SMP="-smp 2"
+        PARAM_MEM="-m ${NESTED_MEM:-4096}"
+        PARAM_SMP="-smp ${NESTED_CPUS:-2}"
     elif [ "$SPREAD_BACKEND" = "google-nested-dev" ]; then
-        PARAM_MEM="${NESTED_PARAM_MEM:--m 8192}"
-        PARAM_SMP="-smp 4"
+        PARAM_MEM="-m ${NESTED_MEM:-8192}"
+        PARAM_SMP="-smp ${NESTED_CPUS:-4}"
     elif [ "$SPREAD_BACKEND" = "qemu-nested" ]; then
-        PARAM_MEM="${NESTED_PARAM_MEM:--m 2048}"
+        PARAM_MEM="-m ${NESTED_MEM:-2048}"
+        PARAM_SMP="-smp ${NESTED_CPUS:-1}"
     else
         echo "unknown spread backend $SPREAD_BACKEND"
         exit 1
@@ -1387,7 +1516,7 @@ nested_start_classic_vm() {
         exit 1
     fi
 
-    PARAM_IMAGE="-drive file=$NESTED_IMAGES_DIR/$IMAGE_NAME,if=virtio"
+    PARAM_IMAGE="-drive file=$NESTED_IMAGES_DIR/$IMAGE_NAME,if=none,id=disk1 -device virtio-blk-pci,drive=disk1,bootindex=1"
     PARAM_SEED="-drive file=$NESTED_ASSETS_DIR/seed.img,if=virtio"
     # Open port 7777 on the host so that failures in the nested VM (e.g. to
     # create users) can be debugged interactively via
@@ -1430,7 +1559,10 @@ nested_start_classic_vm() {
         ${PARAM_EXTRA} \
         ${PARAM_CD} "
 
-    nested_wait_for_ssh
+    if ! nested_wait_vm_ready 60; then
+        echo "failed to wait for the vm becomes ready to receive connections"
+        return 1
+    fi
 
     # Copy tools to be used on tests
     nested_prepare_tools
@@ -1576,3 +1708,25 @@ nested_wait_for_device_initialized_change() {
         sleep "$wait"
     done
 }
+
+nested_check_spread_results() {
+    SPREAD_LOG=$1
+    if [ -z "$SPREAD_LOG" ]; then
+        return 1
+    fi
+
+    if grep -eq "Successful tasks:" "$SPREAD_LOG"; then
+        if grep -E "Failed (task|suite|project)" "$SPREAD_LOG"; then
+            return 1
+        fi
+        if ! grep -eq "Aborted tasks: 0" "$SPREAD_LOG"; then
+            return 1
+        fi
+
+        if [ "$EXIT_STATUS" = "0" ]; then
+            return 0
+        fi    
+    else
+        return 1
+    fi
+}
