@@ -9,6 +9,7 @@ set -eux
 # shellcheck source=tests/lib/state.sh
 . "$TESTSLIB/state.sh"
 
+: "${WORK_DIR:=/tmp/work-dir}"
 
 disable_kernel_rate_limiting() {
     # kernel rate limiting hinders debugging security policy so turn it off
@@ -44,6 +45,37 @@ EOF
     systemctl daemon-reload
 }
 
+# Set of helpers for checking if the test system is expected to be
+# Ubuntu Core. The helpers can be used both in a UC system or in
+# a classic system which will be transformed into UC. Note, the 
+# helpers assume a specific formatting of SPREAD_SYSTEM environment
+# variable which follows this pattern: ubuntu-core-<VERSION>[-ARCH]-<BITS>
+# where arch is "" for amd64, arm for armhf and arm64, etc
+is_test_target_core() {
+    local VERSION=${1:-}
+    [[ "$SPREAD_SYSTEM" = ubuntu-core-${VERSION}* ]]
+}
+
+is_test_target_core_ge() {
+    local VERSION=${1:-}
+    if [ -z "$VERSION" ]; then
+        echo "version id is expected"
+        exit 1
+    fi
+    CURR_VERSION="$(cut -d- -f3 <<< "$SPREAD_SYSTEM")"
+    [ "$CURR_VERSION" -ge "${VERSION}" ]
+}
+
+is_test_target_core_le() {
+    local VERSION=${1:-}
+    if [ -z "$VERSION" ]; then
+        echo "version id is expected"
+        exit 1
+    fi
+    CURR_VERSION="$(cut -d- -f3 <<< "$SPREAD_SYSTEM")"
+    [ "$CURR_VERSION" -le "${VERSION}" ]
+}
+
 ensure_jq() {
     if command -v jq; then
         return
@@ -58,6 +90,9 @@ ensure_jq() {
     elif os.query is-core22; then
         snap install --devmode --edge jq-core22
         snap alias jq-core22.jq jq
+    elif os.query is-core24; then
+        snap install --devmode --edge test-snapd-jq-core24
+        snap alias test-snapd-jq-core24.jq jq
     else
         snap install --devmode jq
     fi
@@ -81,6 +116,7 @@ disable_refreshes() {
     snap remove --purge jq-core18
     snap remove --purge jq-core20
     snap remove --purge jq-core22
+    snap remove --purge test-snapd-jq-core24
 }
 
 setup_systemd_snapd_overrides() {
@@ -89,6 +125,10 @@ setup_systemd_snapd_overrides() {
 [Service]
 Environment=SNAPD_DEBUG_HTTP=7 SNAPD_DEBUG=1 SNAPPY_TESTING=1 SNAPD_REBOOT_DELAY=10m SNAPD_CONFIGURE_HOOK_TIMEOUT=30s SNAPPY_USE_STAGING_STORE=$SNAPPY_USE_STAGING_STORE
 ExecStartPre=/bin/touch /dev/iio:device0
+# The default limit is usually 5, which can be easily hit in 
+# a fast system with few systemd units
+StartLimitBurst=10
+StartLimitIntervalSec=10s
 EOF
 
     # We change the service configuration so reload and restart
@@ -131,7 +171,7 @@ update_core_snap_for_classic_reexec() {
 
     # First of all, unmount the core
     core="$(readlink -f "$SNAP_MOUNT_DIR/core/current" || readlink -f "$SNAP_MOUNT_DIR/ubuntu-core/current")"
-    snap="$(mount | grep " $core" | awk '{print $1}')"
+    snap="$(mount | grep " $core" | head -n 1 | awk '{print $1}')"
     umount --verbose "$core"
 
     # Now unpack the core, inject the new snap-exec/snapctl into it
@@ -211,6 +251,14 @@ update_core_snap_for_classic_reexec() {
 }
 
 prepare_memory_limit_override() {
+    # First time it is needed to save the initial env var value
+    if not tests.env is-set initial SNAPD_NO_MEMORY_LIMIT; then
+        tests.env set initial SNAPD_NO_MEMORY_LIMIT "$SNAPD_NO_MEMORY_LIMIT"
+    # Then if the new value is the same than the initial, then no new configuration needed
+    elif [ "$(tests.env get initial SNAPD_NO_MEMORY_LIMIT)" = "$SNAPD_NO_MEMORY_LIMIT" ]; then
+        return
+    fi
+
     # set up memory limits for snapd bu default unless explicit requested not to
     # or the system is known to be problematic
     local set_limit=1
@@ -228,7 +276,7 @@ prepare_memory_limit_override() {
             set_limit=0
             ;;
         *)
-            if [ -n "${SNAPD_NO_MEMORY_LIMIT:-}" ]; then
+            if [ "$SNAPD_NO_MEMORY_LIMIT" = 1 ]; then
                 set_limit=0
             fi
             ;;
@@ -250,8 +298,7 @@ prepare_memory_limit_override() {
         # systemd is backwards compatible so the limit is still set.
         cat <<EOF > /etc/systemd/system/snapd.service.d/memory-max.conf
 [Service]
-# mvo: disabled because of many failures in restore, e.g. in PR#11014
-#MemoryLimit=100M
+MemoryLimit=200M
 EOF
     fi
     # the service setting may have changed in the service so we need
@@ -260,25 +307,41 @@ EOF
     systemctl restart snapd
 }
 
-prepare_each_classic() {
+prepare_reexec_override() {
+    local reexec_file=/etc/systemd/system/snapd.service.d/reexec.conf
+ 
+    # First time it is needed to save the initial env var value
+    if not tests.env is-set initial SNAP_REEXEC; then
+        tests.env set initial SNAP_REEXEC "$SNAP_REEXEC"
+    # Then if the new value is the same than the initial, then no new configuration needed
+    elif [ "$(tests.env get initial SNAP_REEXEC)" = "$SNAP_REEXEC" ]; then
+        return
+    fi
+
+    # Just update reexec configuration when the SNAP_REEXEC var has been updated
+    # Otherwise it is used the configuration set during project preparation
     mkdir -p /etc/systemd/system/snapd.service.d
     if [ -z "${SNAP_REEXEC:-}" ]; then
-        rm -f /etc/systemd/system/snapd.service.d/reexec.conf
+        rm -f "$reexec_file"
     else
-        cat <<EOF > /etc/systemd/system/snapd.service.d/reexec.conf
+        cat <<EOF > "$reexec_file"
 [Service]
 Environment=SNAP_REEXEC=$SNAP_REEXEC
 EOF
     fi
-    # the re-exec setting may have changed in the service so we need
-    # to ensure snapd is reloaded
+
     systemctl daemon-reload
     systemctl restart snapd
+}
 
+prepare_each_classic() {
     if [ ! -f /etc/systemd/system/snapd.service.d/local.conf ]; then
         echo "/etc/systemd/system/snapd.service.d/local.conf vanished!"
         exit 1
     fi
+
+    prepare_reexec_override
+    prepare_memory_limit_override
 }
 
 prepare_classic() {
@@ -326,17 +389,18 @@ prepare_classic() {
     setup_systemd_snapd_overrides
 
     if [ "$REMOTE_STORE" = staging ]; then
-        # shellcheck source=tests/lib/store.sh
-        . "$TESTSLIB/store.sh"
         # reset seeding data that is likely tainted with production keys
         systemctl stop snapd.service snapd.socket
         rm -rf /var/lib/snapd/assertions/*
         rm -f /var/lib/snapd/state.json
-        setup_staging_store
+        "$TESTSTOOLS"/store-state setup-staging-store
     fi
 
     # Snapshot the state including core.
     if ! is_snapd_state_saved; then
+        # Create the file with the initial environment before saving the state
+        tests.env start initial
+
         # need to be seeded to proceed with snap install
         # also make sure the captured state is seeded
         snap wait system seed.loaded
@@ -363,6 +427,8 @@ prepare_classic() {
         update_core_snap_for_classic_reexec
         systemctl start snapd.{service,socket}
 
+        prepare_reexec_override
+        prepare_memory_limit_override
         disable_refreshes
 
         # Check bootloader environment output in architectures different to s390x which uses zIPL
@@ -378,6 +444,7 @@ prepare_classic() {
         fi
 
         setup_experimental_features
+
         systemctl stop snapd.{service,socket}
         save_snapd_state
         systemctl start snapd.socket
@@ -394,90 +461,89 @@ prepare_classic() {
     fi
 }
 
-repack_snapd_snap_with_deb_content() {
-    local TARGET="$1"
-
-    local UNPACK_DIR="/tmp/snapd-unpack"
-    unsquashfs -no-progress -d "$UNPACK_DIR" snapd_*.snap
-    # clean snap apparmor.d to ensure we put the right snap-confine apparmor
-    # file in place. Its called usr.lib.snapd.snap-confine on 14.04 but
-    # usr.lib.snapd.snap-confine.real everywhere else
-    rm -f "$UNPACK_DIR"/etc/apparmor.d/*
-
-    dpkg-deb -x "$SPREAD_PATH"/../snapd_*.deb "$UNPACK_DIR"
-    cp /usr/lib/snapd/info "$UNPACK_DIR"/usr/lib/snapd
-    snap pack "$UNPACK_DIR" "$TARGET"
-    rm -rf "$UNPACK_DIR"
+ensure_snapcraft() {
+    if ! command -v snapcraft; then
+        snap install --channel="${SNAPCRAFT_SNAP_CHANNEL}" snapcraft --classic
+        "$TESTSTOOLS"/lxd-state prepare-snap
+    fi
 }
 
-repack_core_snap_with_tweaks() {
-    local CORESNAP="$1"
-    local TARGET="$2"
-
-    local UNPACK_DIR="/tmp/core-unpack"
-    unsquashfs -no-progress -d "$UNPACK_DIR" "$CORESNAP"
-
-    mkdir -p "$UNPACK_DIR"/etc/systemd/journald.conf.d
-    cat <<EOF > "$UNPACK_DIR"/etc/systemd/journald.conf.d/to-console.conf
-[Journal]
-ForwardToConsole=yes
-TTYPath=/dev/ttyS0
-MaxLevelConsole=debug
-EOF
-    mkdir -p "$UNPACK_DIR"/etc/systemd/system/snapd.service.d
-cat <<EOF > "$UNPACK_DIR"/etc/systemd/system/snapd.service.d/logging.conf
-[Service]
-Environment=SNAPD_DEBUG_HTTP=7 SNAPD_DEBUG=1 SNAPPY_TESTING=1 SNAPD_CONFIGURE_HOOK_TIMEOUT=30s
-StandardOutput=journal+console
-StandardError=journal+console
-EOF
-
-    cp "${SPREAD_PATH}"/data/completion/bash/complete.sh "${UNPACK_DIR}"/usr/lib/snapd/complete.sh
-
-    snap pack --filename="$TARGET" "$UNPACK_DIR"
-
-    rm -rf "$UNPACK_DIR"
+cleanup_snapcraft() {
+    snap remove --purge lxd || true
+    "$TESTSTOOLS"/lxd-state undo-mount-changes
+    snap remove --purge snapcraft || true
 }
 
-repack_kernel_snap() {
-    local TARGET=$1
-    local VERSION
-    local UNPACK_DIR
-    local CHANNEL
+run_snapcraft() {
+    ensure_snapcraft
+    # maybe read the log path from snapcraft output as an improvement
+    # but this works
+    if ! (cd "${PROJECT_PATH}" && snapcraft "$@"); then
+        # shellcheck disable=SC2012
+        tail -n1000 "${HOME}/.local/state/snapcraft/log/$(ls -t "${HOME}/.local/state/snapcraft/log/" -1 | head -n1)"
+        false
+    fi
+    cleanup_snapcraft
+}
 
-    VERSION=$(nested_get_version)
-    if [ "$VERSION" = 16 ]; then
-        CHANNEL=latest
+build_snapd_snap() {
+    local TARGET
+    local snapd_snap_cache
+    TARGET="${1}"
+
+    snapd_snap_cache="$WORK_DIR/snapd_snap"
+    mkdir -p "${snapd_snap_cache}"
+    for snap in "${snapd_snap_cache}"/snapd_*.snap; do
+        if ! [ -f "${snap}" ]; then
+            if [ "${USE_PREBUILT_SNAPD_SNAP}" = true ]; then
+                if [ -n "${USE_SNAPD_SNAP_URL}" ]; then
+                    wget -q "$USE_SNAPD_SNAP_URL" -O "${snapd_snap_cache}/snapd_from_ci.snap"
+                else
+                    cp "${PROJECT_PATH}/built-snap"/snapd_1337.*.snap.keep "${snapd_snap_cache}/snapd_from_ci.snap"
+                fi
+            else
+                [ -d "${TARGET}" ] || mkdir -p "${TARGET}"
+                chmod -R go+r "${PROJECT_PATH}/tests"
+                run_snapcraft --use-lxd --verbosity quiet --output="snapd_from_snapcraft.snap"
+                mv "${PROJECT_PATH}"/snapd_from_snapcraft.snap "${snapd_snap_cache}"
+            fi
+        fi
+        break
+    done
+
+    cp "${snapd_snap_cache}"/snapd_*.snap "${TARGET}/"
+}
+
+build_snapd_snap_with_run_mode_firstboot_tweaks() {
+    local snapd_snap_cache
+    local TARGET
+
+    TARGET="${1}"
+
+    snapd_snap_cache="$WORK_DIR/snapd_snap_with_tweaks"
+    mkdir -p "${snapd_snap_cache}"
+    for snap in "${snapd_snap_cache}"/snapd_*.snap; do
+        if [ -f "${snap}" ]; then
+            cp "${snap}" "${TARGET}/"
+            return
+        fi
+    done
+
+    if [ "${USE_PREBUILT_SNAPD_SNAP}" = true ]; then
+        if [ -n "${USE_SNAPD_SNAP_URL}" ]; then
+            wget -q "$USE_SNAPD_SNAP_URL" -O /tmp/snapd_from_snapcraft.snap
+        else
+            cp "${PROJECT_PATH}/built-snap"/snapd_1337.*.snap.keep "/tmp/snapd_from_snapcraft.snap"
+        fi
     else
-        CHANNEL=$VERSION
+        chmod -R go+r "${PROJECT_PATH}/tests"
+        run_snapcraft --use-lxd --verbosity quiet --output="snapd_from_snapcraft.snap"
+        mv "${PROJECT_PATH}/snapd_from_snapcraft.snap" "/tmp/snapd_from_snapcraft.snap"
     fi
 
-    echo "Repacking kernel snap"
-    UNPACK_DIR=/tmp/kernel-unpack
-    snap download --basename=pc-kernel --channel="$CHANNEL/edge" pc-kernel
-    unsquashfs -no-progress -d "$UNPACK_DIR" pc-kernel.snap
-    snap pack --filename="$TARGET" "$UNPACK_DIR"
-
-    rm -rf pc-kernel.snap "$UNPACK_DIR"
-}
-
-repack_snapd_snap_with_deb_content_and_run_mode_firstboot_tweaks() {
-    local TARGET="$1"
-
     local UNPACK_DIR="/tmp/snapd-unpack"
-    unsquashfs -no-progress -d "$UNPACK_DIR" snapd_*.snap
-
-    # data/preseed.json is not included in the deb, use the latest
-    # version from source tree to replace the one in the re-packed snapd snap.
-    cp "$PROJECT_PATH/data/preseed.json" "$UNPACK_DIR"/usr/lib/snapd
-
-    # clean snap apparmor.d to ensure we put the right snap-confine apparmor
-    # file in place. Its called usr.lib.snapd.snap-confine on 14.04 but
-    # usr.lib.snapd.snap-confine.real everywhere else
-    rm -f "$UNPACK_DIR"/etc/apparmor.d/*
-
-    dpkg-deb -x "$SPREAD_PATH"/../snapd_*.deb "$UNPACK_DIR"
-    cp /usr/lib/snapd/info "$UNPACK_DIR"/usr/lib/snapd
+    rm -rf "${UNPACK_DIR}"
+    unsquashfs -no-progress -d "$UNPACK_DIR" /tmp/snapd_from_snapcraft.snap
 
     # now install a unit that sets up enough so that we can connect
     cat > "$UNPACK_DIR"/lib/systemd/system/snapd.spread-tests-run-mode-tweaks.service <<'EOF'
@@ -547,12 +613,63 @@ touch /root/spread-setup-done
 EOF
     chmod 0755 "$UNPACK_DIR"/usr/lib/snapd/snapd.spread-tests-run-mode-tweaks.sh
 
+    mkdir -p "${snapd_snap_cache}"
+    snap pack "$UNPACK_DIR" "${snapd_snap_cache}/"
+    rm -rf "$UNPACK_DIR"
+    cp "${snapd_snap_cache}"/snapd_*.snap "${TARGET}/"
+}
+
+repack_core_snap_with_tweaks() {
+    local CORESNAP="$1"
+    local TARGET="$2"
+
+    local UNPACK_DIR="/tmp/core-unpack"
+    unsquashfs -no-progress -d "$UNPACK_DIR" "$CORESNAP"
+
+    mkdir -p "$UNPACK_DIR"/etc/systemd/journald.conf.d
+    cat <<EOF > "$UNPACK_DIR"/etc/systemd/journald.conf.d/to-console.conf
+[Journal]
+ForwardToConsole=yes
+TTYPath=/dev/ttyS0
+MaxLevelConsole=debug
+EOF
+    mkdir -p "$UNPACK_DIR"/etc/systemd/system/snapd.service.d
+cat <<EOF > "$UNPACK_DIR"/etc/systemd/system/snapd.service.d/logging.conf
+[Service]
+Environment=SNAPD_DEBUG_HTTP=7 SNAPD_DEBUG=1 SNAPPY_TESTING=1 SNAPD_CONFIGURE_HOOK_TIMEOUT=30s
+StandardOutput=journal+console
+StandardError=journal+console
+EOF
+
     cp "${SPREAD_PATH}"/data/completion/bash/complete.sh "${UNPACK_DIR}"/usr/lib/snapd/complete.sh
 
-    snap pack "$UNPACK_DIR" "$TARGET"
+    snap pack --filename="$TARGET" "$UNPACK_DIR"
+
     rm -rf "$UNPACK_DIR"
 }
 
+repack_kernel_snap() {
+    local TARGET=$1
+    local VERSION
+    local UNPACK_DIR
+    local CHANNEL
+
+    VERSION=$(nested_get_version)
+    if [ "$VERSION" = 16 ]; then
+        CHANNEL=latest
+    else
+        CHANNEL=$VERSION
+    fi
+
+    echo "Repacking kernel snap"
+    UNPACK_DIR=/tmp/kernel-unpack
+    snap download --basename=pc-kernel --channel="$CHANNEL/${KERNEL_CHANNEL}" pc-kernel
+    unsquashfs -no-progress -d "$UNPACK_DIR" pc-kernel.snap
+    snap pack --filename="$TARGET" "$UNPACK_DIR"
+
+    rm -rf pc-kernel.snap "$UNPACK_DIR"
+}
+
 # Builds kernel snap with bad kernel.efi, in different ways
 # $1: snap we will modify
 # $2: target folder for the new snap
@@ -591,8 +708,9 @@ uc20_build_corrupt_kernel_snap() {
 }
 
 uc20_build_initramfs_kernel_snap() {
+    quiet apt install software-properties-common -y
     # carries ubuntu-core-initframfs
-    add-apt-repository ppa:snappy-dev/image -y
+    quiet add-apt-repository ppa:snappy-dev/image -y
     # On focal, lvm2 does not reinstall properly after being removed.
     # So we need to clean up in case the VM has been re-used.
     if os.query is-focal; then
@@ -601,7 +719,7 @@ uc20_build_initramfs_kernel_snap() {
     # TODO: install the linux-firmware as the current version of
     # ubuntu-core-initramfs does not depend on it, but nonetheless requires it
     # to build the initrd
-    apt install ubuntu-core-initramfs linux-firmware -y
+    quiet apt install ubuntu-core-initramfs linux-firmware -y
 
     local ORIG_SNAP="$1"
     local TARGET="$2"
@@ -649,9 +767,15 @@ uc20_build_initramfs_kernel_snap() {
         # kernel and we don't want to test that, just test our snap-bootstrap
         cp -ar unpacked-initrd skeleton
         # all the skeleton edits go to a local copy of distro directory
-        skeletondir=$PWD/skeleton
-        cp -a /usr/lib/snapd/snap-bootstrap "$skeletondir/main/usr/lib/snapd/snap-bootstrap.real"
-        cat <<'EOF' | sed -E "s/^ {8}//" >"$skeletondir/main/usr/lib/snapd/snap-bootstrap"
+        skeletondir="$PWD/skeleton"
+        snap_bootstrap_file="$skeletondir/main/usr/lib/snapd/snap-bootstrap"
+        clock_epoch_file="$skeletondir/main/usr/lib/clock-epoch"
+        if os.query is-arm; then
+            snap_bootstrap_file="$skeletondir/usr/lib/snapd/snap-bootstrap"
+            clock_epoch_file="$skeletondir/usr/lib/clock-epoch"
+        fi
+        cp -a /usr/lib/snapd/snap-bootstrap "${snap_bootstrap_file}.real"
+        cat <<'EOF' | sed -E "s/^ {8}//" >"$snap_bootstrap_file"
         #!/bin/sh
         set -eux
         if [ "$1" != initramfs-mounts ]; then
@@ -673,20 +797,24 @@ uc20_build_initramfs_kernel_snap() {
         date --utc '+%s' > /run/mnt/ubuntu-seed/test/${mode}-after-snap-bootstrap-date
 EOF
 
-        chmod +x "$skeletondir/main/usr/lib/snapd/snap-bootstrap"
+        chmod +x "$snap_bootstrap_file"
 
         if [ "$injectKernelPanic" = "true" ]; then
             # add a kernel panic to the end of the-tool execution
-            echo "echo 'forcibly panicing'; echo c > /proc/sysrq-trigger" >> "$skeletondir/main/usr/lib/snapd/snap-bootstrap"
+            echo "echo 'forcibly panicking'; echo c > /proc/sysrq-trigger" >> "$snap_bootstrap_file"
         fi
 
         # bump the epoch time file timestamp, converting unix timestamp to 
         # touch's date format
-        touch -t "$(date --utc "--date=@$initramfsEpochBumpTime" '+%Y%m%d%H%M')" "$skeletondir/main/usr/lib/clock-epoch"
+        touch -t "$(date --utc "--date=@$initramfsEpochBumpTime" '+%Y%m%d%H%M')" "$clock_epoch_file"
 
         # copy any extra files to the same location inside the initrd
         if [ -d ../extra-initrd/ ]; then
-            cp -a ../extra-initrd/* "$skeletondir"/main
+            if os.query is-arm; then
+                cp -a ../extra-initrd/* "$skeletondir"
+            else
+                cp -a ../extra-initrd/* "$skeletondir"/main
+            fi
         fi
 
         # XXX: need to be careful to build an initrd using the right kernel
@@ -695,7 +823,13 @@ EOF
         (
             # accommodate assumptions about tree layout, use the unpacked initrd
             # to pick up the right modules
-            cd unpacked-initrd/main
+            if os.query is-arm; then
+                cd unpacked-initrd
+                feature='.'
+            else
+                cd unpacked-initrd/main
+                feature='main'
+            fi
             # XXX: pass feature 'main' and u-c-i picks up any directory named
             # after feature inside skeletondir and uses that a template
             ubuntu-core-initramfs create-initrd \
@@ -703,8 +837,8 @@ EOF
                                   --skeleton "$skeletondir" \
                                   --kerneldir "${unpackeddir}/modules/$kver" \
                                   --firmwaredir "${unpackeddir}/firmware" \
-                                  --feature 'main' \
-                                  --output ../../repacked-initrd
+                                  --feature "$feature" \
+                                  --output "$unpackeddir"/repacked-initrd
         )
 
         # copy out the kernel image for create-efi command
@@ -725,39 +859,8 @@ EOF
         rm -rf unpacked-initrd skeleton initrd repacked-initrd-* vmlinuz-*
     )
 
-    (
-        # XXX: drop ~450MB+ of firmware which should not be needed in under qemu
-        # or the cloud system
-        cd repacked-kernel
-        rm -rf firmware/*
-
-        # the code below drops the modules that are not loaded on the
-        # current host, this should work for most cases, since the image will be
-        # running on the same host
-        # TODO:UC20: enable when ready
-        exit 0
-
-        # drop unnecessary modules
-        awk '{print $1}' <  /proc/modules  | sort > /tmp/mods
-        #shellcheck disable=SC2044
-        for m in $(find modules/ -name '*.ko'); do
-            noko=$(basename "$m"); noko="${noko%.ko}"
-            if echo "$noko" | grep -f /tmp/mods -q ; then
-                echo "keeping $m - $noko"
-            else
-                rm -f "$m"
-            fi
-        done
-
-        #shellcheck disable=SC2010
-        kver=$(ls "config"-* | grep -Po 'config-\K.*')
-
-        # depmod assumes that /lib/modules/$kver is under basepath
-        mkdir -p fake/lib
-        ln -s "$PWD/modules" fake/lib/modules
-        depmod -b "$PWD/fake" -A -v "$kver"
-        rm -rf fake
-    )
+    # drop ~450MB+ of firmware which should not be needed in qemu or the cloud system
+    rm -rf repacked-kernel/firmware/*
 
     # copy any extra files that tests may need for the kernel
     if [ -d ./extra-kernel-snap/ ]; then
@@ -768,6 +871,77 @@ EOF
     rm -rf repacked-kernel
 }
 
+uc24_build_initramfs_kernel_snap() {
+    local ORIG_SNAP="$1"
+    local TARGET="$2"
+    local OPT_ARG="${3:-}"
+
+    injectKernelPanic=false
+    case "$OPT_ARG" in
+        --inject-kernel-panic-in-initramfs)
+            injectKernelPanic=true
+            ;;
+    esac    
+
+    unsquashfs -d pc-kernel "$ORIG_SNAP"
+    objcopy -O binary -j .initrd pc-kernel/kernel.efi initrd.img
+
+    unmkinitramfs initrd.img initrd
+
+    if [ -d ./extra-initrd ]; then
+        if [ -d ./initrd/early ]; then
+            cp -aT ./extra-initrd ./initrd/main
+        else
+            cp -aT ./extra-initrd ./initrd
+        fi
+    fi
+
+    if [ -d ./initrd/early ]; then
+        cp -a /usr/lib/snapd/snap-bootstrap ./initrd/main/usr/lib/snapd/snap-bootstrap
+        chmod +x ./initrd/main/usr/lib/snapd/snap-bootstrap
+        if [ "$injectKernelPanic" = "true" ]; then
+            # add a kernel panic to the end of the-tool execution
+            echo "echo 'forcibly panicking'; echo c > /proc/sysrq-trigger" >> ./initrd/main/usr/lib/snapd/snap-bootstrap
+        fi
+
+        (cd ./initrd/early; find . | cpio --create --quiet --format=newc --owner=0:0) >initrd.img
+        (cd ./initrd/main; find . | cpio --create --quiet --format=newc --owner=0:0 | zstd -1 -T0) >>initrd.img
+    else
+        cp -a /usr/lib/snapd/snap-bootstrap ./initrd/usr/lib/snapd/snap-bootstrap
+        if [ "$injectKernelPanic" = "true" ]; then
+            # add a kernel panic to the end of the-tool execution
+            echo "echo 'forcibly panicking'; echo c > /proc/sysrq-trigger" >> ./initrd/usr/lib/snapd/snap-bootstrap
+        fi
+
+        (cd ./initrd; find . | cpio --create --quiet --format=newc --owner=0:0 | zstd -1 -T0) >initrd.img
+    fi
+
+    quiet apt install -y systemd-boot-efi systemd-ukify
+    objcopy -O binary -j .linux pc-kernel/kernel.efi linux
+
+    /usr/lib/systemd/ukify build --linux=linux --initrd=initrd.img --output=pc-kernel/kernel.efi
+
+    #shellcheck source=tests/lib/nested.sh
+    . "$TESTSLIB/nested.sh"
+    KEY_NAME=$(nested_get_snakeoil_key)
+
+    SNAKEOIL_KEY="$PWD/$KEY_NAME.key"
+    SNAKEOIL_CERT="$PWD/$KEY_NAME.pem"
+
+    # sign the kernel
+    nested_secboot_sign_kernel pc-kernel "$SNAKEOIL_KEY" "$SNAKEOIL_CERT"
+
+    # copy any extra files that tests may need for the kernel
+    if [ -d ./extra-kernel-snap/ ]; then
+        cp -a ./extra-kernel-snap/* ./pc-kernel
+    fi
+
+    snap pack pc-kernel
+    if [ "$(pwd)" != "$TARGET" ]; then
+        mv pc-kernel_*.snap "$TARGET"
+    fi
+    rm -rf pc-kernel
+}
 
 setup_core_for_testing_by_modify_writable() {
     UNPACK_DIR="$1"
@@ -902,19 +1076,16 @@ setup_reflash_magic() {
     # need to be seeded to proceed with snap install
     snap wait system seed.loaded
 
-    # download the snapd snap for all uc systems except uc16
-    if ! os.query is-core16; then
-        snap download "--channel=${SNAPD_CHANNEL}" snapd
-    fi
-
     # we cannot use "snaps.names tool" here because no snaps are installed yet
     core_name="core"
-    if os.query is-core18; then
+    if is_test_target_core 18; then
         core_name="core18"
-    elif os.query is-core20; then
+    elif is_test_target_core 20; then
         core_name="core20"
-    elif os.query is-core22; then
+    elif is_test_target_core 22; then
         core_name="core22"
+    elif is_test_target_core 24; then
+        core_name="core24"
     fi
     # XXX: we get "error: too early for operation, device not yet
     # seeded or device model not acknowledged" here sometimes. To
@@ -928,10 +1099,12 @@ setup_reflash_magic() {
     UNPACK_DIR="/tmp/$core_name-snap"
     unsquashfs -no-progress -d "$UNPACK_DIR" /var/lib/snapd/snaps/${core_name}_*.snap
 
-    if os.query is-core16; then
+    if os.query is-arm; then
+        snap install ubuntu-image --channel="$UBUNTU_IMAGE_SNAP_CHANNEL" --classic
+    elif is_test_target_core 16; then
         # the new ubuntu-image expects mkfs to support -d option, which was not
         # supported yet by the version of mkfs that shipped with Ubuntu 16.04
-        snap install ubuntu-image --channel="$UBUNTU_IMAGE_SNAP_CHANNEL" --classic
+        snap install ubuntu-image --channel="$OLD_UBUNTU_IMAGE_SNAP_CHANNEL" --classic
     else
         # shellcheck source=tests/lib/image.sh
         . "$TESTSLIB/image.sh"
@@ -951,16 +1124,23 @@ setup_reflash_magic() {
     cp /usr/bin/snap "$IMAGE_HOME"
     export UBUNTU_IMAGE_SNAP_CMD="$IMAGE_HOME/snap"
 
-    if os.query is-core18; then
-        repack_snapd_snap_with_deb_content "$IMAGE_HOME"
+    if is_test_target_core 18; then
+        build_snapd_snap "${IMAGE_HOME}"
         # FIXME: fetch directly once its in the assertion service
         cp "$TESTSLIB/assertions/ubuntu-core-18-amd64.model" "$IMAGE_HOME/pc.model"
-    elif os.query is-core20; then
-        repack_snapd_snap_with_deb_content_and_run_mode_firstboot_tweaks "$IMAGE_HOME"
+    elif is_test_target_core 20; then
+        build_snapd_snap_with_run_mode_firstboot_tweaks "$IMAGE_HOME"
         cp "$TESTSLIB/assertions/ubuntu-core-20-amd64.model" "$IMAGE_HOME/pc.model"
-    elif os.query is-core22; then
-        repack_snapd_snap_with_deb_content_and_run_mode_firstboot_tweaks "$IMAGE_HOME"
-        cp "$TESTSLIB/assertions/ubuntu-core-22-amd64.model" "$IMAGE_HOME/pc.model"
+    elif is_test_target_core 22; then
+        build_snapd_snap_with_run_mode_firstboot_tweaks "$IMAGE_HOME"
+        if os.query is-arm; then
+            cp "$TESTSLIB/assertions/ubuntu-core-22-arm64.model" "$IMAGE_HOME/pc.model"
+        else
+            cp "$TESTSLIB/assertions/ubuntu-core-22-amd64.model" "$IMAGE_HOME/pc.model"
+        fi
+    elif is_test_target_core 24; then
+        build_snapd_snap_with_run_mode_firstboot_tweaks "$IMAGE_HOME"
+        cp "$TESTSLIB/assertions/ubuntu-core-24-amd64.model" "$IMAGE_HOME/pc.model"
     else
         # FIXME: install would be better but we don't have dpkg on
         #        the image
@@ -999,30 +1179,50 @@ EOF
     fi
 
     EXTRA_FUNDAMENTAL=
-    IMAGE_CHANNEL=edge
+    IMAGE_CHANNEL=
     if [ "$KERNEL_CHANNEL" = "$GADGET_CHANNEL" ]; then
         IMAGE_CHANNEL="$KERNEL_CHANNEL"
     else
-        # download pc-kernel snap for the specified channel and set
-        # ubuntu-image channel to that of the gadget, so that we don't
-        # need to download it
-        snap download --channel="$KERNEL_CHANNEL" pc-kernel
-
-        EXTRA_FUNDAMENTAL="--snap $PWD/pc-kernel_*.snap"
         IMAGE_CHANNEL="$GADGET_CHANNEL"
+        if is_test_target_core_le 18; then
+            if is_test_target_core 16; then
+                BRANCH=latest
+            elif is_test_target_core 18; then
+                BRANCH=18
+            fi
+            # download pc-kernel snap for the specified channel and set
+            # ubuntu-image channel to that of the gadget, so that we don't
+            # need to download it. Do this only for UC16/18 as the UC20+
+            # case is considered a few lines below.
+            snap download --basename=pc-kernel --channel="$BRANCH/$KERNEL_CHANNEL" pc-kernel
+            # Repack to prevent reboots as the image channel (which will become
+            # the tracked channel) is different to the kernel channel.
+            unsquashfs -d pc-kernel pc-kernel.snap
+            touch pc-kernel/repacked
+            snap pack --filename=pc-kernel-repacked.snap pc-kernel
+            rm -rf pc-kernel
+            mv pc-kernel-repacked.snap pc-kernel.snap
+            EXTRA_FUNDAMENTAL="--snap $PWD/pc-kernel.snap"
+        fi
     fi
 
-    if os.query is-core20 || os.query is-core22; then
-        if os.query is-core20; then
+    if is_test_target_core_ge 20; then
+        if is_test_target_core 20; then
             BRANCH=20
-        elif os.query is-core22; then
+        elif is_test_target_core 22; then
             BRANCH=22
+        elif is_test_target_core 24; then
+            BRANCH=24
         fi
         snap download --basename=pc-kernel --channel="${BRANCH}/${KERNEL_CHANNEL}" pc-kernel
         # make sure we have the snap
         test -e pc-kernel.snap
         # build the initramfs with our snapd assets into the kernel snap
-        uc20_build_initramfs_kernel_snap "$PWD/pc-kernel.snap" "$IMAGE_HOME"
+        if is_test_target_core_ge 24; then
+            uc24_build_initramfs_kernel_snap "$PWD/pc-kernel.snap" "$IMAGE_HOME"
+        else    
+            uc20_build_initramfs_kernel_snap "$PWD/pc-kernel.snap" "$IMAGE_HOME"
+        fi
         EXTRA_FUNDAMENTAL="--snap $IMAGE_HOME/pc-kernel_*.snap"
 
         # also add debug command line parameters to the kernel command line via
@@ -1030,7 +1230,6 @@ EOF
         snap download --basename=pc --channel="${BRANCH}/${KERNEL_CHANNEL}" pc
         test -e pc.snap
         unsquashfs -d pc-gadget pc.snap
-        
         # TODO: it would be desirable when we need to do in-depth debugging of
         # UC20 runs in google to have snapd.debug=1 always on the kernel command
         # line, but we can't do this universally because the logic for the env
@@ -1041,7 +1240,7 @@ EOF
         # so for now, don't include snapd.debug=1, but eventually it would be
         # nice to have this on
 
-        if [ "$SPREAD_BACKEND" = "google" ]; then
+        if [[ "$SPREAD_BACKEND" =~ google ]]; then
             # the default console settings for snapd aren't super useful in GCE,
             # instead it's more useful to have all console go to ttyS0 which we 
             # can read more easily than tty1 for example
@@ -1078,7 +1277,7 @@ EOF
 
     # on core18 we need to use the modified snapd snap and on core16
     # it is the modified core that contains our freshly build snapd
-    if os.query is-core18 || os.query is-core20 || os.query is-core22; then
+    if is_test_target_core_ge 18; then
         extra_snap=("$IMAGE_HOME"/snapd_*.snap)
     else
         extra_snap=("$IMAGE_HOME"/core_*.snap)
@@ -1091,14 +1290,16 @@ EOF
     fi
 
     # download the core20 snap manually from the specified channel for UC20
-    if os.query is-core20 || os.query is-core22; then
-        if os.query is-core20; then
+    if is_test_target_core_ge 20; then
+        if is_test_target_core 20; then
             BASE=core20
-        elif os.query is-core22; then
+        elif is_test_target_core 22; then
             BASE=core22
+        elif is_test_target_core 24; then
+            BASE=core24
         fi
         snap download "${BASE}" --channel="$BASE_CHANNEL" --basename="${BASE}"
-        
+
         # we want to download the specific channel referenced by $BASE_CHANNEL, 
         # but if we just seed that revision and $BASE_CHANNEL != $IMAGE_CHANNEL,
         # then immediately on booting, snapd will refresh from the revision that
@@ -1121,16 +1322,17 @@ EOF
         else 
             mv "${BASE}.snap" "${IMAGE_HOME}/${BASE}.snap"
         fi
-        
+
         EXTRA_FUNDAMENTAL="$EXTRA_FUNDAMENTAL --snap ${IMAGE_HOME}/${BASE}.snap"
     fi
     local UBUNTU_IMAGE="$GOHOME"/bin/ubuntu-image
-    if os.query is-core16; then
+    if is_test_target_core 16 || os.query is-arm; then
         # ubuntu-image on 16.04 needs to be installed from a snap
         UBUNTU_IMAGE=/snap/bin/ubuntu-image
     fi
     # shellcheck disable=SC2086
     "$UBUNTU_IMAGE" snap \
+                    --image-size 5G \
                     -w "$IMAGE_HOME" "$IMAGE_HOME/pc.model" \
                     --channel "$IMAGE_CHANNEL" \
                     $EXTRA_FUNDAMENTAL \
@@ -1138,7 +1340,9 @@ EOF
                     --output-dir "$IMAGE_HOME"
     rm -f ./pc-kernel_*.{snap,assert} ./pc-kernel.{snap,assert} ./pc_*.{snap,assert} ./snapd_*.{snap,assert} ./core{20,22}.{snap,assert}
 
-    if os.query is-core20 || os.query is-core22; then
+    if os.query is-arm; then
+        LOOP_PARTITION=1
+    elif is_test_target_core_ge 20; then
         # (ab)use ubuntu-seed
         LOOP_PARTITION=2
     else
@@ -1148,7 +1352,7 @@ EOF
     # expand the uc16 and uc18 images a little bit (400M) as it currently will
     # run out of space easily from local spread runs if there are extra files in
     # the project not included in the git ignore and spread ignore, etc.
-    if ! (os.query is-core20 || os.query is-core22); then
+    if is_test_target_core_le 18; then
         # grow the image by 400M
         truncate --size=+400M "$IMAGE_HOME/$IMAGE"
         # fix the GPT table because old versions of parted complain about this 
@@ -1172,7 +1376,7 @@ EOF
     dev=$(basename "$devloop")
 
     # resize the 2nd partition from that loop device to fix the size
-    if ! (os.query is-core20 || os.query is-core22); then
+    if os.query is-core-le 18; then
         resize2fs -p "/dev/mapper/${dev}p${LOOP_PARTITION}"
     fi
 
@@ -1184,7 +1388,7 @@ EOF
     # - built debs
     # - golang archive files and built packages dir
     # - govendor .cache directory and the binary,
-    if os.query is-core16 || os.query is-core18; then
+    if is_test_target_core_le 18; then
         mkdir -p /mnt/user-data/
         # we need to include "core" here because -C option says to ignore 
         # files the way CVS(?!) does, so it ignores files named "core" which
@@ -1199,7 +1403,7 @@ EOF
           --exclude /gopath/pkg/ \
           --include core/ \
           /home/gopath /mnt/user-data/
-    elif os.query is-core20 || os.query is-core22; then
+    else
         # prepare passwd for run-mode-overlay-data
 
         # use /etc/{group,passwd,shadow,gshadow} from the core20 snap, merged
@@ -1237,7 +1441,7 @@ EOF
     fi
 
     # now modify the image writable partition - only possible on uc16 / uc18
-    if os.query is-core16 || os.query is-core18; then
+    if is_test_target_core_le 18; then
         # modify the writable partition of "core" so that we have the
         # test user
         setup_core_for_testing_by_modify_writable "$UNPACK_DIR"
@@ -1247,54 +1451,14 @@ EOF
     umount /mnt
     kpartx -d "$IMAGE_HOME/$IMAGE"
 
-    # the reflash magic
-    # FIXME: ideally in initrd, but this is good enough for now
-    cat > "$IMAGE_HOME/reflash.sh" << EOF
-#!/tmp/busybox sh
-set -e
-set -x
-
-# blow away everything
-OF=/dev/sda
-if [ -e /dev/vda ]; then
-    OF=/dev/vda
-fi
-dd if=/tmp/$IMAGE of=\$OF bs=4M
-# and reboot
-sync
-echo b > /proc/sysrq-trigger
-
-EOF
-
-    cat > "$IMAGE_HOME/prep-reflash.sh" << EOF
-#!/bin/sh -ex
-mount -t tmpfs none /tmp
-cp /bin/busybox /tmp
-cp $IMAGE_HOME/reflash.sh /tmp
-cp $IMAGE_HOME/$IMAGE /tmp
-sync
-
-# re-exec using busybox from /tmp
-exec /tmp/reflash.sh
-
-EOF
-    chmod +x "$IMAGE_HOME/reflash.sh"
-    chmod +x "$IMAGE_HOME/prep-reflash.sh"
-
-    DEVPREFIX=""
-    if os.query is-core20 || os.query is-core22; then
-        DEVPREFIX="/boot"
+    gzip "${IMAGE_HOME}/${IMAGE}"
+    if is_test_target_core 16; then
+        "${TESTSLIB}/uc16-reflash.sh" "${IMAGE_HOME}/${IMAGE}.gz"
+    else
+        "${TESTSLIB}/reflash.sh" "${IMAGE_HOME}/${IMAGE}.gz"
     fi
-    # extract ROOT from /proc/cmdline
-    ROOT=$(sed -e 's/^.*root=//' -e 's/ .*$//' /proc/cmdline)
-    cat >/boot/grub/grub.cfg <<EOF
-set default=0
-set timeout=2
-menuentry 'flash-all-snaps' {
-linux $DEVPREFIX/vmlinuz root=$ROOT ro init=$IMAGE_HOME/prep-reflash.sh console=tty1 console=ttyS0
-initrd $DEVPREFIX/initrd.img
-}
-EOF
+
+    rm -rf "$UNPACK_DIR"
 }
 
 # prepare_ubuntu_core will prepare ubuntu-core 16+
@@ -1318,7 +1482,7 @@ prepare_ubuntu_core() {
     fi
 
     # Wait for the snap command to become available.
-    if [ "$SPREAD_BACKEND" != "external" ]; then
+    if [ "$SPREAD_BACKEND" != "external" ] && [ "$SPREAD_BACKEND" != "testflinger" ]; then
         # shellcheck disable=SC2016
         retry -n 120 --wait 1 sh -c 'test "$(command -v snap)" = /usr/bin/snap && snap version | grep -E -q "snapd +1337.*"'
     fi
@@ -1337,7 +1501,7 @@ prepare_ubuntu_core() {
     done
 
     echo "Ensure the snapd snap is available"
-    if os.query is-core18 || os.query is-core20 || os.query is-core22; then
+    if os.query is-core-ge 18; then
         if ! snap list snapd; then
             echo "snapd snap on core18 is missing"
             snap list
@@ -1354,6 +1518,8 @@ prepare_ubuntu_core() {
             rsync_snap="test-snapd-rsync-core20"
         elif os.query is-core22; then
             rsync_snap="test-snapd-rsync-core22"
+        elif os.query is-core24; then
+            rsync_snap="test-snapd-rsync-core24"
         fi
         snap install --devmode --edge "$rsync_snap"
         snap alias "$rsync_snap".rsync rsync
@@ -1365,7 +1531,7 @@ prepare_ubuntu_core() {
 
     echo "Ensure the core snap is cached"
     # Cache snaps
-    if os.query is-core18 || os.query is-core20 || os.query is-core22; then
+    if os.query is-core-ge 18; then
         if snap list core >& /dev/null; then
             echo "core snap on core18 should not be installed yet"
             snap list
@@ -1381,6 +1547,9 @@ prepare_ubuntu_core() {
         if os.query is-core22; then
             cache_snaps test-snapd-sh-core22
         fi
+        if os.query is-core24; then
+            cache_snaps test-snapd-sh-core24
+        fi
     fi
 
     disable_refreshes
@@ -1388,10 +1557,23 @@ prepare_ubuntu_core() {
 
     # Snapshot the fresh state (including boot/bootenv)
     if ! is_snapd_state_saved; then
+        # Create the file with the initial environment before saving the state
+        tests.env start initial
+
+        # save preinstalled snaps when tests are executed in external systems
+        # the preinstalled snaps shouldn't be removed during tests clean up
+        # this is needed just for external devices because those could be using
+        # custom images with pre-installed snaps which cannot be removed, such
+        # as the network-manager.
+        if [ "$SPREAD_BACKEND" = "external" ] || [ "$SPREAD_BACKEND" = "testflinger" ]; then
+            PREINSTALLED_SNAPS="$(snap list | tail -n +2 | awk '{print $1}' | tr '\n' ' ')"
+            tests.env set initial PREINSTALLED_SNAPS "$PREINSTALLED_SNAPS"
+        fi
 
         # important to remove disabled snaps before calling save_snapd_state
         # or restore will break
         remove_disabled_snaps
+        prepare_memory_limit_override
         setup_experimental_features
         systemctl stop snapd.service snapd.socket
         save_snapd_state
