@@ -22,7 +22,6 @@ package wrappers
 import (
 	"bytes"
 	"fmt"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 	"regexp"
@@ -39,7 +38,7 @@ import (
 // catches units that run /usr/bin/snap (with args), or things in /usr/lib/snapd/
 var execStartRe = regexp.MustCompile(`(?m)^ExecStart=(/usr/bin/snap\s+.*|/usr/lib/snapd/.*)$`)
 
-// snapdToolingMountUnit is the name of the mount unit that makes the
+// snapdToolingMountUnit is the name of the mount unit that provides the snapd tooling
 const SnapdToolingMountUnit = "usr-lib-snapd.mount"
 
 func snapdSkipStart(content []byte) bool {
@@ -49,7 +48,7 @@ func snapdSkipStart(content []byte) bool {
 // snapdUnitSkipStart returns true for units that should not be started
 // automatically
 func snapdUnitSkipStart(unitPath string) (skip bool, err error) {
-	content, err := ioutil.ReadFile(unitPath)
+	content, err := os.ReadFile(unitPath)
 	if err != nil {
 		if os.IsNotExist(err) {
 			// no point in starting units that do not exist
@@ -64,12 +63,13 @@ func writeSnapdToolingMountUnit(sysd systemd.Systemd, prefix string, opts *AddSn
 
 	// TODO: the following comment is wrong, we don't need RequiredBy=snapd here?
 
-	// Not using AddMountUnitFile() because we need
+	// Not using EnsureMountUnitFile() because we need
 	// "RequiredBy=snapd.service"
 
 	content := []byte(fmt.Sprintf(`[Unit]
 Description=Make the snapd snap tooling available for the system
 Before=snapd.service
+Before=systemd-udevd.service
 
 [Mount]
 What=%s/usr/lib/snapd
@@ -141,16 +141,60 @@ type AddSnapdSnapServicesOptions struct {
 	Preseeding bool
 }
 
+// SnapdRestart keeps state of services that need a restart after
+// current symlink of snapd snap (/snap/snapd/current) has been
+// updated. Call Restart method then.
+// Currently, the list of services is static and the absence of a
+// SnapdRestart object (nil), means no service requires a restart.
+type SnapdRestart interface {
+	Restart() error
+}
+
+type snapdRestartImpl struct {
+	Sysd systemd.Systemd
+}
+
+// Restart restarts systemd service units. Call this method after
+// symlink /snap/snapd/current has been updated.
+func (r *snapdRestartImpl) Restart() error {
+	if err := r.Sysd.StartNoBlock([]string{"snapd.apparmor.service"}); err != nil {
+		return err
+	}
+
+	// and finally start snapd.service (it will stop by itself and gets
+	// started by systemd then)
+	// Because of the file lock held on the snapstate by the Overlord, the new
+	// snapd will block there until we release it. For this reason, we cannot
+	// start the unit in blocking mode.
+	// TODO: move/share this responsibility with daemon so that we can make the
+	// start blocking again
+	if err := r.Sysd.StartNoBlock([]string{"snapd.service"}); err != nil {
+		return err
+	}
+	if err := r.Sysd.StartNoBlock([]string{"snapd.seeded.service"}); err != nil {
+		return err
+	}
+	// we cannot start snapd.autoimport in blocking mode because
+	// it has a "After=snapd.seeded.service" which means that on
+	// seeding a "systemctl start" that blocks would hang forever
+	// and we deadlock.
+	if err := r.Sysd.StartNoBlock([]string{"snapd.autoimport.service"}); err != nil {
+		return err
+	}
+
+	return nil
+}
+
 // AddSnapdSnapServices sets up the services based on a given snapd snap in the
 // system.
-func AddSnapdSnapServices(s *snap.Info, opts *AddSnapdSnapServicesOptions, inter Interacter) error {
+func AddSnapdSnapServices(s *snap.Info, opts *AddSnapdSnapServicesOptions, inter Interacter) (SnapdRestart, error) {
 	if snapType := s.Type(); snapType != snap.TypeSnapd {
-		return fmt.Errorf("internal error: adding explicit snapd services for snap %q type %q is unexpected", s.InstanceName(), snapType)
+		return nil, fmt.Errorf("internal error: adding explicit snapd services for snap %q type %q is unexpected", s.InstanceName(), snapType)
 	}
 
 	// we never write snapd services on classic
 	if release.OnClassic {
-		return nil
+		return nil, nil
 	}
 
 	if opts == nil {
@@ -165,24 +209,24 @@ func AddSnapdSnapServices(s *snap.Info, opts *AddSnapdSnapServicesOptions, inter
 	}
 
 	if err := writeSnapdToolingMountUnit(sysd, s.MountDir(), opts); err != nil {
-		return err
+		return nil, err
 	}
 
 	serviceUnits, err := filepath.Glob(filepath.Join(s.MountDir(), "lib/systemd/system/*.service"))
 	if err != nil {
-		return err
+		return nil, err
 	}
 	socketUnits, err := filepath.Glob(filepath.Join(s.MountDir(), "lib/systemd/system/*.socket"))
 	if err != nil {
-		return err
+		return nil, err
 	}
 	timerUnits, err := filepath.Glob(filepath.Join(s.MountDir(), "lib/systemd/system/*.timer"))
 	if err != nil {
-		return err
+		return nil, err
 	}
 	targetUnits, err := filepath.Glob(filepath.Join(s.MountDir(), "lib/systemd/system/*.target"))
 	if err != nil {
-		return err
+		return nil, err
 	}
 	units := append(socketUnits, serviceUnits...)
 	units = append(units, timerUnits...)
@@ -192,11 +236,11 @@ func AddSnapdSnapServices(s *snap.Info, opts *AddSnapdSnapServicesOptions, inter
 	for _, unit := range units {
 		st, err := os.Stat(unit)
 		if err != nil {
-			return err
+			return nil, err
 		}
-		content, err := ioutil.ReadFile(unit)
+		content, err := os.ReadFile(unit)
 		if err != nil {
-			return err
+			return nil, err
 		}
 		if execStartRe.Match(content) {
 			content = execStartRe.ReplaceAll(content, []byte(fmt.Sprintf("ExecStart=%s$1", s.MountDir())))
@@ -215,11 +259,11 @@ func AddSnapdSnapServices(s *snap.Info, opts *AddSnapdSnapServicesOptions, inter
 	changed, removed, err := osutil.EnsureDirStateGlobs(dirs.SnapServicesDir, globs, snapdUnits)
 	if err != nil {
 		// TODO: uhhhh, what do we do in this case?
-		return err
+		return nil, err
 	}
 	if (len(changed) + len(removed)) == 0 {
 		// nothing to do
-		return nil
+		return nil, nil
 	}
 
 	// stop all removed units first
@@ -236,7 +280,7 @@ func AddSnapdSnapServices(s *snap.Info, opts *AddSnapdSnapServicesOptions, inter
 	// daemon-reload so that we get the new services
 	if len(changed) > 0 {
 		if err := sysd.DaemonReload(); err != nil {
-			return err
+			return nil, err
 		}
 	}
 
@@ -254,14 +298,14 @@ func AddSnapdSnapServices(s *snap.Info, opts *AddSnapdSnapServicesOptions, inter
 		if !opts.Preseeding {
 			enabled, err := sysd.IsEnabled(unit)
 			if err != nil {
-				return err
+				return nil, err
 			}
 			if enabled {
 				continue
 			}
 		}
 		if err := sysd.EnableNoReload([]string{unit}); err != nil {
-			return err
+			return nil, err
 		}
 	}
 
@@ -282,7 +326,7 @@ func AddSnapdSnapServices(s *snap.Info, opts *AddSnapdSnapServicesOptions, inter
 			// exists before we are fully seeded).
 			isActive, err := sysd.IsActive(unit)
 			if err != nil {
-				return err
+				return nil, err
 			}
 
 			serviceUnits := []string{unit}
@@ -291,53 +335,36 @@ func AddSnapdSnapServices(s *snap.Info, opts *AddSnapdSnapServicesOptions, inter
 				// this will also bring down snapd itself
 				if unit != "snapd.socket" {
 					if err := sysd.Restart(serviceUnits); err != nil {
-						return err
+						return nil, err
 					}
 				}
 			} else {
 				if err := sysd.Start(serviceUnits); err != nil {
-					return err
+					return nil, err
 				}
 			}
 		}
 	}
 
-	// and finally start snapd.service (it will stop by itself and gets
-	// started by systemd then)
-	// Because of the file lock held on the snapstate by the Overlord, the new
-	// snapd will block there until we release it. For this reason, we cannot
-	// start the unit in blocking mode.
-	// TODO: move/share this responsibility with daemon so that we can make the
-	// start blocking again
-	if err := sysd.StartNoBlock([]string{"snapd.service"}); err != nil {
-		return err
-	}
-	if err := sysd.StartNoBlock([]string{"snapd.seeded.service"}); err != nil {
-		return err
-	}
-	// we cannot start snapd.autoimport in blocking mode because
-	// it has a "After=snapd.seeded.service" which means that on
-	// seeding a "systemctl start" that blocks would hang forever
-	// and we deadlock.
-	if err := sysd.StartNoBlock([]string{"snapd.autoimport.service"}); err != nil {
-		return err
-	}
-
 	// Handle the user services
-	if err := writeSnapdUserServicesOnCore(s, inter); err != nil {
-		return err
+	if err := writeSnapdUserServicesOnCore(s, opts, inter); err != nil {
+		return nil, err
 	}
 
 	// Handle D-Bus configuration
 	if err := writeSnapdDbusConfigOnCore(s); err != nil {
-		return err
+		return nil, err
 	}
 
 	if err := writeSnapdDbusActivationOnCore(s); err != nil {
-		return err
+		return nil, err
 	}
 
-	return nil
+	if err := writeSnapdDesktopFilesOnCore(s); err != nil {
+		return nil, err
+	}
+
+	return &snapdRestartImpl{Sysd: sysd}, nil
 }
 
 // undoSnapdUserServicesOnCore attempts to remove services that were deployed in
@@ -431,7 +458,7 @@ func undoSnapdServicesOnCore(s *snap.Info, sysd systemd.Systemd) error {
 	return nil
 }
 
-func writeSnapdUserServicesOnCore(s *snap.Info, inter Interacter) error {
+func writeSnapdUserServicesOnCore(s *snap.Info, opts *AddSnapdSnapServicesOptions, inter Interacter) error {
 	// Ensure /etc/systemd/user exists
 	if err := os.MkdirAll(dirs.SnapUserServicesDir, 0755); err != nil {
 		return err
@@ -456,7 +483,7 @@ func writeSnapdUserServicesOnCore(s *snap.Info, inter Interacter) error {
 		if err != nil {
 			return err
 		}
-		content, err := ioutil.ReadFile(unit)
+		content, err := os.ReadFile(unit)
 		if err != nil {
 			return err
 		}
@@ -500,6 +527,12 @@ func writeSnapdUserServicesOnCore(s *snap.Info, inter Interacter) error {
 		}
 	}
 
+	if !opts.Preseeding {
+		if err := userDaemonReload(); err != nil {
+			logger.Noticef("failed to reload user systemd instances: %v", err)
+		}
+	}
+
 	return nil
 }
 
@@ -625,6 +658,9 @@ func undoSnapdDbusConfigOnCore() error {
 	return err
 }
 
+// Service files that have been written by snapd.
+// Only ever append this list, if a file is no longer installed it needs to
+// remain here so a newer version of snapd can remove it.
 var dbusSessionServices = []string{
 	"io.snapcraft.Launcher.service",
 	"io.snapcraft.Prompt.service",
@@ -657,6 +693,38 @@ func undoSnapdDbusActivationOnCore() error {
 	return err
 }
 
+// Desktop files that have been written by snapd.
+// Only ever append this list, if a file is no longer installed it needs to
+// remain here so a newer version of snapd can remove it.
+var snapdDesktopFileNames = []string{
+	"io.snapcraft.SessionAgent.desktop",
+	"snap-handle-link.desktop",
+}
+
+func writeSnapdDesktopFilesOnCore(s *snap.Info) error {
+	// Ensure /var/lib/snapd/desktop/applications exists
+	if err := os.MkdirAll(dirs.SnapDesktopFilesDir, 0755); err != nil {
+		return err
+	}
+
+	desktopFiles := make(map[string]osutil.FileState, len(snapdDesktopFileNames))
+	for _, fileName := range snapdDesktopFileNames {
+		filePathInSnap := filepath.Join(s.MountDir(), "usr/share/applications", fileName)
+		if !osutil.FileExists(filePathInSnap) {
+			continue
+		}
+		desktopFiles[fileName] = &osutil.FileReference{Path: filePathInSnap}
+	}
+
+	_, _, err := osutil.EnsureDirStateGlobs(dirs.SnapDesktopFilesDir, snapdDesktopFileNames, desktopFiles)
+	return err
+}
+
+func undoSnapdDesktopFilesOnCore(s *snap.Info) error {
+	_, _, err := osutil.EnsureDirStateGlobs(dirs.SnapDesktopFilesDir, snapdDesktopFileNames, nil)
+	return err
+}
+
 // RemoveSnapdSnapServicesOnCore removes the snapd services generated by a prior
 // call to AddSnapdSnapServices. The core snap is used as the reference for
 // restoring the system state, making this undo helper suitable for use when
@@ -673,6 +741,9 @@ func RemoveSnapdSnapServicesOnCore(s *snap.Info, inter Interacter) error {
 
 	sysd := systemd.NewUnderRoot(dirs.GlobalRootDir, systemd.SystemMode, inter)
 
+	if err := undoSnapdDesktopFilesOnCore(s); err != nil {
+		return err
+	}
 	if err := undoSnapdDbusActivationOnCore(); err != nil {
 		return err
 	}
